/*
 * $Id: SimulatedClock.java,v 1.1 2009/04/01 13:57:10 sirocic Exp $
 */
package flexray.clock;

import java.util.ArrayList;
import java.util.TreeMap;


/**
 * A multithreaded master/slave clock.
 *
 * @author Dirk Frsterling
 *
 */
public class SimulatedClock implements Runnable, Clock, ClockSlave, ClockSyncSlave  {

    private long currentTick = 0; // volatile needed?
    private long deviation = 0;
    private long offset;
    private long parentTicksPerOwnTick = 0;
    // get the simulation clock (for pause/resume or as default parent)
    private SimulationClock simulationClock = SimulationClock.getSimulationClock();
    private Clock parentClock = null;

    private static int nameSerials = 0;
    private long tickDelay = 0;
    private int mySlaveID;
    private String name = null;
    private boolean clockActive = false;

    private boolean ready = false;

    private boolean saveToHistory = false;
    private TreeMap<Long,Long> clockHistory = new TreeMap<Long,Long>();

    private Object clockSlaveSync = new Object();
    private ArrayList<ClockSlave> clockSlaves = new ArrayList<ClockSlave>();
    private ArrayList<Integer> slavesActive = new ArrayList<Integer>();
    private int activeSlaveCount = 0;

    /**
     * Hidden default constructor.
     * Just set the default parent Clock here.
     *
     */
    private SimulatedClock() {
        parentClock = simulationClock;
    }

    /**
     * Create clock with a "frequency" measured in ticks of the parent clock
     * per own tick, a time offset and deviation from uniform ticking.
     *
     * If the given parent clock is <tt>null</tt>, then the default parent
     * clock (SimulationClock) will be used.
     *
     * @param pParent parent clock
     * @param pName name of this clock
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     * @param pOffset tick offset measured in ticks of the parent clock
     * @param pDeviation deviation in percent. Resolution is 1%
     */
    public SimulatedClock( Clock pParent, String pName, int pTicks, int pOffset, float pDeviation ) {
        this();
        parentTicksPerOwnTick = pTicks;
        if( pParent != null ) parentClock = pParent;
        name = pName;

        // static numbers for automatic naming.
        nameSerials++;

        setDeviation( pDeviation );
        offset = pOffset;

        //  add myself to the parent
        mySlaveID = parentClock.addSlave( this );
    }

    /**
     * construct a clock just using a name and the "frequency",
     * using the global default parent clock (SimulationClock)
     *
     * @param pName name of this clock
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     */
    public SimulatedClock( String pName, int pTicks ) {
        this( null, pName, pTicks, 0, 0.0f );
    }

    /**
     * construct a clock just using a name and the "frequency",
     * using the global default parent clock (SimulationClock)
     *
     * @param pParent parent clock
     * @param pName name of this clock
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     */
    public SimulatedClock( Clock pParent, String pName, int pTicks ) {
        this( pParent, pName, pTicks, 0, 0.0f );
    }

    /**
     * construct a clock with default parent clock and default name,
     *  defining "frequency", offset and deviation
     *
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     * @param pOffset tick offset measured in ticks of the parent clock
     * @param pDeviation deviation in percent. Resolution is 1%
     */
    public SimulatedClock( int pTicks, int pOffset, float pDeviation ) {
        this( null, "Clock-" + nameSerials, pTicks, pOffset,  pDeviation);
    }

    /**
     * construct a clock with parent clock, default name,
     * defining "frequency", offset and deviation
     *
     * @param pParent parent clock
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     * @param pOffset tick offset measured in ticks of the parent clock
     * @param pDeviation deviation in percent. Resolution is 1%
     */
    public SimulatedClock( Clock pParent, int pTicks, int pOffset, float pDeviation ) {
        this( pParent, "Clock-" + nameSerials, pTicks, pOffset,  pDeviation);
    }

    /**
     * construct a clock with default parent clock and default name,
     * defining "frequency", and deviation
     *
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     * @param pDeviation deviation in percent. Resolution is 1%
     */
    public SimulatedClock( int pTicks, float pDeviation ) {
        this(pTicks, 0, pDeviation );
    }

    /**
     * construct a clock with default parent clock,
     * defining name, "frequency", and deviation
     *
     * @param pName name of this clock
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     * @param pDeviation deviation in percent. Resolution is 1%
     */
    public SimulatedClock( String pName, int pTicks, float pDeviation ) {
        this( null, pName, pTicks, 0, pDeviation );
    }

    /**
     * construct a clock, defining parent clock,
     * "frequency", and deviation
     *
     * @param pParent parent clock
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     * @param pDeviation deviation in percent. Resolution is 1%
     */
    public SimulatedClock( Clock pParent, int pTicks, float pDeviation ) {
        this( pParent, pTicks, 0, pDeviation );
    }

    /**
     * construct a clock, defining parent clock,
     * name, "frequency", and deviation
     *
     * @param pParent parent clock
     * @param pName name of this clock
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     * @param pDeviation deviation in percent. Resolution is 1%
     */
    public SimulatedClock( Clock pParent, String pName, int pTicks, float pDeviation ) {
        this( pParent, pName, pTicks, 0, pDeviation );
    }

    /**
     * construct a non-deviating clock with default parent clock and default name,
     * defining "frequency".
     *
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     * @param pOffset tick offset measured in ticks of the parent clock
     */
    public SimulatedClock( int pTicks, int pOffset ) {
        this( pTicks, pOffset, 0.0f);
    }

    /**
     * construct a non-deviating clock with default parent clock,
     * defining name, "frequency", and offset
     *
     * @param pName name of this clock
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     * @param pOffset tick offset measured in ticks of the parent clock
     */
    public SimulatedClock( String pName, int pTicks, int pOffset ) {
        this( null, pName, pTicks, pOffset, 0.0f );
    }

    /**
     * construct a non-deviating clock,
     * defining parent clock, name, "frequency", and offset
     *
     * @param pParent parent clock
     * @param pName name of this clock
     * @param pTicks ticks that the parent must tick to let this clock tick once.
     * @param pOffset tick offset measured in ticks of the parent clock
     */
    public SimulatedClock( Clock pParent, String pName, int pTicks, int pOffset ) {
        this( pParent, pName, pTicks, pOffset, 0.0f );
    }


    /**
     * Minimum constructor: Create a non-deviating, non-offset
     * clock with default parent and default name, just
     * defining the "frequency".
     *
     * @param pTicks Simulationszeitticks pro Tick
     */
    public SimulatedClock( int pTicks ) {
        this( pTicks, 0, 0.0f  );
    } // Quartz


    /**
     * Minimum constructor: Create a non-deviating, non-offset
     * clock with default name,
     * defining clock parent and "frequency".
     *
     * @param pParent parent clock
     * @param pTicks Simulationszeitticks pro Tick
     */
    public SimulatedClock( Clock pParent, int pTicks ) {
        this( pParent, pTicks, 0, 0.0f );
    }

    /* -------------------------------------------------------------------- */


    /**
     * @return this Clock's name (from interface Clock)
     */
    public String getClockName() {
        return name;
    }

    /**
     * @return this Clock's name (from interface ClockSlave)
     */
    public String getSlaveName() {
        return name;
    }

    /**
     * @return this slave's ID
     */
    public int getSlaveID() {
        return mySlaveID;
    }


    public long getCurrentStep() {
        return currentTick;
    }

    /**
     * stop this Clock
     */
    public void stopClock() {
        clockActive = false;
    }

    public void pauseClock() {
        simulationClock.pauseClock();
    }

    /**
     * Change deviation.
     * A deviation of -100% would mean that this clock stops ticking, so
     * the minumum deviation allowed is -99%
     * The maximum deviation is capped to 99% for symmetry reasons.
     *
     * The parameter is float, just because of consistency reasons which
     * originate in construcor design: deviation was designed as float
     * to allow more distinct constructors. The resolution is 1% (1.0)
     *
     * @param pDeviation deviation in percent.
     */
    private void setDeviation( float pDeviation ) {
        if( pDeviation < -100 )
            throw new IllegalArgumentException( "Cannot run clock slower than stopped. Set deviation  >= -100." );

        if( pDeviation > 100 )
            throw new IllegalArgumentException( "Cannot run clock faster than double speed.");

        deviation = (int)pDeviation;
    }


    // -----------------------------------------------

    /**
     * Save this clock's ticks, relative to the SimulationClock's ticks
     * (comes handy for drawing time graphs)
     */
    private synchronized void saveToHistory() {
        if ( ! clockHistory.containsKey( Long.valueOf( currentTick ) ))
            clockHistory.put( Long.valueOf( currentTick ), Long.valueOf(simulationClock.getCurrentTick()) );
        if( currentTick > 20000 ) clockHistory.clear();
    }


    /**
     * get the saved history
     * @return TreeMap with <simuliationTicks,clockTicks>.
     */
    public TreeMap<Long,Long> getHistory() {
        return clockHistory;
    }


    /**
     * Add a slave clock to this clock
     * @param slave
     * @return the id (hashcode) of the new slave
     */
    public int addSlave( ClockSlave slave ) {
        clockSlaves.add( slave );
        slave.setParent( this );
        return slave.hashCode();
    }

    public void setParent(Clock pParentClock) {
        parentClock = pParentClock;
    }

    /**
     * Receive ready signal from slave. It's work has been done,
     * so - at his behalf - we're allowed to advance the clock.
     * (However, this clock will wait for <u>all</u> clients)
     * @param slaveID the ide of the slave that sent the ready signal.
     */
    public void slaveReadySignal( int slaveID ) {
        synchronized( clockSlaveSync ) {
            if( slavesActive.contains( Integer.valueOf( slaveID ) ) ) {
                activeSlaveCount--;
                slavesActive.remove( Integer.valueOf( slaveID ) );
                clockSlaveSync.notify();
            }
        }
    } // slaveReadySignal()

    /*
     * adjust this clock to its parent.
     *
    public void adjust() {
        currentTick = tickForParentTick( parentClock.getCurrentTick() );
        System.err.println( "adjusted to: " + currentTick + " from " + parentClock.getCurrentTick() );
        for(ClockSlave c: clockSlaves) {
            c.adjust();
        }
    }
    */

    /**
     * Calculate the number of ticks this clock would tick if the parent
     * will tick the number of ticks specified in the paremeter.
     * @param parentTick the parent ticks we want to know this clock's ticks for.
     * @return calculated ticks, this clock would have advanced for <tt>parentTick</tt> parentTicks.
     */
    public long tickForParentTick( long parentTick ) {
        return ((((parentTick - offset)*100L) / parentTicksPerOwnTick) * (deviation + 100L)) / 10000L;
    }

    /**
     * Calculate a basic parent tick value that would be necessary to
     * advance this clock to the specified tick
     * @param tick the number of ticks we want to askt parentTicks for
     * @return parent ticks necessary to advance this clock to the tick specified.
     */
    /*
    public long parentTickForTick( long tick ) {
        return ( (10000*tick*parentTicksPerOwnTick) / ((deviation + 100)*100)) + offset;
    }
    */


    /**
     * from the span of ticks that this clock's parent clock will
     * tick during <tt>tick</tt> of this clock, calculate the lowest
     * possible value.
     *
     * Example: If a slave ticks every 10 parent ticks, then from
     * parent tick 10..19, the slave will be "within" it's second
     * tick (tick 1). The lowest parent tick for this is 10.
     *
     * @return lowest parent tick for that this clock will be in tick <tt>tick</tt>
     */
    /*
    public long lowestParentTickForTick( long tick ) {
        long probeParentTick = parentTickForTick( tick );
        while ( tickForParentTick( probeParentTick ) >= tick ) {
              probeParentTick--;
        }
        probeParentTick++;

        return (probeParentTick<0?0:probeParentTick);
    } // lowestParentTickForTick
    */

    /**
     * from the span of ticks that this clock's parent clock will
     * tick during <tt>tick</tt> of this clock, calculate the highest
     * possible value.
     *
     * Example: If a slave ticks every 10 parent ticks, then from
     * parent tick 10..19, the slave will be "within" it's second
     * tick (tick 1). The highest parent tick for this is 19.
     *
     * @return highest parent tick for that this clock will be in tick <tt>tick</tt>
     */
    /*
    public long highestParentTickForTick( long tick ) {
        long probeParentTick = parentTickForTick( tick );
        while ( tickForParentTick( probeParentTick ) <= tick ) {
            probeParentTick++;
        }
        return probeParentTick-1;
    } // highestParentTickForTick
    */

    /**
     * ask if this clock is ready to proceed
     * @return false if this clock is still busy, or true if not.
     */
    public synchronized boolean isReady() {
        return ready;
    }


    public long getTickDelay() {
        return tickDelay;
    }

    public void setTickDelay( long pTickDelay ) {
        tickDelay = pTickDelay;
    }

    /**
     * return this clocks current tick value
     *
     * @return current tick
     */
    public synchronized long getCurrentTick() {
        //if( id == 0 )
          //  System.err.println( "primary: " + currentTick );
        return currentTick;
    }

    /**
     * wait for this clock's slaves to enter "ready" state.
     */
    private void waitForSlaves() {
        while( activeSlaveCount > 0 ) {
            synchronized( clockSlaveSync ) {
                // limit the wait() in case this method got called
                // after the notify() was called on this object.
                try {
                    clockSlaveSync.wait(5);
                } catch( InterruptedException ie ) {
                    // doesn't matter.
                }
            }
        }
    } // waitForSlaves


    /**
     * This Runnable's run(). This is where the clockwork is done ;-)
     */
    public void run() {
        clockActive = true; // start with an active Clock
        while( clockActive ) {
            try {
                synchronized( parentClock ) {
                    ready = true;

                    // signal: "ready to wait for next tick"
                    parentClock.slaveReadySignal(  mySlaveID );

                    // wait for the simulation clock.
                    parentClock.wait();
                    ready = false;
                }

                long saveTick;
                /*
                 * synchronize because of the <code>long</code> variables.
                 */
                synchronized( this ) {
                    saveTick = currentTick;
                    long offtest = parentClock.getCurrentTick();
                    if( offtest >= offset ) {
                        currentTick = tickForParentTick( offtest );
                    }
                }

                // if "really ticked"
                if( saveTick != currentTick ) {
                    if( saveToHistory ) saveToHistory();
                    // tick changed: notify clients.
                    synchronized ( this ) {
                        activeSlaveCount = this.clockSlaves.size();
                        for( ClockSlave c : clockSlaves ) {
                            slavesActive.add( Integer.valueOf( c.hashCode() ) );
                        }
                        this.notifyAll();
                    }
                    waitForSlaves();
                    if( !simulationClock.isStepping() ) Thread.sleep( tickDelay );
                } // if "really ticked"
            } catch( InterruptedException ie ) {
                // FIXME: it might be that not all work has been done here
                // if the thread was interrupted.
            }
        } // while()
    } // run()

    /*
     * implementation of default toString()
     */
    @Override
    public String toString() {
        return name;
    }

}
