/*
 * $Id: SimulationClock.java,v 1.1 2009/04/01 13:57:10 sirocic Exp $
 */
package flexray.clock;

import static flexray.clock.ClockDirection.*;

import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * SimulationClock is the global time source. By the means of the
 * singleton pattern this clock is forced to be globally unique.
 *
 * @author dirk
 */
public class SimulationClock extends Thread implements Clock {

    // holding variable for singleton pattern
    private static volatile SimulationClock simulationClock = null;

    // is the clock currently running or suspended?
    private boolean clockActive = false;
    private boolean oneMoreTick = false;

    private ClockSyncSlave syncTo = null;
    private long slaveSyncStep = 0;
    private int preSyncTickDelay = 0;
    private boolean stepping = false;

    // has the clock been stopped completely (for discarding)?
    private boolean clockStopped = true;

    // has the clock ever been started ?
    private boolean clockStarted = false;

    // additional delay between ticks (additional to the time
    // that is consumed by waiting for the clients
    private int tickDelay = 0;

    // running direction of the clock.
    private ClockDirection clockDirection = FORWARD;

    // synchronization object for the clock's clients
    private Object clockClientSync = new Object();

    // List of ClockClients to notify and wait for
    private ArrayList<ClockSlave> clockClients = new ArrayList<ClockSlave>();

    // List of IDs of clients that didn't signal ready state since the last tick
    private ArrayList<Integer> clientsActive = new ArrayList<Integer>();

    // counting variable for active Clients.
    private int activeClientCount = 0;

    // current tick count / time value
    private long currentTick = 0; // needs to be volatile?

    // the value at wich the clock stops working.
    // at 1ms delay this is over 14000 years.
    private static final long maxTick = Long.MAX_VALUE/2000;

    /**
     * Keep constructors private (singleton).
     */
    private SimulationClock() {
        // empty
    }

    /**
     * Construct a SimulationClock with initial tick delay.
     * @param pTickDelay initial tick delay
     */
    private SimulationClock( int pTickDelay ) {
        tickDelay = pTickDelay;
    }

    /**
     * Get the instance of the SimulationClock. Create it, if it's
     * not yet available.
     * @return the single instance of the simulation clock
     */
    public static SimulationClock getSimulationClock() {
        if( simulationClock == null ) {
            simulationClock = new SimulationClock();
        }
        return simulationClock;
    }

    /**
     * Get the instance of the SimulationClock and simultaneously
     * set the tick delay. The instance is created if ir does not yet exist.
     * @param pTickDelay the tickDelay to preset
     * @return the instance of the simulation clock, parametrized to a delay of <code>pTickDelay</code>
     */
    public static SimulationClock getSimulationClock( int pTickDelay ) {
        if( simulationClock == null ) {
            // create the instance
            simulationClock = new SimulationClock( pTickDelay );
        } else {
            // just set the delay.
            simulationClock.setTickDelay( pTickDelay );
        }
        return simulationClock;
    } // getSimulationClock

    /**
     * Dummy for implicitly checking the availability
     * of the SimulationClock instance
     */
    private void check() {
        // dummy
    } // check

    /* -------------------------------------------- */

    public String getClockName() {
        return "SimulationClock";
    }

    /**
     * Return the tickDelay which adds a delay after every tick.
     * @return currentyl set tickDelay
     */
    public int getTickDelay() {
        return tickDelay;
    } // getTickDelay()

    public boolean isStepping() {
        return stepping;
    }

    /**
     * Change tickDelay. This method pauses the clock during
     * the operation (if it is not already paused).
     *
     * @param pTickDelay
     */
    public  void setTickDelay( int pTickDelay ) {
        boolean running = clockActive;
        simulationClock.check();
        if (running) pauseClock();
        tickDelay = pTickDelay;
        if (running) resumeClock();
    } // setTickDelay()

    /**
     * Return the current tick value ("time")
     * @return the current system simulation tick
     */
    public long getCurrentTick () {
        return currentTick;
    } // getCurrentTick()


    /**
     * Return the highest possible tick value.
     * @return highest possible tick value
     */
    public long getMaxTick() {
        return maxTick;
    }


    /* -------------------------------------------- */

    /**
     * Return the SimulationClock's pause state. <br>
     * Please note, that if the clock only can be in pause mode if it
     * has been started.
     * @return true if the clock is started but in pause mode, else return false.
     */
    public boolean isPaused() {
        return clockStarted && !clockActive;
    }

    /**
     * Pause the simulation clock. This consists of two steps:
     * <ol><li>Setting the flag to pause the clock</li>
     * <li>waiting for clock clients to call back</li></ol><br>
     * This ensures that subsequent calls to the getters will produce
     * up-to-date values. This method does nothing if the clock already
     * is inactive. I won't rely on isPaused() here, because isPaused()
     * reports pause mode even if the clock never has been started.
     */
    public void pauseClock() {
        simulationClock.check();
        if( clockActive == false ) return;
        clockActive = false;
        waitForSlaves();
    } // pauseClock()

    /**
     * Resume the paused clock. To make the state change known, the thread has
     * to be interrupted (and thus leaving the wait()).<br>
     * Don't do anything if the clock is still active.
     */
    public void resumeClock() {
        simulationClock.check();
        if( clockActive ) return;
        clockActive = true;
        this.interrupt();
    } // resumeClock();

    /**
     * Tick one more step.
     */
    public void tick() {
        if( !isPaused() ) pauseClock();
        oneMoreTick = true;
        resumeClock();
    }

    /**
     * Make the simulation run as fast as possible until the specified
     * ClockSyncSlave "ticks".
     * <br>
     * What does this mean?
     * <br>
     * In <tt>FORWARD</tt> mode, the <tt>ClockSyncSlave</tt> must report
     * <tt>slaveSyncStep+1</tt> through <tt>.getCurrentStep()</tt>. If this happens, the
     * <tt>SimulationClock</tt> gets paused automatically.
     * <br>
     * In <tt>BACKWARD</tt> mode, things are a bit more complicated: The <tt>SimulationClock</tt>
     * does not know when the next backward step will cause <tt>.getCurrentStep()</tt> to
     * report <tt>slaveSyncStep-1</tt>. Therefore, the <tt>SimulationClock</tt> runs backwards until
     * <tt>.getCurrentStep()</tt> reports <tt>slaveSyncStep-2</tt>. The single step error
     * will then be repaired by moving one step forwards. After this step, <tt>.getcurrentStep()</tt>
     * reports <tt>slaveSyncStep-1</tt> for the first time. Having reached this point, the
     * <tt>SimulationClock</tt> will be paused automatically.
     * <br>
     * <b>WARNING</b><br>
     * Unfortunately, this only works if ClockSyncSlaves are implemented in a cooperative way.
     *
     * @param css The ClockSyncSlave which dictates the next pause.
     */
    public void tick( ClockSyncSlave css ) {
            pauseClock();
            preSyncTickDelay = getTickDelay();
            stepping = true;
            setTickDelay( 0 );
            syncTo = css;
            slaveSyncStep = syncTo.getCurrentStep();
            resumeClock(); // will be paused in main loop
    }

    /**
     * Set the clock's ticking direction
     * @param pClockDirection
     */
    public void setDirection( ClockDirection pClockDirection ) {
        // status merken, da pauseClock() aufgerufen werden kann.
        boolean running = clockActive;
        if (running) pauseClock();
        clockDirection = pClockDirection;
        if (running) resumeClock();
    }

    /**
     * return the clock's ticking direction
     * FIXME: synchronized. Really?
     * @return the simulation clock's tick direction
     */
    public synchronized ClockDirection getDirection() {
        return clockDirection;
    } // getDirection()

    /**
     * Start the clock's thread, selecting the running mode.
     * @param active start the clock in paused (false) or running (true) mode.
     */
    public void startClock( boolean active ) {
        if( clockStarted ) return;
        clockActive = active;
        simulationClock.start();
        clockStarted = true;
    } // startClock()

    /**
     * start the Clock, initially being active ("unpaused")
     */
    public void startClock() {
        startClock( true );
    }

    /**
     * stop the clock and prepare it for being discarded.
     */
    public void stopClock() {
        if( !clockStarted ) return;
        pauseClock();
        clockStopped = true;
        interrupt();
    } // stopClock()


    /**
     * Add a clock client and return a client id.
     * The id is the client Object's  hash code.
     * FIXME: synchronized and waitForClients()?
     * @param client
     * @return a client ID.
     */
    public synchronized int addSlave( ClockSlave client ) {
        waitForSlaves();
        simulationClock.clockClients.add( client );
        return client.hashCode();
    }

    /**
     * Receive a client's signal telling that it's ready to
     * receive the next clock tick.
     * @param clientID the id of the client that sent the signal
     */
    public void slaveReadySignal( int clientID ) {
        synchronized( clockClientSync ) {
            if( clientsActive.contains( Integer.valueOf(clientID) ) ) {
                activeClientCount--;
                clientsActive.remove( Integer.valueOf(clientID) );
                clockClientSync.notify();
            }
        }
    } // clientReadySignal()

    /**
     * Wait (blocking) for this clock's clients to call back.
     */
    private void waitForSlaves() {
        while( activeClientCount > 0 ) {
            synchronized( clockClientSync ) {
                try {
                    clockClientSync.wait(5);
                } catch( InterruptedException ie ) {
                    // interrupt is irrelevant
                }
            }
        }
    }

    /* -------------------------------------------- */

    /**
     * the run method of the simulation clock thread.
     * Here lies the control of evertything
     */
    @Override
    public void run() {
        /*
         * Clock thread has just been started: Reset the "stopped"
         * state and enter the main loop.
         */
        simulationClock.check();
        clockStopped = false;
        while( !clockStopped ) {
            try {

                /*
                 * If the clock is active ("unpaused"), then advance and coordinate
                 * all clients.
                 */
                if( clockActive ) {
                    // do the tick delay if in non-pause mode
                    TimeUnit.MILLISECONDS.sleep( tickDelay  );

                    /*
                     * synchronize the following to acheive two things
                     * 1) prevent clients from getting garbled time values
                     *    or different time values even if they ask at
                     *    "the same time"
                     * 2) getting the monitor to allow notifyAll()
                     */
                    synchronized( simulationClock ) {
                        switch( clockDirection ) {
                        case FORWARD:
                            if( currentTick < maxTick ) {
                                currentTick++;
                            } else {
                                pauseClock();
                            }
                            break;
                        case BACKWARD:
                            if( currentTick >= 1 ) {
                                currentTick--;
                            } else {
                                pauseClock();
                            }
                        } // switch

                        // We have ticked. Notify all slaves now.
                        activeClientCount = this.clockClients.size();
                        for( ClockSlave c : clockClients ) {
                            clientsActive.add( Integer.valueOf( c.hashCode() ) );
                        }
                        simulationClock.notifyAll();
                    } // synchronized

                    // wait for all slaves (... which wait for theirs ) to be ready again.
                    waitForSlaves();

                    /* oneMoreTick indicates, that one more tick had to take place.
                     * If the syncTo object is not null, then there was a backward
                     * skip in place. In this case, just one step forward had to be
                     * done before switching back to BACKWARD mode.
                     * This is ugly, but allows us to go back to the first simulationTick
                     * of the "step" that is one "step" before the current "step" of the
                     * ClockSyncSlave.
                     */
                    if( oneMoreTick ) {
                        pauseClock();
                        oneMoreTick = false;
                        stepping = false;
                        if( syncTo != null ) {
                            syncTo = null;
                            setDirection( BACKWARD );
                            setTickDelay( preSyncTickDelay );
                        }
                    }

                    if( syncTo != null ) {
                        switch( clockDirection ) {
                        case FORWARD:
                            if ( syncTo.getCurrentStep() == (slaveSyncStep + 1) ) {
                                pauseClock();
                                stepping = false;
                                syncTo = null;
                                setTickDelay( preSyncTickDelay );
                            }
                            break;
                        case BACKWARD:
                            if( syncTo.getCurrentStep() == ( slaveSyncStep - 2 ) ) {
                                setDirection( FORWARD );
                                oneMoreTick = true;
                            }
                            break;
                        }
                    }
                } else { // if( clockActive )
                    // don't waste CPU cycles in pause mode. The resume() method will
                    // interrupt(), so we can continue later.
                    synchronized( this ) {
                        this.wait();
                    }
                } // if( clockActive )
            } catch( InterruptedException ie ) {
                // FIXME: work done? -- Do we need to check if this interrupt
                // was generated by ourselves or is it safe to assume so?
            } catch( IllegalMonitorStateException ime ) {
                ime.printStackTrace();
            } // try
        } // while

        /* The clock has been stopped. As this invalidates the thread, we have to make
         * sure our singleton clock cannot be used again.
         */
        simulationClock = null;
    }

} // SimulationClock
