/*
 * $Id: Cluster.java,v 1.1 2009/04/01 13:57:09 sirocic Exp $
 *
 * Created on 25.04.2006 14:57:34
 * by Dirk Frsterling <r@zorbla.de>
 */
package flexray.cluster;


import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.TreeMap;
import java.util.Vector;
import java.util.Map.Entry;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import flexray.communication.CommCycle;
import flexray.configuration.ChannelName;
import flexray.configuration.ClusterConfiguration;
import flexray.xml.XMLTags;

/**
 * @author Dirk Frsterling <r@zorbla.de>
 *
 */
public class Cluster implements XMLTags {

    private ClusterConfiguration clusterConfiguration = new ClusterConfiguration();

    private TreeMap<Integer,Node> nodes = new TreeMap<Integer,Node>();
    private TreeMap<Integer,Connector> connectors = new TreeMap<Integer,Connector>();

    /*
     * Initial values have to be one before the first macroTick
     */
    private long macroTick = -1;
    private int cycleCounter = 0;
    private int macroTickInCycle = -1;

    private int slotCounter[] = { -1, -1 };

//    private int slotCounterA = -1;
//    private int slotCounterB = -1;
    private int maxSlot = -1;


    private int lastStaticMT;
    private int lastDynamicMT;

    private CommCycle currentCommCycle = null;

    private TreeMap<Long, ClusterState> clusterStates = new TreeMap<Long, ClusterState>();

    private class ClusterState {

        private long cs_macroTick;
        private int  cs_cycleCounter;
        private int  cs_macroTickInCycle;
        private int  cs_slotCounter[] = new int[2];
        private int  cs_maxSlot;
        private int  cs_lastStaticMT;
        private int  cs_lastDynamicMT;
        private CommCycle cs_currentCommCycle;

        @SuppressWarnings("synthetic-access")
        public ClusterState(Cluster c) {
//            DebugUtil.println("storing: " + c);
            this.cs_macroTick        = c.getCurrentMacroTick();
            this.cs_cycleCounter     = c.getCycleCounter();
            this.cs_macroTickInCycle = c.getMacroTickInCycle();
            this.cs_slotCounter[ChannelName.A.ordinal()] = c.getSlotCounter(ChannelName.A);
            this.cs_slotCounter[ChannelName.B.ordinal()] = c.getSlotCounter(ChannelName.B);
            this.cs_maxSlot          = c.maxSlot;
            this.cs_lastDynamicMT    = c.lastDynamicMT;
            this.cs_lastStaticMT     = c.lastStaticMT;
            this.cs_currentCommCycle = c.getCurrentCommCycle().copy();
        }

        @SuppressWarnings("synthetic-access")
        public void restore(Cluster c) {
            c.macroTick        = this.cs_macroTick;
            c.cycleCounter     = this.cs_cycleCounter;
            c.macroTickInCycle = this.cs_macroTickInCycle;
            c.setSlotCounter(ChannelName.A, this.cs_slotCounter[ChannelName.A.ordinal()]);
            c.setSlotCounter(ChannelName.B, this.cs_slotCounter[ChannelName.B.ordinal()]);
            c.maxSlot          = this.cs_maxSlot;
            c.lastDynamicMT    = this.cs_lastDynamicMT;
            c.lastStaticMT     = this.cs_lastStaticMT;
            c.currentCommCycle = this.cs_currentCommCycle.copy();
//            DebugUtil.println("loading: " + c);
        }
    } // class ClusterState

    @Override
    public String toString() {
        StringBuffer text = new StringBuffer();
        text.append("tick " + macroTick);
        text.append(" cycle " + cycleCounter + "(" + macroTickInCycle + ")");
        text.append(" A." + getSlotCounter(ChannelName.A));
        text.append(" B." + getSlotCounter(ChannelName.B));
        text.append("\nA - ");
        int prevSlot = -2;
        int slot = 0;
        for( int i = 0 ; i < clusterConfiguration.gMacroPerCycle(); i++ ) {
            slot = currentCommCycle.getSlotNumberFromMacroTick(ChannelName.A, i);
            if( prevSlot == -2 ) {
                text.append(""+i);
            } else if( slot != prevSlot ) {
                text.append( "-" + (i-1) + ":" + prevSlot + ", " + i);
            }
            prevSlot = slot;
        }
        text.append("-" + clusterConfiguration.gMacroPerCycle() + ":" + slot +"\nB - ");

        prevSlot = -2;
        slot = 0;
        for( int i = 0 ; i < clusterConfiguration.gMacroPerCycle(); i++ ) {
            slot = currentCommCycle.getSlotNumberFromMacroTick(ChannelName.B, i);
            if( prevSlot == -2 ) {
                text.append(""+i);
            } else if( slot != prevSlot ) {
                text.append( "-" + (i-1) + ":" + prevSlot + ", " + i);
            }
            prevSlot = slot;
        }
        text.append("-" + clusterConfiguration.gMacroPerCycle() + ":" + slot );
        return text.toString();
    }
    /**
     * Construct a cluster
     *
     */
    public Cluster() {
        // empty
    }

    /*
     * ---------------- cluster configuration  ------------------------
     */

    /**
     * Return the node that is configured for the static slot slotID
     *
     * @param channel
     * @param slotID
     * @return the Node.
     */
    public Node getNodeForSlot(ChannelName channel, int slotID ) {
        if( slotID > ClusterConfiguration.cSlotIDMax ) {
            return null;
        }

        for( Node n : nodes.values() ) {
            if( n.isTransmittingInSlot(slotID) &&
                n.getNodeConfiguration().pChannels().contains(channel))
                return n;
        }
        return null;
    }
    
    /**
     * Return the node that is configured for the static slot slotID
     *
     * @param channel
     * @param slotID
     * @return the Node.
     */
    public ArrayList getNodesForSlot(ChannelName channel, int slotID ) {
    	ArrayList nodesForSlot=null;
    	
        if( slotID > ClusterConfiguration.cSlotIDMax ) {
            return null;
        }

        nodesForSlot = new ArrayList();
        for( Node n : nodes.values() ) {
            if( n.isTransmittingInSlot(slotID) &&
                n.getNodeConfiguration().pChannels().contains(channel))
            	nodesForSlot.add(n);
        }
        return nodesForSlot;
    }

    /**
     * Get the cluster configuration.
     *
     * Would be better to return a copy.
     * @return the cluster configuration object.
     */
    public ClusterConfiguration getClusterConfiguration() {
        return clusterConfiguration;
    }

    public void setClusterConfiguration(ClusterConfiguration config) {
        this.clusterConfiguration = config;
    }


    /**
     * Adding a Node to the cluster.
     *
     *  @param pNode the node to add
     */
    public void addNode(Node pNode) {
        nodes.put(Integer.valueOf(pNode.getID()),pNode);
    }

    public void removeNode(int nodeID) {
        nodes.remove(Integer.valueOf(nodeID));
    }

    public Collection<Node> getNodes() {
        return nodes.values();
    }

    public int getNumberOfNodes() {
        return nodes.size();
    }

    public Node getNode(int id) {
        return nodes.get(Integer.valueOf(id));
    }

//    public Node getNode(ChannelName channel, int id) {
//        return nodes.get(Integer.valueOf(id));
//    }

    //        connectors

    /**
     * Get a List of connectors for a channel
     * @param c
     * @return the list of connectors
     */
    public List<Connector> getConnectorsForChannel(ChannelName c) {
        Vector<Connector> connectorList = new Vector<Connector>();
        for( Entry<Integer, Connector> conn : connectors.entrySet() ) {
            if( conn.getValue().getChannel().contains(c)) {
                connectorList.add(conn.getValue());
            }
        }
        return connectorList;
    }

    /*
     * ------------------ cluster operation  ------------------------
     */

    /**
     * This receives the macrotick "event". This method should
     * block the caller until every child has done its macrotick.
     *
     */
    static long average = 0;
    static long avgcount = 0;
    public void receiveMacroTick() {
//        DebugUtil.println("Cluster received MT (" + macroTick + "/" + macroTickInCycle + ")");

        Long currentTick = Long.valueOf(macroTick);
        if( clusterStates.containsKey(currentTick)) {
            clusterStates.get(currentTick).restore(this);
            for( Node node : nodes.values() ) {
                node.receiveMacroTick();
            }
            return;
        }

        if(currentCommCycle == null) currentCommCycle = new CommCycle(clusterConfiguration.gMacroPerCycle());

        lastStaticMT = clusterConfiguration.gNumberOfStaticSlots()*clusterConfiguration.gdStaticSlot();
        lastDynamicMT = lastStaticMT+(clusterConfiguration.gNumberOfMinislots()*clusterConfiguration.gdMinislot());

        macroTick++;
        macroTickInCycle++;


        if( macroTickInCycle >= clusterConfiguration.gMacroPerCycle()) {
            cycleCounter++;
            macroTickInCycle = 0;
            setSlotCounter(ChannelName.AB, 0);
        }

        if( inStaticSegment() ) {
            // synchronous in static segment
            setSlotCounter(ChannelName.AB,  (macroTickInCycle / clusterConfiguration.gdStaticSlot()));
        }


        if( inDynamicSegment() ) {
            if( macroTickInCycle == clusterConfiguration.gdStaticSlot()*clusterConfiguration.gNumberOfStaticSlots()) {
                incSlotCounter(ChannelName.AB);
            }
            else {
	            if ((macroTickInCycle % clusterConfiguration.gdMinislot()) == 0) {
	            	if (getNodeForSlot(ChannelName.A, getSlotCounter(ChannelName.A)) == null)
	            		incSlotCounter(ChannelName.A);
	                if (getNodeForSlot(ChannelName.B, getSlotCounter(ChannelName.B)) == null)
	                	incSlotCounter(ChannelName.B);
	            }
            }
        }

        /*
         * Correction of Slot Counters (from nodes, basically during dynamic)
         */
        int cA = 0;
        int cB = 0;
        for( Node n : nodes.values() ) {
            cA += n.getAndResetSlotCounterCorr(ChannelName.A);
            cB += n.getAndResetSlotCounterCorr(ChannelName.B);
        }
        while( cA > 0 ) {
            incSlotCounter(ChannelName.A);
            cA--;
        }
        while( cB > 0 ) {
            incSlotCounter(ChannelName.B);
            cB--;
        }
        
        if( inNIT() ) {
            // FIXME: What's the slot counter in the NIT?
            setSlotCounter(ChannelName.AB, -1);
        }

        /*
         * Do the work in the nodes
         */
        for( Node node : nodes.values() ) {
            node.receiveMacroTick();
        }

        /*
         * store the data.
         */
        currentCommCycle.setSlotNumberForMacroTick(ChannelName.A, macroTickInCycle, getSlotCounter(ChannelName.A));
        currentCommCycle.setSlotNumberForMacroTick(ChannelName.B, macroTickInCycle, getSlotCounter(ChannelName.B));

        clusterStates.put(currentTick, new ClusterState(this));
    } // receiveMacroTick

    public CommCycle getCurrentCommCycle() {
        return currentCommCycle;
    }

    /**
     * Beware of side effect. This counter is only accurate if called
     * from within one tick. :(
     * @return the last dynamic macrotick counted
     */
    public int getLastDynamicMT() {
        return lastDynamicMT;
    }

    /**
     * Increment the slot counter of the given channel by 1
     * @param channel A, B or AB for both
     */
    protected void incSlotCounter(final ChannelName channel) {
        switch (channel) {
        case A:
        case B:
            slotCounter[channel.ordinal()]++;
            break;
        case AB:
            slotCounter[ChannelName.A.ordinal()]++;
            slotCounter[ChannelName.B.ordinal()]++;
            break;
        default:
            break;
        }
    }

    /**
     * set the slot counter(s) to the given value
     * @param channel A, B or AB for both
     * @param value the value to set the slot counter(s) to
     */
    private void setSlotCounter(final ChannelName channel, final int value) {
        switch (channel) {
        case A:
        case B:
            slotCounter[channel.ordinal()] = value;
            break;
        case AB:
            slotCounter = new int[]{value, value};
            break;
        default:
            break;
        }
    }

    /**
     * Get the current value of a slot counter
     * @param channel the channel for which to get the slot counter
     * @return the slot counter for channel <tt>channel</tt>
     */
    public int getSlotCounter(ChannelName channel) {
        switch(channel) {
        case A:
        case B:
            return slotCounter[channel.ordinal()];
        default:
            return 0;
        }
    }


    /**
     * Return the number of the current macrotick relative to the current
     * communication cycle
     * @return the current macrotick position within the current cycle.
     */
    protected int getMacroTickInCycle() {
        return macroTickInCycle;
    }

    /**
     * This receives the backward macrotick "event".
     */
    public void undoMacroTick() {
        if(macroTick <= 0) return;
        // exchanged throw for return because leviSimulator fails to catch exceptions
        // from execute* methode.
        // throw new IllegalStateException("cannot step back beyond origin");

        Long currentTick = Long.valueOf(macroTick-2);
        clusterStates.get(currentTick).restore(this);

        // now the nodes
        for( Node node : nodes.values() ) {
            node.undoMacroTick();
        }

    } // undoMacroTick

    public boolean inStaticSegment() {
        if( macroTickInCycle < lastStaticMT ) return true;
        return false;
    }

    public boolean inDynamicSegment() {
        if( (macroTickInCycle >= lastStaticMT) &&
                (macroTickInCycle < lastDynamicMT))
            return true;
        return false;
    }

    public boolean inNIT() {
        if( macroTickInCycle >= lastDynamicMT) return true;
        return false;
    }

    /**
     * Reset all runtime data
     *
     */
    public void reset() {
        macroTick = -1;
        cycleCounter = 0;
        macroTickInCycle = -1;
        slotCounter = new int[]{ -1, -1 };
        maxSlot = -1;
        lastStaticMT = 0;
        lastDynamicMT = 0;
        currentCommCycle = null;
        for(Node node : nodes.values() ) {
            node.reset();
        }
        clusterStates.clear();
    }

    public long getCurrentMacroTick() {
        return macroTick;
    }


    public int getCycleCounter() {
        return cycleCounter;
    }


    /*
     * ---------------------- storage ------------------------
     */

    public Element toXML(final Document xmlDoc) {
        Element element = xmlDoc.createElement(XT_CLUSTER);
        for( Node n : getNodes() ) {
            element.appendChild(n.toXML(xmlDoc));
        }
        element.appendChild(clusterConfiguration.toXML(xmlDoc));
        return element;
    }

    public boolean restoreFromXML(final Element element) {
        NodeList clusterTreeNodes = element.getChildNodes();
        // for all childnodes
        for(int index = 0 ; index < clusterTreeNodes.getLength(); index++) {
        	if (clusterTreeNodes.item(index) instanceof Element) {
	            Element nodeElement = (Element)clusterTreeNodes.item(index);
	            // One Cluster Node
	            if( nodeElement.getNodeName().equals(XT_CNODE)) {
	                Integer id = Integer.valueOf(nodeElement.getAttribute(XT_CNODE_ID));
	                Node node = new Node(this, id.intValue());
	                addNode(node);
	//                createNode(id.intValue());
	//                Node node = getNode(id.intValue());
	                node.restoreFromXML(nodeElement);
	            } // if CLUSTERNODE
	            if(nodeElement.getNodeName().equals(XT_CLUSTER_CONFIG)) {
	                clusterConfiguration.restoreFromXML(nodeElement);
	            }
        	}
        } // for all XML elements
        return true;
    }

}
