/*
 * $Id: Node.java,v 1.1 2009/04/01 13:57:09 sirocic Exp $
 *
 * Created on 25.04.2006 14:56:47
 * by Dirk Frsterling <r@zorbla.de>
 */
package flexray.cluster;


import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Observable;
import java.util.Observer;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.Map.Entry;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import flexray.communication.FlexRayFrame;
import flexray.configuration.ChannelName;
import flexray.configuration.ClusterConfiguration;
import flexray.configuration.NodeConfiguration;
import flexray.gui.Property;
import flexray.util.DebugUtil;
import flexray.xml.XMLTags;

/**
 * Abstract scenario of a FlexRay node.
 *
 * Node contains local Configuration as well as functionality of a FlexRay
 * node necessary to show how the MAC works.
 *
 * @author Dirk Frsterling <r@zorbla.de>
 *
 */
public class Node implements Observer,
                             XMLTags {

    private int nodeID;

    private NodeConfiguration nodeConfiguration = new NodeConfiguration();
    private Cluster cluster = null;
    private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);

    private String name = null;

    /**
     * the slots this node transmits in
     */
    private HashSet<Integer> transmitSlots = new HashSet<Integer>();


    private Random random = new Random();

    private boolean dynamicSend[] = { true, true };
    private boolean wantSendStatic[] = { true, true };
    private boolean dynamicPauseSending = false;

    private int sendLengthMin[] = { 0, 0 };
    private int sendLengthMax[] = { 0, 0}; // default values
    //private boolean sendRandom = true;


    // There is no clean way of creating an array of generic collections.
    private HashMap<Integer,FlexRayFrame>[] dynFrame = new HashMap[2];
    {
        dynFrame[ChannelName.A.ordinal()] = new HashMap<Integer,FlexRayFrame>();
        dynFrame[ChannelName.B.ordinal()] = new HashMap<Integer,FlexRayFrame>();
    }

    private int[] requestLength = {0,0};
    private boolean requestAllowed = true;

    private int[] slotCounterCorr = new int[2];
    private int[] messageID = new int[2];
    private int[] sendDynCycleMin = { 0, 0 }; // 0,0
    private int[] sendDynCycleMax = { 0, 0 }; // 0,0
    private int[] wantedDiff = {0,0};
    private int[] lastSendCycle = {-1,-1};

    private int[] sendStatCycleMin = { 0, 0 }; // 0,0
    private int[] sendStatCycleMax = { 0, 0 }; // 0,0
    private int[] statWantedDiff = { 0 , 0 };
    private int[] lastStatSendCycle = {-1,-1};

    private int[] dynMessageCount = {0, 0};
    private int[] dynRequestMin = { 0, 0 };
    private int[] dynRequestMax = { 0, 0 };

    private TreeMap<Long, NodeState> nodeStates = new TreeMap<Long, NodeState>();
    private Long currentTick;


    /**
     * node state
     */
    private class NodeState {
        private HashMap<Integer,FlexRayFrame>[] dynFrame = new HashMap[2];
        {
            dynFrame[ChannelName.A.ordinal()] = new HashMap<Integer,FlexRayFrame>();
            dynFrame[ChannelName.B.ordinal()] = new HashMap<Integer,FlexRayFrame>();
        }
        private int[] ns_slotCounterCorr = new int[2];
        private int[] ns_messageID = new int[2];
        private Long ns_currentTick;
        private boolean[] ns_dynamicSend  = new boolean[2];
        private boolean[] ns_wantSendStatic  = new boolean[2];
        private int[] ns_sendLengthMin = new int[2];
        private int[] ns_sendLengthMax = new int[2];
        private int[] ns_sendCycleMin = new int[2];
        private int[] ns_sendCycleMax = new int[2];
        private int[] ns_lastSendCycle = new int[2];
        private int[] ns_wantedDiff = new int[2];
        private int[] ns_sendStatCycleMin = new int[2];
        private int[] ns_sendStatCycleMax = new int[2];
        private int[] ns_lastSendStatCycle = new int[2];
        private int[] ns_statWantedDiff = new int[2];
        private int[] ns_dynMessageCount = new int[2];
        private int[] ns_dynRequestMin = new int[2];
        private int[] ns_dynRequestMax = new int[2];

        public NodeState(Node node) {
            this.ns_currentTick = node.currentTick;
            for( ChannelName channel : ChannelName.channels() ) {
                int channelOrd = channel.ordinal();
                for( Entry<Integer, FlexRayFrame> oneFrame : node.dynFrame[channelOrd].entrySet() ) {
//                    if( oneFrame.getValue() != null ) DebugUtil.println(this.currentTick + ": storing " + oneFrame.getValue());
                    this.dynFrame[channelOrd].put(oneFrame.getKey(),
                            oneFrame.getValue().copy());
                }
                this.ns_slotCounterCorr[channelOrd] = node.slotCounterCorr[channelOrd];
                this.ns_messageID[channelOrd] = node.messageID[channelOrd];
                this.ns_dynamicSend[channelOrd] = node.isDynamicSend(channel);
                this.ns_wantSendStatic[channelOrd] = node.wantSendStatic[channelOrd];
                this.ns_sendLengthMin[channelOrd] = node.getSendLengthMin(channel);
                this.ns_sendLengthMax[channelOrd] = node.getSendLengthMax(channel);
                this.ns_sendCycleMin[channelOrd] = node.getSendCycleMin(channel);
                this.ns_sendCycleMax[channelOrd] = node.getSendCycleMax(channel);
                this.ns_lastSendCycle[channelOrd] = node.lastSendCycle[channelOrd];
                this.ns_wantedDiff[channelOrd] = node.wantedDiff[channelOrd];
                this.ns_sendStatCycleMin[channelOrd] = node.getSendStatCycleMin(channel);
                this.ns_sendStatCycleMax[channelOrd] = node.getSendStatCycleMax(channel);
                this.ns_lastSendStatCycle[channelOrd] = node.lastStatSendCycle[channelOrd];
                this.ns_statWantedDiff[channelOrd] = node.statWantedDiff[channelOrd];
                this.ns_dynMessageCount[channelOrd] = node.getDynMessageCount(channel);
                this.ns_dynRequestMin[channelOrd] = node.getDynRequestMin(channel);
                this.ns_dynRequestMax[channelOrd] = node.getDynRequestMax(channel);
            }
        }


        public void restore( Node node ) {
            for( ChannelName channel : ChannelName.channels() ) {
                int channelOrd = channel.ordinal();
                node.messageID[channelOrd] = this.ns_messageID[channelOrd];
                node.slotCounterCorr[channelOrd] = this.ns_slotCounterCorr[channelOrd];
                node.dynFrame[channelOrd].clear();
                for( Entry<Integer,FlexRayFrame> oneFrame : this.dynFrame[channelOrd].entrySet()) {
//                    if( oneFrame.getValue() != null ) DebugUtil.println(this.currentTick + ": loading " + oneFrame.getValue());
                    node.dynFrame[channelOrd].put(oneFrame.getKey(),
                                this.dynFrame[channelOrd].get(oneFrame.getKey()).copy());
                }
                node.setDynamicSend(channel,ns_dynamicSend[channelOrd]);
                node.wantSendStatic[channelOrd] = ns_wantSendStatic[channelOrd];
                node.setSendLengthMin(channel,ns_sendLengthMin[channelOrd]);
                node.setSendLengthMax(channel,ns_sendLengthMax[channelOrd]);
                node.setSendDynCycleMin(channel,ns_sendCycleMin[channelOrd]);
                node.setSendDynCycleMax(channel,ns_sendCycleMax[channelOrd]);
                node.lastSendCycle[channelOrd]=ns_lastSendCycle[channelOrd];
                node.wantedDiff[channelOrd]=ns_wantedDiff[channelOrd];
                node.setSendStatCycleMin(channel,ns_sendStatCycleMin[channelOrd]);
                node.setSendStatCycleMax(channel,ns_sendStatCycleMax[channelOrd]);
                node.lastStatSendCycle[channelOrd]=ns_lastSendStatCycle[channelOrd];
                node.statWantedDiff[channelOrd]=ns_statWantedDiff[channelOrd];
                node.dynMessageCount[channelOrd]=ns_dynMessageCount[channelOrd];
                node.dynRequestMin[channelOrd]=ns_dynRequestMin[channelOrd];
                node.dynRequestMax[channelOrd]=ns_dynRequestMax[channelOrd];
                if( node.currentTick != ns_currentTick)
                    node.currentTick = ns_currentTick;
            }
        } // restore

    }

    /**
     * Create a new Node.
     * @param c cluster the node is being added to
     * @param id id nof the new node
     *
     */
    public Node( Cluster c, int id ) {
        this.cluster = c;
        nodeID = id;
    }

    /*
     *   ------------------- node configuration ----------------------
     */


    /**
     * return a reference to the cluster this node belongs to
     * @return the cluster this node belongs to
     */
    public Cluster getCluster() {
        return cluster;
    }
    /**
     * Get the node's id.
     * @return the integer id of this node
     */
    public int getID() {
        return nodeID;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }


    /**
     * check for a specific slot id if this node wants to transmit
     * @param slotID
     * @return true if this node transmits in the given slot
     */
    public boolean isTransmittingInSlot( int slotID ) {
        if( transmitSlots.contains(Integer.valueOf(slotID))) return true;
        return false;
    }

    /**
     * Add a slot id to the list of slots this node transmits in
     * @param slotID
     */
    public void addTransmitSlot(int slotID ) {
        Integer iSlotID = Integer.valueOf(slotID);
        if( !transmitSlots.contains(iSlotID)) {
            transmitSlots.add(iSlotID);
        }
    }

    /**
     * Remove one slot id from the list of ids of slots this node transmits in
     * @param slotID
     */
    public void removeTransmitSlot( int slotID ) {
        transmitSlots.remove(Integer.valueOf(slotID));
    }

    /**
     * Return a set of slot ids which this node transmits in
     * @return a set of slots in which this node transmits
     */
    public Set<Integer> getTransmitSlots() {
        return transmitSlots;
    }

    /**
     * set the channels this node is using
     * @param channels
     */
    public void setChannels(final ChannelName channels) {
        nodeConfiguration.pChannels(channels);
    }

    /**
     *
     * @param channel
     */
    public void addChannel(final ChannelName channel) {
        ChannelName currentChannels = nodeConfiguration.pChannels();

        // already AB set - more is impossible
        if(currentChannels.contains(ChannelName.AB)) return;

        // current = requested. Nothing to do
        if( currentChannels.equals(channel)) return;

        // AB is to be added or there is nothing set yet.
        if( channel.contains(ChannelName.AB) ||
            currentChannels.equals(ChannelName.NONE)) {
                nodeConfiguration.pChannels(channel);
        } else {
            // Now we have A and want to set B or vice versa.
            nodeConfiguration.pChannels(ChannelName.AB);
        }

        propertyChangeSupport.firePropertyChange(Property.NODE_CHANNELS,currentChannels,channel);
    }

    public void setDynamicSend(ChannelName channel, boolean dynamicSendRandom) {
        switch( channel ) {
        case A:
        case B:
            this.dynamicSend[channel.ordinal()] = dynamicSendRandom;
            break;
        case AB:
            this.dynamicSend[ChannelName.A.ordinal()] = dynamicSendRandom;
            this.dynamicSend[ChannelName.B.ordinal()] = dynamicSendRandom;
            break;
        default:
        }
    }

    public boolean isDynamicSend(ChannelName channel) {
        return dynamicSend[channel.ordinal()];
    }

    /*
     *   ------------------- node operation ----------------------
     */

    /**
     * "step forward"
     *
     */
    public void receiveMacroTick() {
        currentTick = Long.valueOf(cluster.getCurrentMacroTick());
        if( nodeStates.containsKey(currentTick)) {
            nodeStates.get(currentTick).restore(this);
        } else {
            if( cluster.getClusterConfiguration().gChannels().contains(ChannelName.A))
                receiveMacroTickForChannel(ChannelName.A);
            if( cluster.getClusterConfiguration().gChannels().contains(ChannelName.B))
                receiveMacroTickForChannel(ChannelName.B);

            nodeStates.put(currentTick, new NodeState(this));
        }
    } // receiveMacroTick

    /**
     * "work" to be done when a macrotick passed by.
     *
     * @param channel
     */
    private void receiveMacroTickForChannel(ChannelName channel) {
        int channelOrd = channel.ordinal();
        Integer slot = Integer.valueOf(cluster.getSlotCounter(channel));

        // if this is the first MacroTick, initialize.
        // In real FlexRay networks, this is done
        // during the NIT. I do it here, because
        // startup is not simulated and so there will
        // be no NIT before the first cycle.
        // Accordiing to the open definition in Chapter 9.3.2 I decided
        // that a node can have different message sending behaviour
        // for each channel.
        if( cluster.inDynamicSegment() ) {
            int length = 0;
            boolean emptyFrame;
            // when I am connected to channel
            if( nodeConfiguration.pChannels().contains(channel) && (isTransmittingInSlot(slot.intValue()))) {
                // if first cycle or frame is sent
                FlexRayFrame frame = dynFrame[channelOrd].get(slot);
                if( (frame == null) ||
                    ((frame.getTickInFrame() == 0) && (frame.getCycle() < cluster.getCycleCounter())) ||
                    (frame.isSent())
                   ) {
                    if( (sendLengthMax[channelOrd]-sendLengthMin[channelOrd]) > 0 ) {
                    	// Angabe der mglichen Nachrichten enthlt variablen Anteil z.B. zwischen 0 und 1 Slots
                    	// kann sein, dass nicht gesendet wird (zufallsentscheidung)
                        length = random.nextInt(sendLengthMax[channelOrd]-sendLengthMin[channelOrd]+1);
                    }
                    length+=sendLengthMin[channelOrd];

                    /*
                     * FlexRay Protocol Specification v2.1, Chapter 5.1.4.2
                     * The Node performs slot counting in the following way:
                     *  * The dynamic slot consists of one empty minislot if no communication takes place [...]
                     *  * The dynamic slot consists of one or multiple minislots if communication takes place [...]
                     *
                     */
                    if( !dynamicSend[channelOrd] ) {
                    	length=FlexRayFrame.EMPTY_FRAME_LENGTH;
                    	emptyFrame=true;
                    }
                    else
                    	emptyFrame=false;

                    // Macht der Knoten eine Pause?
                    if( sendDynCycleMax[channelOrd] > 0 ) {
                        int cycleDiff = cluster.getCycleCounter() - lastSendCycle[channelOrd] - 1;
                        if( cycleDiff != wantedDiff[channelOrd] ) {
                            length=FlexRayFrame.EMPTY_FRAME_LENGTH;
                            emptyFrame=true;
                        }
                    }

                    // dynamic segment message request
                    if( requestLength[channelOrd] == 0) {
                        dynFrame[channelOrd].put(slot,
                                new FlexRayFrame(cluster.getCycleCounter(),
                                        slot.intValue(),
                                        messageID[channelOrd],
                                        length, emptyFrame, length*cluster.getClusterConfiguration().gdMinislot()));
                    } else {
                        dynFrame[channelOrd].put(slot,
                                new FlexRayFrame(cluster.getCycleCounter(),
                                        slot.intValue(),
                                        messageID[channelOrd],
                                        requestLength[channelOrd],
                                        emptyFrame,
                                        length*cluster.getClusterConfiguration().gdMinislot()));
                        requestLength[channelOrd] = 0;
                        requestAllowed = true;
                    }
                    messageID[channelOrd]++;
                }
            }
        } // inDynamic

        // side effects in use. The following
        // "in*" Methods return the correct values
        // only for the current macrotick _after_ the
        // Cluster#receiveMacroTick method was entered!
        if( cluster.inStaticSegment() ) {
            // after this calculation, the wantSendStatic array contains the correct values
            if( nodeConfiguration.pChannels().contains(channel) ) {
                if( isTransmittingInSlot(slot.intValue())) {
                    if( sendStatCycleMax[channelOrd] > 0 ) {
                        // if first slotTick
                        if( cluster.getMacroTickInCycle() == (slot.intValue() * cluster.getClusterConfiguration().gdStaticSlot())) {
                            int cycle = cluster.getCycleCounter();
                            if( wantSendStatic[channelOrd] ) { // sent in the previous cycle ?
                                lastStatSendCycle[channelOrd] = cycle - 1;
                            }
                            // how long ago did I send?
                            int cycleDiff = cycle - lastStatSendCycle[channelOrd] - 1;
                            if( statWantedDiff[channelOrd] == cycleDiff ) {
                                int ra = random.nextInt(sendStatCycleMax[channelOrd]-sendStatCycleMin[channelOrd]+1);
                                statWantedDiff[channelOrd] = ra;
                                statWantedDiff[channelOrd] += sendStatCycleMin[channelOrd];
                                wantSendStatic[channelOrd] = true;
                            } else {
                                wantSendStatic[channelOrd] = false;
                            }
                        }
                    } // if max is set
                } // isTransmittingInSlot
            } // containsChannel

        } // in static segment


        if( cluster.inDynamicSegment() ) {
            if( nodeConfiguration.pChannels().contains(channel) ) {
                FlexRayFrame frame = dynFrame[channelOrd].get(slot);
                long frameEnd = 0;
                if( frame != null ) frameEnd = Math.max(1,frame.getLength()) * cluster.getClusterConfiguration().gdMinislot();

                if( (frame != null) && (isTransmittingInSlot(slot.intValue())) ) {
                    frame.incrementTickInFrame();
                }

                // there is a frame, but the node has not yet finished sending it.
                if( (frame != null) && (!frame.isSent()) ) {
                    // just reached the end ?
                    if( (frame.getTickInFrame() == frameEnd)
                      ) {
                        slotCounterCorr[channelOrd]++;
                        frame.setSent(true);
                        if (!frame.isEmptyFrame()) {
                            dynMessageCount[channelOrd]++;
                            if( sendDynCycleMax[channelOrd] > 0 ) {
                                lastSendCycle[channelOrd] = cluster.getCycleCounter();
                                int ra = random.nextInt(sendDynCycleMax[channelOrd]-sendDynCycleMin[channelOrd]+1);
                                wantedDiff[channelOrd] = ra;
                                wantedDiff[channelOrd]+=sendDynCycleMin[channelOrd];
                            }
                        }
                    }
                    // not finished -> do a fragment
                    if(!frame.isSent() && ((cluster.getMacroTickInCycle()+1) == cluster.getLastDynamicMT()) ) {
                        frame.incrementFragment();
                    }
                }
            } // channel
        }// if inDynamic

        /*
         * After Work Party
         */
        if( cluster.inNIT() ) {
            // Network Idle Time
        }
    } // receiveMacroTickForChannel


    /**
     * "step backward"
     *
     */
    public void undoMacroTick() {
        currentTick = Long.valueOf(cluster.getCurrentMacroTick());
        if( nodeStates.containsKey(currentTick)) {
            nodeStates.get(currentTick).restore(this);
        } else {
            DebugUtil.println("Error: Tick #" + currentTick + " not stored");
        }
    }

    /**
     *
     * @param channel the channel to send the message to
     * @param length the length of the message in MacroTicks
     */
    public void requestMessage(ChannelName channel) {
        if( !nodeConfiguration.pChannels().contains(channel) )
            return; // ignore silently
        requestAllowed = false;
        int channelOrd = channel.ordinal();
        int dynLength = 0;

        // determine length of request
        if( (dynRequestMax[channelOrd]-dynRequestMin[channelOrd]) > 0 )
            dynLength = random.nextInt(dynRequestMax[channelOrd]-dynRequestMin[channelOrd]+1);
        dynLength+=dynRequestMin[channelOrd];

        requestLength[channel.ordinal()] = dynLength;
    }

    /**
     * determine if dynamic requests via action button are allowed
     * @return true if action button is allowed to start a request
     */
    public boolean isRequestAllowed() {
        return requestAllowed;
    }
    
    /**
     * berprft, ob der Knoten im aktuelle Zyklus eine Sendepause macht oder nicht.
     * @return true wenn der Knoten eine Sendepause macht.
     */
    public boolean isDynamicPauseSending() {
        return dynamicPauseSending;
    }
    

    /**
     * Reset the overall state of the node. This is done when
     * stopping the simulation to re-start it.
     * reset() calls reset for each channel and deletes the
     * mementoes.
     *
     */
    public void reset() {
        reset(ChannelName.A);
        reset(ChannelName.B);
        nodeStates.clear();
        requestAllowed = true;
    }

    /**
     * reset for one channel
     * @param channel
     * @see Node#reset()
     */
    public void reset(ChannelName channel) {
        dynFrame[channel.ordinal()].clear();
        slotCounterCorr[channel.ordinal()] = 0;
        messageID[channel.ordinal()] = 0;
        wantedDiff[channel.ordinal()] = 0;
        statWantedDiff[channel.ordinal()] = 0;
        lastSendCycle[channel.ordinal()] = -1;
        wantSendStatic[channel.ordinal()] = true;
        lastStatSendCycle[channel.ordinal()] = -1;
        dynMessageCount[channel.ordinal()] = 0;
        requestLength[channel.ordinal()] = 0;
    }

    public int getAndResetSlotCounterCorr(ChannelName channel) {
        int r = slotCounterCorr[channel.ordinal()];
        slotCounterCorr[channel.ordinal()] = 0;
        return r;
    }

    /**
     * convenience method to get the "current" (depending on the
     * running cluster) frame.
     * @param channel
     * @return the current frame for the given channel
     * @see #getFrameForSlot(ChannelName, int)
     */
    public FlexRayFrame getCurrentFrame(ChannelName channel) {
        return getFrameForSlot(channel,cluster.getSlotCounter(channel));
    }

    /**
     * Get the frame this node sends a specific channel/slot
     * @param channel
     * @param slot
     * @return the frame for the given channel and slot
     */
    public FlexRayFrame getFrameForSlot(ChannelName channel, int slot) {
        if( dynFrame[channel.ordinal()] != null )
            return dynFrame[channel.ordinal()].get(Integer.valueOf(slot));
        return null;
    }

    /*
     * @param o
     * @param arg
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    public void update(Observable o, Object arg) {
        // nothing yet
    }

    public NodeConfiguration getNodeConfiguration() {
        return nodeConfiguration;
    }

    /*
     * -------------------- property change support -------------------------
     */

    /*
     * @see java.beans.PropertyChangeSupport#addPropertyChangeListener(PropertyChangeListener)
     */
    public void addPropertyChangeListener( PropertyChangeListener l ) {
        propertyChangeSupport.addPropertyChangeListener( l );
    }

    /*
     * @see java.beans.PropertyChangeSupport#removePropertyChangeListener(PropertyChangeListener)
     */
    public void removePropertyChangeListener( PropertyChangeListener l ) {
        propertyChangeSupport.removePropertyChangeListener( l );
    }

    /*
     * ---------------------- storage ------------------------
     */

    public void restoreFromXML(final Element element) {
        NodeList nodeList = element.getChildNodes();
        // cluster node data
        for(int i = 0 ; i < nodeList.getLength() ; i++) {
        	if (nodeList.item(i) instanceof Element) {
	            Element item = (Element)nodeList.item(i);
	            if( item.getNodeName().equals(XT_CNODE_TRANSMITSLOT)) {
	                addTransmitSlot(Integer.parseInt(item.getTextContent()));
	            }
	            if( item.getNodeName().equals(XT_CNODE_NAME)) {
	                name = item.getTextContent();
	            }
	            if( item.getNodeName().equals(XT_CNODE_CONFIG)) {
	                nodeConfiguration.restoreFromXML(item);
	                // configuration read, now check the cluster
	                ChannelName nodeChannels = nodeConfiguration.pChannels();
	                ClusterConfiguration clusterConfig = cluster.getClusterConfiguration();
	                ChannelName clusterChannels = clusterConfig.gChannels();
	                if( !clusterChannels.contains(nodeChannels) ) {
	                    clusterConfig.gChannels(clusterChannels.add(nodeChannels));
	                }
	            }
	            if(item.getNodeName().equals(XT_CNODE_DYNFRAME)) {
	                // old file without channel info
	                if( item.getAttribute(XT_CNODE_DYNFRAME_CHANNEL).equals("")) {
	                    setDynamicSend(ChannelName.A,Boolean.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_SEND)).booleanValue());
	                    setSendLengthMin(ChannelName.A,Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_RANGE_MIN)).intValue());
	                    setSendLengthMax(ChannelName.A,Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_RANGE_MAX)).intValue());
	                    setDynamicSend(ChannelName.B,Boolean.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_SEND)).booleanValue());
	                    setSendLengthMin(ChannelName.B,Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_RANGE_MIN)).intValue());
	                    setSendLengthMax(ChannelName.B,Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_RANGE_MAX)).intValue());
	                    if( !item.getAttribute(XT_CNODE_DYNFRAME_DIST_MIN).equals("")) {
	                        setSendDynCycleMin(ChannelName.A,Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_DIST_MIN)).intValue());
	                        setSendDynCycleMax(ChannelName.A,Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_DIST_MAX)).intValue());
	                        setSendDynCycleMin(ChannelName.B,Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_DIST_MIN)).intValue());
	                        setSendDynCycleMax(ChannelName.B,Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_DIST_MAX)).intValue());
	                    }
	                } else {
	                    ChannelName channel = ChannelName.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_CHANNEL));
	                    setDynamicSend(channel,Boolean.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_SEND)).booleanValue());
	                    int value = Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_RANGE_MIN)).intValue();
	                    setSendLengthMin(channel,value);
	                    value = Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_RANGE_MAX)).intValue();
	                    setSendLengthMax(channel,value);
	                    if( !item.getAttribute(XT_CNODE_DYNFRAME_DIST_MIN).equals("")) {
	                        setSendDynCycleMin(channel,Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_DIST_MIN)).intValue());
	                        setSendDynCycleMax(channel,Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_DIST_MAX)).intValue());
	                    }
	                    if( !item.getAttribute(XT_CNODE_DYNFRAME_REQUEST_MIN).equals("")) {
	                        value = Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_REQUEST_MIN)).intValue();
	                        setDynRequestMin(channel,value);
	                        value =Integer.valueOf(item.getAttribute(XT_CNODE_DYNFRAME_REQUEST_MAX)).intValue();
	                        setDynRequestMax(channel,value);
	                    }
	                }
	            } // if item
	            if(item.getNodeName().equals(XT_CNODE_STATIC_SEND)) {
	                ChannelName channel = ChannelName.valueOf(item.getAttribute(XT_CNODE_STATIC_CHANNEL));
	                setSendStatCycleMin(channel, Integer.valueOf(item.getAttribute(XT_CNODE_STATIC_RANGE_MIN)));
	                setSendStatCycleMax(channel, Integer.valueOf(item.getAttribute(XT_CNODE_STATIC_RANGE_MAX)));
	            }
        	}
        } //  for... nodeList
    } // restoreFromXML

    /**
     * Create an XML Element representing this node
     *
     * @param xmlDoc
     * @return the XML element
     */
    public Element toXML(final Document xmlDoc) {
        Element nodeElement = xmlDoc.createElement(XT_CNODE);
        nodeElement.setAttribute(XT_CNODE_ID, Integer.valueOf(nodeID).toString());
        if(name != null) {
            Element nameElement = xmlDoc.createElement(XT_CNODE_NAME);
            nameElement.setTextContent(name);
            nodeElement.appendChild(nameElement);
        }

        // Channel A
        Element dynFrameConfig = xmlDoc.createElement(XT_CNODE_DYNFRAME);
        // reverse order to make it look more logical
        int channelOrd = ChannelName.A.ordinal();
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_REQUEST_MAX,
                                    Integer.valueOf(dynRequestMax[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_REQUEST_MIN,
                                    Integer.valueOf(dynRequestMin[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_DIST_MAX,
                                    Integer.valueOf(sendDynCycleMax[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_DIST_MIN,
                                    Integer.valueOf(sendDynCycleMin[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_RANGE_MAX,
                                    Integer.valueOf(sendLengthMax[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_RANGE_MIN,
                                    Integer.valueOf(sendLengthMin[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_SEND,
                                    Boolean.valueOf(dynamicSend[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_CHANNEL,
                                    ChannelName.A.toString());

        nodeElement.appendChild(dynFrameConfig);


        // Channel B
        dynFrameConfig = xmlDoc.createElement(XT_CNODE_DYNFRAME);
        // reverse order to make it look more logical
        channelOrd = ChannelName.B.ordinal();
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_REQUEST_MAX,
                                    Integer.valueOf(dynRequestMax[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_REQUEST_MIN,
                                    Integer.valueOf(dynRequestMin[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_DIST_MAX,
                                    Integer.valueOf(sendDynCycleMax[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_DIST_MIN,
                                    Integer.valueOf(sendDynCycleMin[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_RANGE_MAX,
                                    Integer.valueOf(sendLengthMax[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_RANGE_MIN,
                                    Integer.valueOf(sendLengthMin[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_SEND,
                                    Boolean.valueOf(dynamicSend[channelOrd]).toString());
        dynFrameConfig.setAttribute(XT_CNODE_DYNFRAME_CHANNEL,
                                    ChannelName.B.toString());

        nodeElement.appendChild(dynFrameConfig);

        Element statConfig = xmlDoc.createElement(XT_CNODE_STATIC_SEND);
        channelOrd = ChannelName.A.ordinal();
        statConfig.setAttribute(XT_CNODE_STATIC_RANGE_MAX, Integer.valueOf(sendStatCycleMax[channelOrd]).toString());
        statConfig.setAttribute(XT_CNODE_STATIC_RANGE_MIN, Integer.valueOf(sendStatCycleMin[channelOrd]).toString());
        statConfig.setAttribute(XT_CNODE_STATIC_CHANNEL, ChannelName.A.toString());
        nodeElement.appendChild(statConfig);

        statConfig = xmlDoc.createElement(XT_CNODE_STATIC_SEND);
        channelOrd = ChannelName.B.ordinal();
        statConfig.setAttribute(XT_CNODE_STATIC_RANGE_MAX, Integer.valueOf(sendStatCycleMax[channelOrd]).toString());
        statConfig.setAttribute(XT_CNODE_STATIC_RANGE_MIN, Integer.valueOf(sendStatCycleMin[channelOrd]).toString());
        statConfig.setAttribute(XT_CNODE_STATIC_CHANNEL, ChannelName.B.toString());
        nodeElement.appendChild(statConfig);

        for( Integer slot :  transmitSlots ) {
             Element transmitSlot = xmlDoc.createElement(XT_CNODE_TRANSMITSLOT);
             transmitSlot.setTextContent(slot.toString());
             nodeElement.appendChild(transmitSlot);
        }

        nodeElement.appendChild(nodeConfiguration.toXML(xmlDoc));

        return nodeElement;
    } // toXML

    public int getSendLengthMin(ChannelName channel) {
        return sendLengthMin[channel.ordinal()];
    }

    public void setSendLengthMin(ChannelName channel, int sendFrom) {
        this.sendLengthMin[channel.ordinal()] = sendFrom;
    }

    public int getSendLengthMax(ChannelName channel) {
        return sendLengthMax[channel.ordinal()];
    }

    public void setSendLengthMax(ChannelName channel, int sendTo) {
        this.sendLengthMax[channel.ordinal()] = sendTo;
    }

    public int getSendCycleMin(ChannelName channel) {
        return sendDynCycleMin[channel.ordinal()];
    }

    public void setSendDynCycleMin(ChannelName channel, int sendTo) {
        this.sendDynCycleMin[channel.ordinal()] = sendTo;
    }

    public int getSendStatCycleMin(ChannelName channel) {
        return sendStatCycleMin[channel.ordinal()];
    }

    public void setSendStatCycleMin(ChannelName channel, int sendTo) {
        this.sendStatCycleMin[channel.ordinal()] = sendTo;
    }

    public int getSendCycleMax(ChannelName channel) {
        return sendDynCycleMax[channel.ordinal()];
    }

    public void setSendDynCycleMax(ChannelName channel, int sendTo) {
        this.sendDynCycleMax[channel.ordinal()] = sendTo;
    }


    public int getSendStatCycleMax(ChannelName channel) {
        return sendStatCycleMax[channel.ordinal()];
    }

    public void setSendStatCycleMax(ChannelName channel, int sendTo) {
        this.sendStatCycleMax[channel.ordinal()] = sendTo;
    }

    public int getDynMessageCount(ChannelName channel) {
        return dynMessageCount[channel.ordinal()];
    }

    public int getDynRequestMin(ChannelName channel) {
        return dynRequestMin[channel.ordinal()];
    }

    public void setDynRequestMin(ChannelName channel, int length) {
        this.dynRequestMin[channel.ordinal()] = length;
    }

    public int getDynRequestMax(ChannelName channel) {
        return dynRequestMax[channel.ordinal()];
    }

    public void setDynRequestMax(ChannelName channel, int length) {
        this.dynRequestMax[channel.ordinal()] = length;
    }

    public boolean isSendingStatic( ChannelName channel ) {
        return wantSendStatic[channel.ordinal()];
    }

}
