/*
 * $Id: ClusterConfiguration.java,v 1.1 2009/04/01 13:57:10 sirocic Exp $
 *
 */
package flexray.configuration;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import flexray.xml.XMLTags;




/**
 * Class for FlexRay Cluster configuration data. This contains at most the data
 * that is relevant to run a simulation that suffices to show how the media access
 * control works in FlexRay.
 *
 * Intentionally violating Java naming conventions to make the whole thing more readable
 * in respect to the FlexRay Protocol Specification.
 *
 * @author Dirk Frsterling <r@zorbla.de>
 *
 */
public class ClusterConfiguration implements Cloneable,
                                             XMLTags {


    // Some protocol constants from FlexRay Protocol Specification, Appendix A
    // keeping the Names as in the specs, violating java naming conventions
    // (of writing constants in all uppercase)
    /** Maximum number of Controllers in a cluster */
    public static final int  cControllerMax       =   64;
    /** Maximum cycle counter value in a cluster */
    public static final int  cCycleCountMax       =   63;
    public static final int  cPayloadLengthMax    =  127; // two-byte-words
    public static final int  cMicroPerMacroNomMin =   40; // at least 40 T per MT
    /** Highest slot ID number */
    public static final int  cSlotIDMax           = 2047;
    /** Highest static slot ID number */
    public static final int  cStaticSlotIDMax     = 1023;

    // cluster parameters
    /** Duration of a Minislot (measured in Macroticks) */
    private int             gdMinislot;
    /** Minimum duration of a minislot */
    public static final int gdMinislot_min                  =  2;
    /** Maximum duration of a minislot */
    public static final int gdMinislot_max                  = 63;

    private int             gdMiniSlotActionPointOffset;
    public static final int gdMiniSlotActionPointOffset_min =  1;
    public static final int gdMiniSlotActionPointOffset_max = 31;

    /**
     * Duration of the Network Idle Time
     */
    private int             gdNIT;
    /** Minumum duration of the Network Idle Time */
    public static final int gdNIT_min                       =  1;
    /** Maximum duration of the Network Idle Time */
    public static final int gdNIT_max                       = 805;

    /**
     * Duration of a static slot (measured in Macroticks)
     *
     *  gdStaticSlot depends on the bit rate. See appendix B.4.9
     */
    private int             gdStaticSlot;
    /** Minimum duration of a static slot */
    public static final int gdStaticSlot_min                =   4; // @ 10MBit
    /** Maximum duration of a static slot */
    public static final int gdStaticSlot_max                = 661; // @  5MBit

    private int             gdSymbolWindow;
    public static final int gdSymbolWindow_min              =   0;
    public static final int gdSymbolWindow_max              = 142;

    /**
     * Number of Macroticks in a communication cycle
     * Appendix B.4.11
     */
    private int             gMacroPerCycle;
    public static final int gMacroPerCycle_min              =    10;
    public static final int gMacroPerCycle_max              = 16000;

    /**
     * Number of Minislots in the dynamic segment
     * Appendix B.4.14
     */
    private int             gNumberOfMinislots;
    /** Minimum number of Minislots in dynamic segment */
    public static final int gNumberOfMinislots_min          =    0;
    /** Maximum number of Minislots in dynamic segment */
    public static final int gNumberOfMinislots_max          = 7986;

    /** Number of static slots in the static segment */
    private int             gNumberOfStaticSlots;
    /** Minimum number of static slots */
    public static final int gNumberOfStaticSlots_min        = 2;
    /** Maximum number of static slots */
    public static final int gNumberOfStaticSlots_max        = cStaticSlotIDMax;

    private int             gPayloadLengthStatic;
    public static final int gPayloadLengthStatic_min        = 0;
    public static final int gPayloadLengthStatic_max        = cPayloadLengthMax;

    private ChannelName     gChannels                       = ChannelName.A;


    /**
     * Create a basic cluster configuration with some default Values.
     *
     */
    public ClusterConfiguration() {

        gdMinislot = gdMinislot_min;

        gdMiniSlotActionPointOffset = gdMiniSlotActionPointOffset_min;

        gdStaticSlot = gdStaticSlot_min;
        gdSymbolWindow = gdSymbolWindow_min;
        gNumberOfMinislots = gNumberOfMinislots_min;
        gNumberOfStaticSlots = gNumberOfStaticSlots_min;
        gPayloadLengthStatic = gPayloadLengthStatic_min;
        gdNIT = gdNIT_min;

        calculateMacros();
    }

    /**
     * (re)calculate the number of MacroTicks of one communication cycle
     *
     */
    private void calculateMacros() {
        gMacroPerCycle = gdMinislot * gNumberOfMinislots +
        gdStaticSlot * gNumberOfStaticSlots + gdNIT;
    }

    /**
     * @return the duration of a minislot (in macroticks)
     */
    public int gdMinislot() {
        return gdMinislot;
    }

    /**
     * Set the duration of a minislot in macroticks
     *
     * @param macroTicks
     */
    public void gdMinislot(int macroTicks) {
        if ((macroTicks >= gdMinislot_min) && (macroTicks <= gdMinislot_max)) {
            this.gdMinislot = macroTicks;
        } else {
            throw new IllegalArgumentException( "value out of allowed range" );
        }
        calculateMacros();
    }

    /**
     * @return the number of macroticks the minislot action point is offset from the beginning of a minislot.
     */
    public int gdMiniSlotActionPointOffset() {
        return gdMiniSlotActionPointOffset;
    }

    /**
     * Set the number of macroticks the minislot action point is offset from the beginning of a minislot.
     *
     * @param macroTicks
     */
    public void gdMiniSlotActionPointOffset(int macroTicks) {
        if ((macroTicks >= gdMiniSlotActionPointOffset_min) && (macroTicks <= gdMiniSlotActionPointOffset_max)) {
            this.gdMiniSlotActionPointOffset = macroTicks;
        } else {
            throw new IllegalArgumentException( "value out of allowed range" );
        }
    }

    /**
     * @return the duration of the static slot (in macroticks)
     */
    public int gdStaticSlot() {
        return gdStaticSlot;
    }

    /**
     * Set the duration of the static slot (in macroticks)
     *
     * @param macroTicks
     */
    public void gdStaticSlot(int macroTicks) {
        if ((macroTicks >= gdStaticSlot_min) && (macroTicks <= gdStaticSlot_max)) {
            this.gdStaticSlot = macroTicks;
        } else {
            throw new IllegalArgumentException( "value out of allowed range" );
        }
        calculateMacros();
    }

    /**
     * @return the duration of the symbol window (in macroticks)
     */
    public int gdSymbolWindow() {
        return gdSymbolWindow;
    }

    /**
     * Set the duration of the symbol window (in macroticks)
     *
     * @param macroTicks
     */
    public void gdSymbolWindow(int macroTicks) {
        if ((macroTicks >= gdSymbolWindow_min) && (macroTicks <= gdSymbolWindow_max)) {
            this.gdSymbolWindow = macroTicks;
        } else {
            throw new IllegalArgumentException( "value out of allowed range" );
        }
        calculateMacros();
    }

    /**
     * @return the duration of the network idle time (in macroticks)
     */
    public int gdNIT() {
        return gdNIT;
    }

    /**
     * Set the duration of the network idle time (in macroticks)
     *
     * @param macroTicks
     */
    public void gdNIT(int macroTicks) {
        if ((macroTicks >= gdNIT_min) && (macroTicks <= gdNIT_max)) {
            this.gdNIT = macroTicks;
        } else {
            throw new IllegalArgumentException( "value out of allowed range" );
        }
        calculateMacros();
    }

    /**
     * @return the number of maroticks in a communication cycle
     */
    public int gMacroPerCycle() {
        return gMacroPerCycle;
    }

    /**
     * Set the number of macroticks in a communication cycle
     *
     * @param macroTicks
     */
//    public void gMacroPerCycle(int macroTicks) {
//        throw new IllegalArgumentException();
//        if ((macroTicks > gMacroPerCycle_min) && (macroTicks < gMacroPerCycle_max)) {
//            this.gMacroPerCycle = macroTicks;
//        } else {
//            throw new IllegalArgumentException( "value out of allowed range" );
//        }
//    }

    /**
     * @return the number of minislots in the dynamic segment
     */
    public int gNumberOfMinislots() {
        return gNumberOfMinislots;
    }

    /**
     * Set the number of minislots in the dynamic segment
     *
     * @param numberOfMiniSlots
     */
    public void gNumberOfMinislots(int numberOfMiniSlots) {
        if ((numberOfMiniSlots >= gNumberOfMinislots_min) && (numberOfMiniSlots <= gNumberOfMinislots_max)) {
            int oldMiniSlotMacroTicks = this.gNumberOfMinislots * this.gdMinislot;
            int newMiniSlotMacroTicks = numberOfMiniSlots * this.gdMinislot;
            this.gNumberOfMinislots = numberOfMiniSlots;
            this.gMacroPerCycle += newMiniSlotMacroTicks - oldMiniSlotMacroTicks;
        } else {
            throw new IllegalArgumentException( "value out of allowed range" );
        }
        calculateMacros();
    }

    /**
     * @return the number of static slots in the static segment
     */
    public int gNumberOfStaticSlots() {
        return gNumberOfStaticSlots;
    }

    /**
     * Set the number of static slots in the static segment
     * The number of Macroticks per Cycle gets automatically updated
     *
     * @param numberOfStaticSlots
     */
    public void gNumberOfStaticSlots(int numberOfStaticSlots) {
        if ((numberOfStaticSlots >= gNumberOfStaticSlots_min) && (numberOfStaticSlots <= gNumberOfStaticSlots_max)) {
            int oldMiniSlotMacroTicks = this.gNumberOfStaticSlots * this.gdStaticSlot;
            int newMiniSlotMacroTicks = numberOfStaticSlots * this.gdStaticSlot;
            this.gNumberOfStaticSlots = numberOfStaticSlots;
            this.gMacroPerCycle += newMiniSlotMacroTicks - oldMiniSlotMacroTicks;
        } else {
            throw new IllegalArgumentException( "value out of allowed range" );
        }
        calculateMacros();
    }

    /**
     * @return the payload length of a static frame (in 2 byte words)
     */
    public int gPayloadLengthStatic() {
            return gPayloadLengthStatic;
    }

    /**
     * Set the payload length of a static frame (in 2 byte words)
     *
     * @param twoByteWords
     */
    public void gPayloadLengthStatic( int twoByteWords ) {
        if( (twoByteWords >= gPayloadLengthStatic_min) && (twoByteWords <= gPayloadLengthStatic_max) ) {
            this.gPayloadLengthStatic = twoByteWords;
        } else {
            throw new IllegalArgumentException( "value out of allowed range" );
        }
    }


    /**
     * @return the channels used by the cluster
     */
    public ChannelName gChannels() {
        return gChannels;
    }

    /**
     * Set the channels used by this cluster
     *
     * @param channels
     */
    public void gChannels( ChannelName channels ) {
        gChannels = channels;
    }

    /**
     * Clone the cluster configuration
     * @return a copy of this configuration
     * @throws CloneNotSupportedException
     * @see java.lang.Object#clone()
     */
    @Override
    public ClusterConfiguration clone() throws CloneNotSupportedException {
        // TODO Auto-generated method stub
        return (ClusterConfiguration)super.clone();
    }



    public void restoreFromXML(final Element element) {
        NodeList nodeList = element.getChildNodes();
        for( int i = 0 ; i < nodeList.getLength() ; i++ ) {
        	if (nodeList.item(i) instanceof Element) {
	            Element item = (Element)nodeList.item(i);
	            if( item.getNodeName().equals(XT_CLUSTER_CONFIG_GDSTATICSLOT)) {
	                gdStaticSlot(Integer.valueOf(item.getTextContent()).intValue());
	            } else if( item.getNodeName().equals(XT_CLUSTER_CONFIG_GDMINISLOT)) {
	                    gdMinislot(Integer.valueOf(item.getTextContent()).intValue());
	            } else if( item.getNodeName().equals(XT_CLUSTER_CONFIG_GDNIT)) {
	                        gdNIT(Integer.valueOf(item.getTextContent()).intValue());
	            } else if( item.getNodeName().equals(XT_CLUSTER_CONFIG_GNUMBEROFSTATICSLOTS)) {
	                    gNumberOfStaticSlots(Integer.valueOf(item.getTextContent()).intValue());
	            } else if( item.getNodeName().equals(XT_CLUSTER_CONFIG_GNUMBEROFMINISLOTS)) {
	                gNumberOfMinislots(Integer.valueOf(item.getTextContent()).intValue());
	            }
        	}
        }
    }

    /**
     *
     * @param xmlDoc
     * @return the XML Element representing this cluster configuration
     */
    public Element toXML(Document xmlDoc) {
        Element config =  xmlDoc.createElement(XT_CLUSTER_CONFIG);

        Element element = xmlDoc.createElement(XT_CLUSTER_CONFIG_GDSTATICSLOT);
        element.setTextContent(Integer.valueOf(gdStaticSlot).toString());
        config.appendChild(element);

        element = xmlDoc.createElement(XT_CLUSTER_CONFIG_GDMINISLOT);
        element.setTextContent(Integer.valueOf(gdMinislot).toString());
        config.appendChild(element);

        element = xmlDoc.createElement(XT_CLUSTER_CONFIG_GDNIT);
        element.setTextContent(Integer.valueOf(gdNIT).toString());
        config.appendChild(element);

        element = xmlDoc.createElement(XT_CLUSTER_CONFIG_GNUMBEROFSTATICSLOTS);
        element.setTextContent(Integer.valueOf(gNumberOfStaticSlots).toString());
        config.appendChild(element);

        element = xmlDoc.createElement(XT_CLUSTER_CONFIG_GNUMBEROFMINISLOTS);
        element.setTextContent(Integer.valueOf(gNumberOfMinislots).toString());
        config.appendChild(element);

        return config;
    }
 }
