/*
 * $Id: NodeConfiguration.java,v 1.1 2009/04/01 13:57:10 sirocic Exp $
 *
 */
package flexray.configuration;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import flexray.xml.XMLTags;


/**
 * Storage for FlexRay node configuration parameters.
 *
 * This contains the data that is relevant to run a simulation that
 * suffices to show how the media access control works in FlexRay.
 *
 * Intentionally violating Java naming conventions to make the whole thing more readable
 * in respect to the FlexRay Protocol Specification.
 *
 * NOTE: In this context "this Node" means the node which uses this configuration.
 *
 * @author Dirk Frsterling <r@zorbla.de>
 *
 */
public class NodeConfiguration implements Cloneable,
                                          XMLTags {

    public NodeConfiguration() {
        // nothing to construct yet
    }

    private ChannelName         pChannels                = ChannelName.NONE;

    // pMicroPerCycle is a Node configuration parameter because the length of a
    // microTick usually varies between nodes.
    private int                 pMicroPerCycle;          // calculates as sum(macroperCycle)*gMicroPerMacroTick
    private static final int    pMicroPerCycle_min       = 640;
    private static final int    pMicroPerCycle_max       = 640000;

    private int                 pPayloadLengthDynMax;
    private static final int    pPayloadLengthDynMax_min = 0;
    private static final int    pPayloadLengthDynMax_max =  ClusterConfiguration.cPayloadLengthMax;


    /**
     * @return the channels this node is connected to
     */
    public ChannelName pChannels() {
        return this.pChannels;
    }

    /**
     * Set this node's channel connectivity
     *
     * @param channels
     */
    public void pChannels( ChannelName channels ) {
        this.pChannels = channels;
    }

    /**
     * Get the number of microticks per communication cycle.
     * @return the value of pMicroPerCycle
     */
    public int pMicroPerCycle() {
        return pMicroPerCycle;
    }

    /**
     * Set the number of microtics per communication cycle.
     * @param microPerCycle
     */
    public void pMicroPerCycle( int microPerCycle ) {
        if( (microPerCycle > pMicroPerCycle_min) && (microPerCycle < pMicroPerCycle_max) ) {
            this.pMicroPerCycle = microPerCycle;
        } else {
            throw new IllegalArgumentException( "value out of allowed range" );
        }
    }

    /**
     * Get the maximum payload length (in 2 byte words) of one frame for the dynamic segment
     *
     * @return the maximum payload length for the dynamic segment
     */
    public int pPayloadLengthDynMax() {
        return pPayloadLengthDynMax;
    }

    /**
     * Set the maximum paload length (in 2 byte words) of one frame for the dynamic segment
     * @param payloadLengthDynMax
     */
    public void pPayloadLengthDynMax( int payloadLengthDynMax ) {
        if( (payloadLengthDynMax > pPayloadLengthDynMax_min) && (payloadLengthDynMax < pPayloadLengthDynMax_max) ) {
            this.pPayloadLengthDynMax = payloadLengthDynMax;
        } else {
            throw new IllegalArgumentException( "value out of allowed range" );
        }
    }

    /**
     * Clone this object. This is done mainly for the configuration dialogs, because
     * I did not want to manipulate the configuration object by call-by-reference
     * @return a copy of this NodeConfiguration
     * @throws CloneNotSupportedException
     * @see java.lang.Object#clone()
     */
    @Override
    public NodeConfiguration clone() throws CloneNotSupportedException {
        return (NodeConfiguration)super.clone();
    }


    /**
     * Restore the configuration from an XML element tree.
     * @param element
     */
    public void restoreFromXML(final Element element) {
        NodeList nodeList = element.getChildNodes();
        for( int i = 0 ; i < nodeList.getLength() ; i++ ) {
        	if (nodeList.item(i) instanceof Element) {
	            Element item = (Element)nodeList.item(i);
	            if( item.getNodeName().equals(XT_CNODE_CONFIG_PCHANNELS)) {
	                String text = item.getTextContent();
	                if( text.equals("--") ) {
	                    pChannels(ChannelName.NONE);
	                } else {
	                    pChannels(ChannelName.valueOf(text));
	                }
	            } 
        	}
        }
    }

    /**
     * Encode the configuration into a tree of XML nodes
     * @param xmlDoc
     * @return the root element of this Class' XML representation
     */
    public Element toXML(Document xmlDoc) {
        Element config =  xmlDoc.createElement(XT_CNODE_CONFIG);

        Element element = xmlDoc.createElement(XT_CNODE_CONFIG_PCHANNELS);
        element.setTextContent(pChannels.toString());
        config.appendChild(element);

        element = xmlDoc.createElement(XT_CNODE_CONFIG_PMICROPERCYCLE);
        element.setTextContent(Integer.valueOf(pMicroPerCycle).toString());
        config.appendChild(element);

        element = xmlDoc.createElement(XT_CNODE_CONFIG_PPAYLOADLENGTHDYNMAX);
        element.setTextContent(Integer.valueOf(pMicroPerCycle).toString());
        config.appendChild(element);

        return config;
    }
}
