package flexray.gui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import learningVisualizations.LeviConfiguration;
import learningVisualizations.LeviResourceManager;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import flexray.cluster.Node;
import flexray.configuration.ChannelName;
import flexray.scenario.FlexRayScenario;
import flexray.util.DebugUtil;
import flexray.xml.XMLTags;

/**
 * Component to represent an active Star Coupler
 *
 * This is also abused as a holding point for bus topology.
 *
 * @author Dirk Frsterling <r@zorbla.de>
 */
@SuppressWarnings("serial")
public class ActiveStarCouplerComponent extends AlphaBlendingComponent
                                     implements XMLTags,
                                                PropertyChangeListener,
                                                MouseMotionListener,
                                                ChangeListener {
	
    private Rectangle sizeRect = new Rectangle(25,25);
    private int width = sizeRect.width;
    private int height = sizeRect.height;
    private ChannelName channel;
    private final FlexRayScenario scenario;

    private Color moduleForeground = Color.WHITE;
    private Color moduleBackground = Color.BLACK;
    private Color brighterBackground = moduleBackground.brighter().brighter();
    private Color highlightColor = Color.RED;

    /**
     * construct an "active star"
     * @param channel
     * @param scenario
     */
    public ActiveStarCouplerComponent(final ChannelName channel, final FlexRayScenario scenario) {
        super();
        this.scenario = scenario;
       
        if (channel.contains(ChannelName.AB))
            throw new IllegalArgumentException("only one channel allowed");
        this.channel = channel;
        setColorsFromConfig();
        addMouseMotionListener(this);
    }

    /**
     * Paint-Method for the little cirles that represent the active star couplers
     * or the bus position.
     *
     * @param g
     * @see flexray.gui.AlphaBlendingComponent#paintComponent(java.awt.Graphics)
     */
    @Override
    public void paintComponent( Graphics g ) {
        super.paintComponent( g );

        Graphics2D g2d = (Graphics2D)g;

        g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);

        g2d.setColor( brighterBackground );
        g2d.fillOval( 3, 3, width-7, height-7 );
        Stroke s = g2d.getStroke();
        Point mousePosition = getMousePosition();
        if( ((mousePosition != null) && (sizeRect != null)) && sizeRect.contains(mousePosition)) {
            g2d.setColor( highlightColor );
            g2d.setStroke( new BasicStroke(2));
        } else {
            g2d.setColor( moduleForeground );
            g2d.setStroke( new BasicStroke(4));
        }
        g2d.drawOval( 3, 3, width-7, height-7 );
        g2d.setStroke(s);
        g2d.setColor(moduleForeground);
        g2d.drawString(channel.toString(),9,17);
    } // paintmodule

    @Override
    public Dimension getPreferredSize() {
        return new Dimension( width, height );
    }

    /**
     * Return the channel of this ActiveStarCouplerComponent
     * @return ChannelName.A or ChannelName.B
     */
    public ChannelName getChannel() {
        return channel;
    }

    @Override
    public void setLocation(int x, int y) {
        int oldx = this.getX();
        int oldy = this.getY();
        if( (oldx != x) || (oldy != y)) {
            super.setLocation(x,y);
            firePropertyChange(Property.LOCATION, new Point(oldx,oldy), new Point(x,y));
        }
    }

    /**
     * Create an XML representation of this NodeComponent's parameters
     * @param xmlDoc
     * @return the XML Element representing this coupler's parameters
     */
    public Element toXML(final Document xmlDoc) {
        Element myXML = xmlDoc.createElement(XT_STARCOUPLERVIEW);
        myXML.setAttribute(XT_STARCOUPLERVIEW_CHANNEL, channel.toString());
        Element screenpos = xmlDoc.createElement(XT_SCREENPOS);
        screenpos.setAttribute(XT_SCREENPOS_X, Integer.valueOf(getX()).toString());
        screenpos.setAttribute(XT_SCREENPOS_Y, Integer.valueOf(getY()).toString());
        screenpos.setAttribute(XT_SCREENPOS_Z,
                Integer.valueOf(getParent().getComponentZOrder(this)).toString());
        myXML.appendChild(screenpos);
        return myXML;
    }


    /**
     * Restore this node module's parameters from an XML element
     * @param element
     * @return the stored z-order number of the coupler module
     */
    public int restoreFromXML(final Element element) {
        NodeList nodeList = element.getChildNodes();
        int z = 0;
        for( int i = 0 ; i < nodeList.getLength() ; i++ ) {
        	if (nodeList.item(i) instanceof Element) {
	            Element item = (Element)nodeList.item(i);
	            if( item.getNodeName().equals(XT_SCREENPOS)) {
	                int x = Integer.parseInt(item.getAttribute(XT_SCREENPOS_X));
	                int y = Integer.parseInt(item.getAttribute(XT_SCREENPOS_Y));
	                z = Integer.parseInt(item.getAttribute(XT_SCREENPOS_Z));
	                setLocation(x, y);
	            } // SCREENPOS
        	}
        }
        return z;
    } // restoreFromXML


    /*
     * Set the colors of this module from the configuration values
     */
    private void setColorsFromConfig() {
    	Object o; 
    	
        LeviConfiguration config = scenario.getLearningModule().getConfiguration();
        if (config.getSettingByKey(LeviConfiguration.COLOR_COMPONENT_BACKGROUND)!= null)
        	o = config.getSettingByKey(LeviConfiguration.COLOR_COMPONENT_BACKGROUND);
        else
        	o = config.getSettingByKey(LeviConfiguration.COLOR_MODULE_BACKGROUND);
        if( o instanceof Color ) {
            moduleBackground = (Color)o;
        }

        if (config.getSettingByKey(LeviConfiguration.COLOR_COMPONENT_FOREGROUND)!= null)
        	o = config.getSettingByKey(LeviConfiguration.COLOR_COMPONENT_FOREGROUND);
        else
        	o = config.getSettingByKey(LeviConfiguration.COLOR_MODULE_FOREGROUND);
        if( o instanceof Color ) {
            moduleForeground = (Color)o;
        }

        if (config.getSettingByKey(LeviConfiguration.COLOR_COMPONENT_HIGHLIGHT)!= null)
        	o = config.getSettingByKey(LeviConfiguration.COLOR_COMPONENT_HIGHLIGHT);
        else
        	o = config.getSettingByKey(LeviConfiguration.COLOR_MODULE_HIGHLIGHT);
        if( o instanceof Color ) {
            highlightColor =(Color)o;
        }
        brighterBackground = moduleBackground.brighter().brighter();
    } // setColorsFromConfig

    @Override
    public void propertyChange(PropertyChangeEvent evt) {
        super.propertyChange(evt);
        if( evt.getPropertyName().startsWith("Color")) {
            setColorsFromConfig();
            repaint();
        }
    }


    boolean myVisibility;
    /**
     * Enable / disable view based on used channels
     * @param e
     * @see javax.swing.event.ChangeListener#stateChanged(javax.swing.event.ChangeEvent)
     */
    public void stateChanged(ChangeEvent e) {
        myVisibility = false;
        if( e.getSource() instanceof FlexRayScenario ) {
            FlexRayScenario escenario = (FlexRayScenario)e.getSource();
            for( Node n : escenario.getCluster().getNodes()) {
                if( n.getNodeConfiguration().pChannels().contains(channel)) {
                    myVisibility = true;
                    break;
                }
            } // for the nodes
            if( myVisibility != isVisible()) {
                EventQueue.invokeLater(new Runnable() {
                    public void run() {
                        setVisible(myVisibility);
                    }
                });
            } // if not visible
        } // if event source is flexray scenario
    } // stateChanged

    public void mouseDragged(MouseEvent e) {/**/}

    public void mouseMoved(MouseEvent e) {
    	Component hitComponent = getComponentAt(e.getPoint());
    	String tooltip;
    	
       if (scenario.getLearningModule().isTooltipsVisible())
       {
    	   // Tooltip fr den Knoten setzen
    	   tooltip = "<html><br>";
    	   tooltip=tooltip +scenario.getLearningModule().getRessources().getResourceByKey("flexray.network.tip")+"<br><br>";
    	   tooltip=tooltip+"</html>";
    	   ((ActiveStarCouplerComponent)hitComponent).setToolTipText(tooltip);
    	
       }
    }
    
}
