/*
 * $Id: AlphaBlendingComponent.java,v 1.1 2009/04/01 13:57:09 sirocic Exp $
 */
package flexray.gui;

import java.awt.AlphaComposite;
import java.awt.EventQueue;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import javax.swing.JComponent;

/**
 * Replacement for JComponent with alpha blending as well as automatic
 * alpha blending FX. Please note that this Component conflicts with
 * anti-aliasing in some cases.
 *
 * This module cooperates with ComponentDragger, so that a dragged
 * AlphaBlendingComponent can automatically be blended out and in on
 * dragging operations.
 *
 * IDEA:
 * should this not only implement a setDefault( alpha ) and a blendTo( alpha, direction, speed )?
 *
 * @author Dirk Frsterling
 *
 */
@SuppressWarnings("serial")
public class AlphaBlendingComponent extends JComponent
                                 implements PropertyChangeListener {

    // current alpha value.
    protected float alphaValue = 1.0f;

    //  blend delays in milliSeconds
    private long blendDelay = 5;
    private long blendLightDelay = 25;

    // blend operation flag: blending should be done
    protected boolean blend = false;

    // indicate if we are blending to our hilight alpha and if so, in which direction.
    protected BlendMode blendLight = BlendMode.NONE;

    /*
     * bounds for blending operations. protected versus private to speed
     * things up (because of the anonymous inner classes)
     */
    protected float alphaMin = 0.07f;   // minimum alpha for auto blending
    protected float alphaMax = 1.00f;   // maximum alpha for auto blending
    protected float alphaDefault = this.alphaMax; // default alpha value
    protected float alphaLight = 0.7f; // default highlight alpha value

    // the automatic blending mode of the module.
    protected BlendMode autoBlendMode = BlendMode.NONE;
    protected BlendMode savedAutoBlendMode = null;

    // "misuse" of the BlendMode enumeration: store the
    // direction in which to blend.
    protected BlendMode blendDirection = BlendMode.IN;

    private boolean blendOnDrag;
    private static Boolean defaultBlendOnDrag = Boolean.TRUE;
    private static final String DEFAULTBLENDONDRAG = "blendOnDrag";
    private static final JComponent propertyChangeInstance;
    static {
        propertyChangeInstance = new JComponent() {/* tricked you! */};
    }

    /**
     *  Enumerate the automatic blending "direction"
     *
     * @author Dirk Frsterling
     */
    public enum BlendMode {
        /** do not (auto) blend  */
        NONE,
        /** blend in */
        IN,
        /** blend out */
        OUT,
        /** blend in first, and then out again */
        IN_OUT,
        /** blend out first, and then in again */
        OUT_IN;

        /**
         * check if the blend mode contains blending in
         * @return true if the blendmode contains an "IN" direction, false otherwise
         */
        public boolean isIN() {
            if( this.equals(OUT) || this.equals(NONE) ) return false;
            return true;
        }

        /**
         * check if the blend mode contains blending out
         * @return true if the blendmode contains an "OUT" direction, false otherwise
         */
        public boolean isOUT() {
            if( this.equals(IN) || this.equals(NONE) ) return false;
            return true;
        }

    } // enum BlendMode

    /*
     * Auto blending scheduler
     */
    protected ScheduledExecutorService blendScheduler = null;
    ScheduledFuture<?> scheduledBlend = null;

    public AlphaBlendingComponent() {
        super();
        synchronized (defaultBlendOnDrag) {
            blendOnDrag = defaultBlendOnDrag.booleanValue();
        }
        propertyChangeInstance.addPropertyChangeListener(DEFAULTBLENDONDRAG,this);
        setDoubleBuffered(false);
    }

    /**
     * Set the default behaviour for blendOnDrag
     *
     * @param b true = on, false = off
     */
    public static void setDefaultBlendOnDrag(final boolean b) {
        boolean oldValue = true;
        synchronized (defaultBlendOnDrag) {
            if( b != defaultBlendOnDrag.booleanValue()) {
                oldValue = defaultBlendOnDrag.booleanValue();
                defaultBlendOnDrag = Boolean.valueOf(b);
                propertyChangeInstance.firePropertyChange(DEFAULTBLENDONDRAG, oldValue, b);
            }
        }
    }

    public static boolean isDefaultBlendOnDrag() {
        return defaultBlendOnDrag.booleanValue();
    }

    /**
     * Stop the autoblending. This stops the scheduler and
     * resets any state information.
     */
    protected void blendStop() {
        if( blendScheduler != null ) blendScheduler.shutdown();
        blend = false;
        blendLight = BlendMode.NONE;
        scheduledBlend = null;
        blendScheduler = null;
        blendDirection = BlendMode.NONE;
        if( savedAutoBlendMode != null ) {
            autoBlendMode = savedAutoBlendMode;
            savedAutoBlendMode = null;
        }
    } // blendStop()

    /**
     * paint this module. This includes softly blending in and out
     * if requested.
     *
     * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
     */
    @Override
    public void paintComponent( Graphics g ) {
        super.paintComponent( g );

        /*
         * If auto blending is requestet but not yet started,
         * create a scheduler and run the blending.
         */
        if( blend && isShowing() && (scheduledBlend == null )) {
            blendScheduler = Executors.newScheduledThreadPool( 1 );
            scheduledBlend = blendScheduler.scheduleAtFixedRate(
                    new Runnable() {
                        public void run() {
                            // Wait with blending until next visible.
                            if( !isShowing() ) return;
                            /*
                             * Determine the alpha value based on the current condition:
                             * Based on the direction: increase or decrease the alpha.
                             * If blending into hilight, stop at the hilight, else stop
                             * at min or max.
                             * After blending is complete, stop the process.
                             */
                            if( blendDirection.isIN() ) {
                                alphaValue += 0.02f;
                                switch( autoBlendMode ) {
                                case IN:
                                case OUT_IN:
                                    if( blendLight.equals(BlendMode.OUT) ) {
                                        if( alphaValue > alphaLight ) {
                                            alphaValue = alphaLight;
                                            blendStop();
                                        }
                                    } else if ( blendLight.equals(BlendMode.IN)) {
                                        if( alphaValue > alphaDefault ) {
                                            alphaValue = alphaDefault;
                                            blendStop();
                                        }
                                    } else if( alphaValue > alphaMax ) {
                                            alphaValue = alphaMax;
                                            blendStop();
                                    }
                                    break;
                                case IN_OUT:
                                    if( alphaValue > alphaMax ) {
                                        alphaValue = alphaMax;
                                        blendDirection = BlendMode.OUT;
                                    }
                                    break;
                                case NONE:
                                    // no blending
                                case OUT:
                                    // OUT not applicable to IN direction
                                    break;
                                }
                            } else if( blendDirection.isOUT() ) {
                                alphaValue -= 0.02f;
                                switch( autoBlendMode ) {
                                case OUT:
                                case IN_OUT:
                                    if( blendLight.equals(BlendMode.OUT) ) {
                                        if( alphaValue < alphaLight ) {
                                            alphaValue = alphaLight;
                                            blendStop();
                                        }
                                    } else  if( blendLight.equals(BlendMode.IN) ) {
                                        if( alphaValue < alphaDefault ) {
                                            alphaValue = alphaDefault;
                                            blendStop();
                                        }
                                    } else {
                                        if( alphaValue < alphaMin ) {
                                            alphaValue = alphaMin;
                                            blendStop();
                                        }
                                    }
                                    break;
                                case OUT_IN:
                                    if( alphaValue < alphaMin ) {
                                        alphaValue = alphaMin;
                                        blendDirection = BlendMode.IN;
                                    }
                                    break;
                                case NONE:
                                    // no blending
                                case IN:
                                    // IN not applicable to OUT direction
                                    break;
                                }
                            }
                            EventQueue.invokeLater( new Runnable(){
                                public void run() {
                                    repaint();
                                }
                            });
                        }
                    }, 0, ((blendLight!=BlendMode.NONE)?blendLightDelay:blendDelay), TimeUnit.MILLISECONDS);
        }
        if( (alphaValue != 1.0f) && (isShowing()) ) {
            Graphics2D g2d = (Graphics2D)g;
            if( isOpaque() ) setOpaque(false);
            g2d.setComposite( AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alphaValue ));
            g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            g2d.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
            g2d.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
        }
    }

    /**
     * Set the module's default alpha value
     * @param alpha
     */
    public void setAlphaDefault( float alpha ) {
        if( alpha > alphaMax ) {
            alphaValue = alphaMax;
            alphaDefault = alphaMax;
            alphaLight = alphaMax/2;
        } else if( alpha/2 < alphaMin ) {
            alphaValue = alphaMin*2;
            alphaDefault = alphaMin*2;
            alphaLight = alphaMin;
        } else {
            alphaValue = alpha;
            alphaDefault = alpha;
            alphaLight = alpha/2;
        }
    }

    /**
     * Set the alpha value of this module
     * @param alpha
     */
    public void setAlphaValue( float alpha ) {
        alphaValue = alpha;
    }

    /**
     * get this module's current alpha value
     * @return the module's current alpha value
     */
    public float getAlphaValue() {
        return alphaValue;
    }

    /**
     * Check if this module's alpha is the "light" value
     * @return true if this module is currently at light alpha
     */
    public boolean isLight() {
        if( alphaValue == alphaLight ) return true;
        return false;
    }

    /**
     * Set this module to "light" alpha
     * @param light
     */
    public void setLight( boolean light ) {
        if( light ) {
            alphaValue = alphaLight;
        } else {
            alphaValue = alphaDefault;
        }
    }

    /**
     * Request (temporary) blending to the hilight alpha value
     *
     * @param light
     */
    public void blendToLight ( boolean light ) {
        if( savedAutoBlendMode != null ) blendStop();
        savedAutoBlendMode = autoBlendMode;
        if( light ) {
            if( alphaLight < alphaDefault ) {
                autoBlendMode = BlendMode.OUT;
                blendDirection = BlendMode.OUT;
            } else {
                autoBlendMode = BlendMode.IN;
                blendDirection = BlendMode.IN;
            }
            blendLight = BlendMode.OUT;
        } else {
            if( alphaLight < alphaDefault ) {
                autoBlendMode = BlendMode.IN;
                blendDirection = BlendMode.IN;
            } else {
                autoBlendMode = BlendMode.OUT;
                blendDirection = BlendMode.OUT;
            }
            blendLight = BlendMode.IN;
        }
        blend = true;
    }

    /**
     * Start an auto blending the next time this component is showing.
     * If the component is not showing and more than one <code>blendOnNextShow</code>
     * is issued, the latest one will be in effect. So if you issue a blend request
     * with <code>OUT</code> blend mode while the component is not showing and you
     * later issue another blend request in <code>IN</code> mode while the component
     * is still not showing, the component will just blend in (or appear to do nothing
     * if it already was at the end alpha value).
     * @param reset
     */
    public void blendOnNextShow( boolean reset ) {
        // new blending requested, stop and delete previous
        if( scheduledBlend != null ) blendStop();
        // reset the alpha value to its default before blending?
        if( reset ) alphaValue = alphaDefault;

        // set the starting blend direction according to the
        // auto blend mode.
        switch( autoBlendMode ) {
        case IN:
        case IN_OUT:
            blendDirection = BlendMode.IN;
            break;
        case OUT:
        case OUT_IN:
            blendDirection = BlendMode.OUT;
            break;
        case NONE:
            // no blending
            break;
        }
        // flag the auto blending request to paintComponent
        blend = true;
    }

    /**
     * Set this component's blend delay. The blend delay ist the
     * time it takes between setting different shades of alpha
     *
     * @param blendDelay
     */
    public void setBlendDelay( long blendDelay ) {
        this.blendDelay = blendDelay;
    }

    /**
     * Get this component's current blend delay value
     * @return the blend delay of this component
     */
    public long getBlendDelay() {
        return this.blendDelay;
    }

    /**
     * get this component's upper bound for alpha values.
     * Trying to set the alpha value to anything above this value
     * will result in setting just this upper bound.
     * @return the upper bound for this component's alpha values
     */
    public float getAlphaMax() {
        return alphaMax;
    }

    /**
     * Set this component's upper bound for alpha values.
     * Trying to set the alpha value to anything above the value
     * set here will result in setting it just to this value.
     * If the current alpha value is already higher, it will be
     * re-set tho the value given here.
     * @param alphaMax
     */
    public void setAlphaMax(float alphaMax) {
        this.alphaMax = alphaMax;
        if ( alphaValue > alphaMax ) {
            alphaValue = alphaMax;
        }
    }

    /**
     * Get this component's lower bound for alpha values.
     * Trying to set the alpha value to anything below this value
     * will result in setting just this lower bound.
     * @return the lowe bound of this component's alpha value
     */
    public float getAlphaMin() {
        return alphaMin;
    }


    /**
     * Set this component's lower bound for alpha values.
     * Trying to set the alpha value to anything below the value
     * set here will result in setting it just to this value.
     * If the current alpha value is already lower, it will be
     * re-set tho the value given here.
     * @param alphaMin
     */
    public void setAlphaMin(float alphaMin) {
        this.alphaMin = alphaMin;
        if( alphaValue < alphaMin ) {
            alphaValue = alphaMin;
        }
    }

    /**
     * Return this component's auto blend mode.
     * See {@link BlendMode} for modes.
     * @return this component's auto blend mode
     */
    public BlendMode getAutoBlendMode() {
        return autoBlendMode;
    }

    /**
     * Set this component's auto blend mode.
     * See {@link BlendMode} for modes.
     * @param autoBlendMode
     */
    public void setAutoBlendMode(BlendMode autoBlendMode) {
        this.autoBlendMode = autoBlendMode;
    }

    /**
     * Get this components alpha value for "light" setting.
     * @return this component's "light" alpha value
     */
    public float getAlphaLight() {
        return alphaLight;
    }

    /**
     * Set this components alpha value for "light" setting
     * @param alphaLight
     */
    public void setAlphaLight(float alphaLight) {
        this.alphaLight = alphaLight;
    }

    /**
     * Get blend delay for blending to or from light setting
     * @return light blend delay
     */
    public long getBlendLightDelay() {
        return blendLightDelay;
    }

    /**
     * Set blend delay for blending to or from light setting
     * @param blendLightDelay
     */
    public void setBlendLightDelay(long blendLightDelay) {
        this.blendLightDelay = blendLightDelay;
    }

    /**
     * Check if this component blends on dragging.
     * @return true if this component wants to blend on dragging
     */
    public boolean isBlendOnDrag() {
        return blendOnDrag;
    }

    /**
     * Set if this component should blend on dragging
     * @param blendOnDrag
     */
    public void setBlendOnDrag(boolean blendOnDrag) {
        this.blendOnDrag = blendOnDrag;
    }

    /*
     * @param evt
     * @see java.beans.PropertyChangeListener#propertyChange(java.beans.PropertyChangeEvent)
     */
    public void propertyChange(PropertyChangeEvent evt) {
        if(evt.getPropertyName().equals(DEFAULTBLENDONDRAG)) {
            blendOnDrag = defaultBlendOnDrag.booleanValue();
        }
    }

}
