/*
 * $Id: ClockSymbol.java,v 1.1 2009/04/01 13:57:09 sirocic Exp $
 */
package flexray.gui;

import static flexray.gui.ClockSymbol.ClockSymbolDisplayMode.DOUBLECIRCLE;
import static java.lang.Math.PI;
import static java.lang.Math.cos;
import static java.lang.Math.sin;

import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;

import javax.swing.SwingConstants;


/**
 * A module to display a clock like symbol, which is similar to
 * a progress bar, but displays the current state in a circular,
 * clock-like way.<br>
 * The ClockSymbol has three display modes:<br>
 * <ul>
 * <li><tt>NEEDLE</tt>:
 *      Display just a needle.</li>
 * <li><tt>SINGLECIRCLE</tt>:
 * Display a filled arc that goes from the top of the clock symbol
 * clockwise up to the angle that represents the current state, so that
 * <tt>tick()</tt> slowly fills the whole symbol. At 100% (360 degrees), the next
 * <tt>tick()</tt> resets the symbol to be empty.</li>
 *
 * <li><tt>DOUBLECIRCLE</tt>:
 * Just like <tt>SINGLECIRCLE</tt> with one difference: After the circle is
 * completely filled, the next <tt>tick()</tt> will not reset the whole
 * symbol, but slowly &quot;unfill&quot; the symbol.
 * </li><br>
 * If the ClockSymbol contains text, then this text is placed within a centered
 * rectangle that is slightly bigger as the bounds of the text.
 *
 * @author Dirk Frsterling
 *
 */
public class ClockSymbol extends AlphaBlendingComponent implements SwingConstants {

    private static final long serialVersionUID = 2097343226015299931L;

    /**
     * Enumeration for the ClockSymbol's display modes
     * @author Dirk Frsterling
     */
    public enum ClockSymbolDisplayMode {
        NEEDLE, SINGLECIRCLE, DOUBLECIRCLE
    }

    private int posX = 0;
    private int posY = 0;
    private int width = 15;
    private int height = 15;
    private int angle = 0;
    private String text = null;
    private boolean growing = true;
    private ClockSymbolDisplayMode mode = DOUBLECIRCLE;

    /**
     * Must define explicit default constructor to allow
     * default ClockSymbols
     */
    public ClockSymbol() {
        super();
        setOpaque( false );
    }

    /**
     * Constructor for a ClockSymbol with a specific size
     * @param size size of the bounding rectangle of this ClockSymbol
     */
    public ClockSymbol( int size ) {
        this();
        width = size;
        height = size;
    }

    /**
     * Construct a ClockSymbol with a text.
     * The text is displayed in the center of the ClockSymbol.
     * This constructor is unsafe, because the size of the text remains
     * unchecked. The text may exceed the bounds of the ClockSymbol.
     * @param text
     */
    public ClockSymbol( String text ) {
        this();
        this.text = text;
    }

    /**
     * Construct a ClockSymbol with text and a specified size.
     * @param size
     * @param text
     */
    public ClockSymbol( int size, String text ) {
        this( size );
        this.text = text;
    }

    /*  (non-javadoc)
     * @see java.awt.Component#getPreferredSize()
     */
    @Override
    public Dimension getPreferredSize() {
        return new Dimension( width, height );
    }

//    /*  (non-javadoc)
//     * @see java.awt.Component#contains(int, int)
//     */
//    @Override public boolean contains(int x, int y) {
//        return new Ellipse2D.Double( 0,0, this.width+1, this.height+1 ).contains(x+1,y+1);
//    }

    /**
     * Paint method for the ClockSymbol<br>
     * <br>
     * This uses fillArc, which cannot correctly paint filled arcs with small angles if
     * the overall size of the bounding rectangle is small, too. <br>
     * <pre>
     * width/height    minimum angle
     *     5                12
     *     6                10
     *     7                 9
     *     8                 8
     *     9                 7
     *    10                 6
     *    11                 5
     *    15                 4
     *    20                 3
     *    29                 2
     *    58                 1    <-- the first size at which 1 does work.
     * </pre>
     * <br>
     * Workaround: Run the jvm with <tt>-Dsun.java2d.ddoffscreen=false</tt><br>
     * See <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6357387">bug #6357387</a>
     *
     * @author Dirk Frsterling
     * @param g the <code>Graphics</code> context in which to paint
     */
    @Override
    public void paintComponent( Graphics g ) {
        super.paintComponent( g );

        // Kreise hbsch machen.
        ((Graphics2D)g).setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);

        /*
         * fill the complete background (opaque) or just the
         * circular shape (not opaque)
         */
        g.setColor( getBackground() );
        if( isOpaque() ) {
            g.fillRect( posX, posY, width, height );
        } else {
            g.fillOval( posX, posY, width, height );
        }
        /*
         * draw an oval to display the shape of the symbol.
         */
        g.setColor( getForeground() );
        g.drawOval( posX , posY , width-1, height-1 );

        /*
         * Depending on the drawing mode, fill the needed parts of the symbol
         * or draw a needle.
         */
        switch( mode ) {
        case SINGLECIRCLE:
            g.fillArc(  posX , posY , width, height, 90 , -angle );
            break;
        case DOUBLECIRCLE:
            if( growing )
                g.fillArc(  posX , posY , width, height, 90 , -angle );
            else
                g.fillArc(  posX , posY , width, height, 90 , 360-angle );
            break;
        case NEEDLE:
            int radius = width / 2;
            int cx = posX + radius;
            int cy = posY + radius;
            double rad = angle * ( PI / 180.0 );
            double x = sin( rad ) * radius;
            double y = cos( rad ) * radius;
            // paint a small dot if no text is to be displayed
            if ( text != null ) g.fillOval( cx - 1, cy - 1, 3, 3 );
            g.drawLine( cx, cy, cx + (int)x, cx-(int)y );
            break;
        }

        /*
         * if there is text, make a small box in background color
         * and draw the text into it.
         */
        if( text != null ) {
            // set a boldfaced font
            Font fo = g.getFont();
            g.setFont( fo.deriveFont( Font.BOLD ) );

            // get the bounds of the string to be drawn and
            // calculate the center position.
            FontMetrics f = g.getFontMetrics();
            int fx, fy , fw, fh;
            fw = (int)(f.getStringBounds( text, g ).getWidth());
            fh = f.getLeading() + f.getAscent();
            fx = posX + ( width / 2 )-( fw / 2 );
            fy = posY + ( height / 2 ) + ( f.getLeading() + f.getAscent() ) / 2;

            g.setColor( getBackground() );

            g.fillRect( fx - 1, fy - fh + 2, fw + 1, fh );
            g.setColor( getForeground() );

            // g.setXORMode( Color.BLACK ); // alternative to the two above

            g.drawString( text, fx , fy );
        }
    } // paint()

    /**
     * Set the angle that should be displayed by this ClockSymbol.<br>
     * Please note that the actual display depends on the <tt>growing</tt>
     * value if in <tt>DOUBLECIRCLE</tt> mode.
     *
     * @param angle the angle to display
     */
    public void setAngle( int angle ) {
        if( (angle < 0) || angle > 360 )
            throw new IllegalArgumentException( "angle must be between 0 and 360" );
        this.angle = angle;
        repaint();
    }

    /**
     * Let this ClockSymbol display an angle that corresponds to
     * the given percent value.<br>
     * Please note that the actual display depends on the <tt>growing</tt>
     * value if in <tt>DOUBLECIRCLE</tt> mode.
     *
     * @param percent percentage this symbol should draw.
     */
    public void setPercent( int percent ) {
        if( (percent < 0) || percent > 100 )
            throw new IllegalArgumentException( "percent must be between 0 and 100" );
        this.angle = (int)(3.6 * percent);
        repaint();
    }

    /**
     * Change <tt>growing</tt> status for DOUBLECIRCLE mode.
     * @param growing
     */
    public void setGrowing( boolean growing ) {
        this.growing = growing;
    }

    /**
     * Return the current <tt>growing</tt> status.
     *
     * @return false if the ClockSymbol is in DOUBLECIRCLE mode and the inverse
     * arc is to be painted. true otherwise.
     */
    public boolean isGrowing() {
        return growing;
    }

    /**
     * "Advance" the display by the given percentage of a full circle.
     *
     * @param percent
     */
    public void tickPercent( int percent ) {
        if( (percent < -100) || percent > 100 )
            throw new IllegalArgumentException( "percent must be between -100 and 100" );
        tick( (360/100)*percent );
    }

    /**
     * "Advance" the display by the given angle.<br>
     * This method respects the <tt>growing</tt> status
     *
     * @param tickAngle
     */
    public void tick( int tickAngle ) {
        angle += tickAngle;
        if( angle >= 360 ) {
            growing = !growing;
            angle -= 360;
        } else if( angle < 0 ) {
            growing = !growing;
            angle +=360;
        }
        repaint();
    }

    /**
     * Set the display mode. There are three:
     * <ul>
     * <li><tt>NEEDLE</tt>:
     *      Display just a needle.</li>
     * <li><tt>SINGLECIRCLE</tt>:
     * Display a filled arc that goes from the top of the clock symbol
     * clockwise up to the angle that represents the current state, so that
     * <tt>tick()</tt> slowly fills the whole symbol. At 100% (360 degrees), the next
     * <tt>tick()</tt> resets the symbol to be empty.</li>
     *
     * <li><tt>DOUBLECIRCLE</tt>:
     * Just like <tt>SINGLECIRCLE</tt> with one difference: After the circle is
     * completely filled, the next <tt>tick()</tt> will not reset the whole
     * symbol, but slowly &quot;unfill&quot; the symbol.
     * </li><br>
     *
     * @param pMode the display mode to set
     */
    public void setDisplayMode( ClockSymbolDisplayMode pMode ) {
        mode = pMode;
        repaint();
    }

} // class ClockSymbol
