/*
 * $Id: CommCycleComponent.java,v 1.3 2009/04/28 04:48:55 sirocic Exp $
 *
 * Created on 17.05.2006 07:47:12
 * by Dirk Frsterling <r@zorbla.de>
 */
package flexray.gui;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.font.FontRenderContext;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.lang.Math;
import java.util.Locale;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JToggleButton;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import learningVisualizations.LeviConfiguration;
import learningVisualizations.LeviModule;
import learningVisualizations.LeviResourceManager;
import learningVisualizations.simulator.LeviSimulator;
import learningVisualizations.simulator.LeviTimer;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import flexray.ModuleFR;
import flexray.cluster.Cluster;
import flexray.cluster.Node;
import flexray.communication.FlexRayFrame;
import flexray.configuration.ChannelName;
import flexray.configuration.ClusterConfiguration;
import flexray.configuration.NodeConfiguration;
import flexray.gui.actions.ChangeAutoMoveAction;
import flexray.gui.dialogs.CommCycleConfigDialog;
import flexray.gui.dialogs.SlotDialog;
import flexray.levi.gui.FrameFR;
import flexray.scenario.FlexRayScenario;
import flexray.xml.XMLTags;


/**
 * This component implements the visualization and control of
 * the communication cycle.
 *
 * - Resize is handled by this (MouseListener#mouseDragged)
 *
 *
 * @author Dirk Frsterling <r@zorbla.de>
 *
 */
public class CommCycleComponent extends AlphaBlendingComponent
                             implements ChangeListener,
                                        ComponentListener,
                                        AutoMoveable,
                                        DragZone,
                                        MouseMotionListener,
                                        MouseListener,
                                        XMLTags {

    /*
     *   -----------------------
     *    view/control related
     *   -----------------------
     *
     */

    private final LeviResourceManager res;
    private ModuleFR learningModul;
    private String staticSegmentString;
    private String dynamicSegmentString;

    private final JLabel sizeButton;
    private final JToggleButton autoMoveButton;

    private static final AlphaComposite transparentComposite =
        AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.333f);

    private static Color slotHighlightInnerColor = null;
    private static Color slotHighlightBorderColor = Color.BLACK;

    private static final Rectangle2D.Double emptySlotRect = new Rectangle2D.Double(-1,-1,-1,-1);

    private Rectangle2D.Double highlightSlotRect = emptySlotRect;
    private Rectangle2D.Double titleRect = null;

    private boolean resized = true;

    private boolean drawStaticLabel = false;
    private String staticLabelString ="";
    private int staticLabelStart = 0;
    private int staticLabelWidth = 0;

    private boolean drawDynamicLabel = false;
    private String dynamicLabelString;
    private int dynamicLabelStart = 0;
    private int dynamicLabelWidth = 0;

    private int nitborder;
    private int staticSegmentWidth;
    private int dynamicSegmentWidth;
    private int nitwidth;


    // ----------------------- layout related

    private int autoMoveButtonSize;

    /**
     * preferred width of this component
     */
    private int prefWidth = 0;

    /**
     * Minimum (and initial) width of a macrotick
     * (this implicitly overrides the minWidth setting)
     */
    private int minMTWidth = 2;
    private int mtWidth = 15; // initial MacroTick width

    /**
     * Minimum width for this component
     */
    private final int minWidth = 90;

    private int maxStaticSlotX = 0;

    private int maxMiniSlotX = 0;

    int dynborder = 0;

    private int slotStart = 26;

    /**
     * Preferred height of this component
     */
    private int prefHeight = 0;

    /**
     * minumum height for this component
     */
    private final int minHeight = 105;

    /**
     * Height of the MacroTick-scale
     */
    private int scaleHeight = 10;

    /**
     * Height of the slot drawing area
     */
    private int slotsHeight = 0;

    /**
     * Overall padding
     */
    private final int pad = 4;

    /**
     * Size of the dropshadow
     */
    private final int shadowSize = 4;

    /**
     * The "inner" dimension (without shadow)
     */
    private int innerX = 0;
    private int innerY = 0;
    private int innerWidth = 0;
    private int innerHeight = 0;

    /**
     *  padding between slots
     */
    private final int slot_padding = 2; // pixels

    /**
     * Vertical start of channel A drawing area
     */
    int channelAY;
    /**
     * Vertical size of channel A drawing area
     */
    int channelAH;

    /**
     * Vertical start of channel B drawing area
     */
    int channelBY;
    /**
     * Vertical size of channel B drawing area
     */
    int channelBH;

    int staticSlotWidth = 0;

    int miniSlotWidth = 0;


    // ---------------------- control related

    /**
     * A ComponentDragger to be used for dragging this component
     */
    private final ComponentDragger componentDragger;

    /**
     * autoMove: true if the component attaches itself to
     * the lower left corner of the window
     */
    private boolean autoMove = true;

    /**
     * Storage for the previous non-autoMove location
     */
    private Point prevLocation = null;

    private Point mousePos = null;

    private int prevSlot[] = {-1, -1};

    JButton configButton;
    int locationPad = 3;

    // ----------------------- scenario related

    private final Cluster cluster;
    private final FlexRayScenario scenario;
    private ClusterConfiguration clusterConfiguration;

    protected LeviModule module;
    
    private Font  labelFont;
    private Color forgroundColor;
    
    
    
    /**
     * Create a graphical component to display the communication cycle
     *
     * @param pScenario
     * @param pComponentDragger
     */
    public CommCycleComponent( ModuleFR learningModul, FlexRayScenario pScenario, ComponentDragger pComponentDragger ) {
//        super();
    	this.learningModul = learningModul;
        this.scenario = pScenario;
        this.module = pScenario.getLearningModule();
        this.cluster = pScenario.getCluster();
        this.clusterConfiguration = cluster.getClusterConfiguration();
        this.res = pScenario.getLearningModule().getRessources();
        this.componentDragger = pComponentDragger;
        addMouseListener(componentDragger);
        this.setOpaque(false);
        
        if( slotHighlightInnerColor == null )
            slotHighlightInnerColor = (Color)ModuleFR.getModule().getConfiguration().getSettingByKey(LeviConfiguration.COLOR_MODULE_HIGHLIGHT);
        if( slotHighlightInnerColor == null ) // still nothing
            slotHighlightInnerColor = Color.YELLOW;

        autoMoveButton = new JToggleButton();
        autoMoveButton.setAction(new ChangeAutoMoveAction(this));

        autoMoveButton.setLocation(0,0);
        autoMoveButton.setEnabled(true);
        autoMoveButton.setSelected(true);

        autoMoveButtonSize = 10;
        autoMoveButton.setSize(autoMoveButtonSize,autoMoveButtonSize);
        this.add(autoMoveButton);

        // some initial width
        prefWidth = pad + slotStart +
        (mtWidth * clusterConfiguration.gMacroPerCycle()) +
        pad + +autoMoveButtonSize + shadowSize;

        prefHeight = 120;

        sizeButton = new JLabel( " " );
        sizeButton.setBorder(BorderFactory.createLineBorder(Color.BLACK, 1));
        sizeButton.setSize(autoMoveButtonSize,autoMoveButtonSize);
        sizeButton.setLocation(prefWidth-shadowSize-autoMoveButtonSize, 0);
        sizeButton.setOpaque(true);
        sizeButton.setBackground(Color.CYAN);

        sizeButton.addMouseMotionListener(this);
        sizeButton.setCursor(Cursor.getPredefinedCursor(Cursor.NE_RESIZE_CURSOR));
        this.add(sizeButton);

        innerWidth = prefWidth-shadowSize-pad-pad-autoMoveButtonSize;
        innerHeight = prefHeight-autoMoveButtonSize-shadowSize-pad-pad;
        innerX = pad;
        innerY = autoMoveButtonSize + pad;

        titleRect = new Rectangle2D.Double();

        titleRect.setRect(autoMoveButtonSize,
                          1,
                          prefWidth-shadowSize-autoMoveButtonSize,
                          autoMoveButtonSize-1);

        addMouseListener(this);
        addMouseMotionListener(this);

        configButton = new JButton();
        ImageIcon configIcon = NodeComponent.configIcon;
        int iconPad = 6;
        configButton.setIcon(configIcon);
        configButton.setActionCommand("config");
        configButton.setFocusPainted(false);
        configButton.setFocusable(false);
        configButton.setSize(configIcon.getIconWidth()+iconPad,configIcon.getIconHeight()+iconPad);
        configButton.setLocation(new Point(locationPad,prefHeight-5-configButton.getHeight()-locationPad));
       
        add(configButton);
        configButton.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                JButton source = (JButton)e.getSource();
                int x = source.getX() + source.getWidth()/2;
                int y = source.getY() + source.getHeight()/2;
                mouseRightClicked(new MouseEvent(source,e.getID(),e.getWhen(),e.getModifiers(),x,y,1,false));
            }
        });
        
        // Font, Sprache und Farben intialisieren...
        updateLanguageColorFonts();
        
    }

    /**
     * paint the comm cycle component
     * TODO: This <strong>has</strong> to be optimized for performance!
     * @param g
     * @see flexray.gui.AlphaBlendingComponent#paintComponent(java.awt.Graphics)
     */
    @Override
    public void paintComponent(Graphics g) {

        Rectangle2D boundsLabel;
  	    FontRenderContext context;
  	    String slotText;
  	    int slotTextX, slotTextY;
        
        super.paintComponent(g);
        Dimension preferredSize = getPreferredSize();
        if( !preferredSize.equals(getSize())) {
            setSize(preferredSize);
        }

        Graphics2D g2d = (Graphics2D)g;
        context = g2d.getFontRenderContext();
        g2d.setColor(getBackground().darker().darker());

        fitPanel();
        
        /*
         * a dropshadow
         */
        // bottom
        g2d.fillRect(shadowSize,
                     preferredSize.height-shadowSize,
                     preferredSize.width,
                     shadowSize);
        // right
        g2d.fillRect(innerWidth+pad+pad,
                     shadowSize+autoMoveButtonSize+pad,
                     shadowSize,
                     innerHeight-shadowSize+pad);

        // the inner of the component
        g2d.setColor(new Color(255,255,200));
        g2d.fillRect(0,
                     autoMoveButtonSize,
                     innerWidth+pad+pad,
                     innerHeight+pad+pad);

        // the edge of the component
        g2d.setColor(getForeground());
        g2d.drawRect(0,
                     autoMoveButtonSize,
                     innerWidth - 1 + pad + pad,
                     innerHeight - 1 + pad + pad );

        Color savedColor = g2d.getColor();

        /*
         * start drawing something real
         */

        /*
         * Draw the backgrounds of the segments
         */

        /*
         * The macrotick scale
         */
        for( int i = 0 ; i <= clusterConfiguration.gMacroPerCycle() ; i++ ) {
            if( (i % 5) == 0 ) {
                g2d.setColor(Color.RED);
                g2d.drawLine(innerX + slotStart + i * mtWidth,
                             innerY + innerHeight - (scaleHeight / 5) * 4,
                             innerX + slotStart + i * mtWidth,
                             innerY + innerHeight );
            } else {
                g2d.setColor(Color.BLACK);
                g2d.drawLine(innerX + slotStart + i * mtWidth,
                             innerY + innerHeight - scaleHeight / 3,
                             innerX + slotStart + i * mtWidth,
                             innerY + innerHeight );
            }
        }
        g2d.setColor(Color.BLACK);
        g2d.drawString("M",
                innerX + slotStart + clusterConfiguration.gMacroPerCycle() * mtWidth + 2,
                innerY + innerHeight - 12);
        g2d.drawString("T",
                innerX + slotStart + clusterConfiguration.gMacroPerCycle() * mtWidth + 3,
                innerY + innerHeight);

        if( !autoMove ) {
            g2d.setColor(Color.WHITE);
            g2d.fill(titleRect);
            g2d.setColor(Color.BLACK);
            g2d.draw(titleRect);
            g2d.setColor(Color.GRAY);
            g2d.drawLine((int)titleRect.getX()+2, 4,
                    (int)(titleRect.getX()+titleRect.getWidth())-4, 4);
            g2d.drawLine((int)titleRect.getX()+2, 7,
                    (int)(titleRect.getX()+titleRect.getWidth())-4, 7);
        }

        /*
         * static and dynamic segment labels
         */
        Font savedFont = g2d.getFont();
        Font staticFont = savedFont.deriveFont(10f);
        g2d.setFont(staticFont);
        FontMetrics fm = g2d.getFontMetrics();
        int strlen = staticSegmentString.length();
        int cut = 1;
        int strmid = strlen/2;


        /*
         * if resized, calculate everything for the labels
         */
        if( resized ) {
            staticLabelString = staticSegmentString;
            staticLabelWidth = fm.stringWidth(staticSegmentString) + 4;
            staticLabelStart = (dynborder/2) - (staticLabelWidth/2) + (slotStart/2);
            while( ((staticLabelWidth + autoMoveButtonSize + shadowSize + pad + slotStart + 1) > staticSegmentWidth) &&
                    (strmid >= cut)) {
                String strleft = staticSegmentString.substring(0,strlen/2-cut);
                String strright = staticSegmentString.substring(strlen/2+cut);
                cut++;
                staticLabelString = strleft + ".." + strright;
                staticLabelWidth = fm.stringWidth(staticLabelString) + 4;
                staticLabelStart = (dynborder/2) - (staticLabelWidth/2) + (slotStart/2);
            }
            if ((staticLabelWidth + autoMoveButtonSize  < staticSegmentWidth) &&
                    (strmid >= cut)) drawStaticLabel = true;
            else drawStaticLabel = false;

            if( clusterConfiguration.gNumberOfMinislots() > 0 ) {
                drawDynamicLabel = false;
                dynamicLabelString = dynamicSegmentString;
                dynamicLabelWidth = fm.stringWidth(dynamicSegmentString) + 4;
                dynamicLabelStart = dynborder + ((nitborder-dynborder+pad)/2) - (dynamicLabelWidth/2);

                strlen = dynamicSegmentString.length();
                cut = 1;
                strmid = strlen / 2;

                while( ((dynamicLabelWidth + autoMoveButtonSize + shadowSize + pad + 1) > dynamicSegmentWidth) &&
                        strmid >= cut) {
                    String strleft = dynamicSegmentString.substring(0,strlen/2-cut);
                    String strright = dynamicSegmentString.substring(strlen/2+cut);
                    cut++;
                    dynamicLabelString = strleft + ".." + strright;
                    dynamicLabelWidth = fm.stringWidth(dynamicLabelString) + 4;
                    dynamicLabelStart = dynborder + ((nitborder-dynborder+pad)/2) - (dynamicLabelWidth/2);
                }
                if( (dynamicLabelWidth > 0) && (strmid >= cut) ) drawDynamicLabel = true;
            }
            resized = false;
        } else {
            // drawing cascade after loading default scenario might have left the button at
            // a wrong position :(
            if( configButton.getY() > prefHeight-5-configButton.getHeight() - locationPad)
                configButton.setLocation(new Point(locationPad,prefHeight-5-configButton.getHeight()-locationPad));
        }

        /*
         * static segment label
         */
        if(drawStaticLabel) {
            g2d.setColor(Color.WHITE);
            g2d.fillRect(staticLabelStart, 0, staticLabelWidth, autoMoveButtonSize + 2);
            g2d.setColor(Color.BLACK);
            g2d.drawRect(staticLabelStart, 0, staticLabelWidth, autoMoveButtonSize + 2);
            g2d.drawString(staticLabelString, staticLabelStart+2, autoMoveButtonSize);
        }


        /*
         * dynamic segment label
         */
        if(drawDynamicLabel) {
            g2d.setColor(Color.WHITE);
            g2d.fillRect(dynamicLabelStart, 0, dynamicLabelWidth, autoMoveButtonSize + 2);
            g2d.setColor(Color.BLACK);
            g2d.drawRect(dynamicLabelStart, 0, dynamicLabelWidth, autoMoveButtonSize + 2);
            g2d.drawString(dynamicLabelString, dynamicLabelStart+2, autoMoveButtonSize);
        }
        g2d.setFont(savedFont);


        /*
         * Draw the communication structure
         */
        LeviSimulator sim = scenario.getLearningModule().getSimulator();

        int macroTickInCycle = 0;
        if( !sim.isStop() ) { // calculate the cycle relative macrotick
            LeviTimer timer = sim.getTimer();
            int time = timer.getTime() - 1;
            int mpc = clusterConfiguration.gMacroPerCycle();
            int cycle = time / mpc;
            macroTickInCycle = time - cycle*mpc + 1;
        }

        Composite composite = g2d.getComposite();

        /*
         * Draw static slots
         */
        if( cluster != null ) {
            /*
             * default coordinates and sizes (dual channel)
             */
            channelAY = innerY;
            channelAH = (slotsHeight/2) - pad/2;
            channelBY = channelAY + channelAH + pad;
            channelBH = channelAH;
            ChannelName channels = clusterConfiguration.gChannels();

            /*
             * set the coordinates and sizes according to the channel config
             */
            switch( channels ) {
            case A:
                channelAY = innerY;
                channelAH = slotsHeight;
                channelBY = 0;
                channelBH = 0;
                break;
            case B:
                channelAY = 0;
                channelAH = 0;
                channelBY = innerY;
                channelBH = slotsHeight;
                break;
            case AB:
                // defaults already set
                break;
            case NONE:
                channelAY = 0;
                channelAH = 0;
                channelBY = 0;
                channelBH = 0;
                break;
            }

            /*
             * fetch the static slot configuration and calculate the width to
             * display the static slots
             */
            int staticSlots = clusterConfiguration.gNumberOfStaticSlots();

            Node node;

            int sx = 0;
            /*
             * Basics for all slots
             */
            Stroke savedStroke = g2d.getStroke();
            BasicStroke lightDashStroke = new BasicStroke(1,
                    BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL,
                    1, new float[]{ 2,2 }, 1 );
            BasicStroke boldStroke = new BasicStroke(4f);

            /*
             * Channel A + B label
             */
            if(channels.contains(ChannelName.A)) {
                g2d.setFont(g2d.getFont().deriveFont(Font.BOLD));
                fm = g2d.getFontMetrics();
                Rectangle2D strBounds = fm.getStringBounds("A", g2d);
                int strwidth = (int)strBounds.getWidth() + 4;
                int t = (slotStart-strwidth)/2 + 2;
                g2d.setColor(Color.BLACK);
                g2d.drawString("A", innerX + t,
                        channelAY + channelAH/2 + (int)strBounds.getHeight()/2 - pad );
                g2d.setFont(savedFont);
            }
            if( channels.contains(ChannelName.B)) {
                g2d.setFont(g2d.getFont().deriveFont(Font.BOLD));
                fm = g2d.getFontMetrics();
                Rectangle2D strBounds = fm.getStringBounds("B", g2d);
                int strwidth = (int)strBounds.getWidth() + 4;
                int t = (slotStart-strwidth)/2 + 2;
                g2d.setColor(Color.BLACK);
                g2d.drawString("B", innerX + t,
                        channelBY + (channelBH/2) + (int)strBounds.getHeight()/2 - pad );
                g2d.setFont(savedFont);
            }

            /*
             * for all static slots...
             */
            for( int i = 0; i < staticSlots; i++ ) {
                // calculate x position of static slot graphic
                sx = innerX + slotStart + (i * (staticSlotWidth + slot_padding));
                int firstMTforSlot = i * clusterConfiguration.gdStaticSlot() + 1;
                /*
                 * Channel A (static slots)
                 */
                node = cluster.getNodeForSlot(ChannelName.A, i);
                NodeConfiguration nodeConfig = null;
                if( node != null ) nodeConfig = node.getNodeConfiguration();
                if(channels.contains(ChannelName.A)) {
                    Rectangle2D.Double slotRect = new Rectangle2D.Double(sx,channelAY,staticSlotWidth,channelAH);
                    if( (nodeConfig != null) && (nodeConfig.pChannels().contains(ChannelName.A)) ) {
                        assert(node!=null);
                        g2d.setColor(scenario.getNodeComponent(node).getNodeColor());
                        
                        if( sim.isStop() || (macroTickInCycle >= firstMTforSlot) ) {
                            if( node.isSendingStatic(ChannelName.A)) {
                                g2d.fillRect(sx+1,channelAY+1,staticSlotWidth-2,channelAH-2);
                            } else {
                                g2d.setStroke(boldStroke);
                                g2d.drawRect(sx+2, channelAY+2, staticSlotWidth-1-3, channelAH-1-3);
                                g2d.setStroke(savedStroke);
                            }
                        }
                        
                        if( (!sim.isStop()) && (macroTickInCycle >= firstMTforSlot)) {
                        	g2d.setColor(Color.BLACK);
                            g2d.drawRect(sx, channelAY, staticSlotWidth-1, channelAH-1);
                        }
                    }
                    if( sim.isStop() || (macroTickInCycle < firstMTforSlot) ) {
                        g2d.setStroke(lightDashStroke);
                        g2d.setColor(Color.LIGHT_GRAY);
                        g2d.drawRect(sx,channelAY,staticSlotWidth-1,channelAH-1);
                        g2d.setStroke(savedStroke);
                    }
                    if( !sim.isStop() && (macroTickInCycle >= firstMTforSlot)&& (node ==null)) {
                        g2d.setStroke(lightDashStroke);
                        g2d.setColor(forgroundColor);
                        g2d.drawRect(sx,channelAY,staticSlotWidth-1,channelAH-1);
                        g2d.setStroke(savedStroke);
                    }
                    //Beschriftung der Static Slots fr Kanal A
                    if( sim.isStop() || (macroTickInCycle >= firstMTforSlot) ) {
                    	if (node == null)
                    		g2d.setColor(Color.LIGHT_GRAY);
                    	else
                    		g2d.setColor(forgroundColor);
                    }
                    else {
                    		g2d.setColor(Color.LIGHT_GRAY);
                    }
                    if( (!sim.isStop()) && (macroTickInCycle >= firstMTforSlot)) 
                		g2d.setColor(forgroundColor);
                    g2d.setFont(labelFont);
                    slotText = "SS "+(i+1);
                    boundsLabel = getFont().getStringBounds(slotText, context);
              	    slotTextX = sx + staticSlotWidth/2 - (int)boundsLabel.getWidth()/2; 
              	    slotTextY = channelAY+channelAH/2+ (int)boundsLabel.getHeight()/2;
                    g2d.drawString(slotText, slotTextX, slotTextY);

                    if((mousePos != null) && slotRect.contains(new Point2D.Double(mousePos.x,mousePos.y)) ) {
                        highlightSlotRect = slotRect;
                        g2d.setComposite(transparentComposite);
                        g2d.setColor(slotHighlightInnerColor);
                        final int delta=1;
                        g2d.fillRect(sx-delta,channelAY-delta,staticSlotWidth+delta+delta,channelAH+delta+delta);
                        g2d.setColor(slotHighlightBorderColor);
                        g2d.setComposite(composite);
                        g2d.drawRect(sx-delta,channelAY-delta,staticSlotWidth+delta+delta-1,channelAH+delta+delta-1);
                    }

                } // if (channel A)

                node = cluster.getNodeForSlot(ChannelName.B, i);
                nodeConfig = null;
                if( node != null ) nodeConfig = node.getNodeConfiguration();
                /*
                 * Channel B (static slots)
                 */
                if( channels.contains(ChannelName.B)) {
                    Rectangle2D.Double slotRect = new Rectangle2D.Double(sx,channelBY,staticSlotWidth,channelBH);
                    if( (nodeConfig != null) && (nodeConfig.pChannels().contains(ChannelName.B)) ) {
                        assert(node!=null);
                        g2d.setColor(scenario.getNodeComponent(node).getNodeColor());
                        if( sim.isStop() || (macroTickInCycle >= firstMTforSlot) ) {
                            if( node.isSendingStatic(ChannelName.B)) {
                                g2d.fillRect(sx+1,channelBY+1,staticSlotWidth-2,channelBH-2);
                            } else {
                                g2d.setStroke(boldStroke);
                                g2d.drawRect(sx+2, channelBY+2, staticSlotWidth-1-3, channelBH-1-3);
                                g2d.setStroke(savedStroke);
                            }
                        }
                        if( (!sim.isStop()) && (macroTickInCycle >= firstMTforSlot)) {
                        	g2d.setColor(Color.BLACK);
                            g2d.drawRect(sx, channelBY, staticSlotWidth-1, channelBH-1);
                        }
                    }
                    if( sim.isStop() || (macroTickInCycle < firstMTforSlot) ) {
                        g2d.setStroke(lightDashStroke);
                        g2d.setColor(Color.LIGHT_GRAY);
                        g2d.drawRect(sx,channelBY,staticSlotWidth-1,channelBH-1);
                        g2d.setStroke(savedStroke);
                    }
                    if( !sim.isStop() && (macroTickInCycle >= firstMTforSlot)&& (node ==null)) {
                        g2d.setStroke(lightDashStroke);
                        g2d.setColor(forgroundColor);
                        g2d.drawRect(sx,channelBY,staticSlotWidth-1,channelBH-1);
                        g2d.setStroke(savedStroke);
                    }
                    //  Beschriftung statische Slots Kanal B
                    if( sim.isStop() || (macroTickInCycle >= firstMTforSlot) ) {
                    	if (node == null)
                    		g2d.setColor(Color.LIGHT_GRAY);
                    	else
                    	g2d.setColor(forgroundColor);
                    }
                    else
                    	g2d.setColor(Color.LIGHT_GRAY);
                    if( (!sim.isStop()) && (macroTickInCycle >= firstMTforSlot)) 
                		g2d.setColor(forgroundColor);
                    g2d.setFont(labelFont);
                    slotText = "SS "+(i+1);
                    boundsLabel = getFont().getStringBounds(slotText, context);
              	    slotTextX = sx + staticSlotWidth/2 - (int)boundsLabel.getWidth()/2; 
              	    slotTextY = channelBY+channelBH/2+(int)boundsLabel.getHeight()/2;
                    g2d.drawString(slotText, slotTextX, slotTextY);
                    
                    if((mousePos != null) && slotRect.contains(new Point2D.Double(mousePos.x,mousePos.y)) ) {
                        highlightSlotRect = slotRect;
                        g2d.setComposite(transparentComposite);
                        g2d.setColor(slotHighlightInnerColor);
                        final int delta=1;
                        g2d.fillRect(sx-delta,channelBY-delta,staticSlotWidth+delta+delta,channelBH+delta+delta);
                        g2d.setColor(slotHighlightBorderColor);
                        g2d.setComposite(composite);
                        g2d.drawRect(sx-delta,channelBY-delta,staticSlotWidth+delta+delta-1,channelBH+delta+delta-1);
                    }
                } // if (channel B)
            } // for all static slots
            maxStaticSlotX = sx + staticSlotWidth-1;



            /*
             * dynamic slots (minislots)
             */

            if( sim.isStop() ) {
                /*
                 * If sim stopped: edit the dynamic slot allocation
                 */
                for(int i = 0 ; i < clusterConfiguration.gNumberOfMinislots(); i++) {
                    sx = innerX + dynborder + (i * mtWidth * clusterConfiguration.gdMinislot());
                    node = cluster.getNodeForSlot(ChannelName.A,i+clusterConfiguration.gNumberOfStaticSlots());

                    /*
                     * Channel A (minislots)
                     */
                    Rectangle2D.Double slotRect;
                    if(channels.contains(ChannelName.A)) {
                    	
                        // Beschriftung Minislot Kanal A
                        g2d.setColor(Color.LIGHT_GRAY);
                        g2d.setFont(labelFont);
                        slotText = "MS "+(clusterConfiguration.gNumberOfStaticSlots()+i+1);
                        boundsLabel = getFont().getStringBounds(slotText, context);
                  	    slotTextX = sx + miniSlotWidth/2 - (int)boundsLabel.getWidth()/2; 
                  	    slotTextY = channelAY+channelAH/2+ (int)boundsLabel.getHeight()/2;
                        g2d.drawString(slotText, slotTextX, slotTextY);
                        
                        g2d.setStroke(lightDashStroke);
                        g2d.setColor(Color.LIGHT_GRAY);
                        g2d.drawRect(sx, channelAY, miniSlotWidth-1, channelAH-1);
                        g2d.setStroke(savedStroke);
                        slotRect = new Rectangle2D.Double(sx, channelAY, miniSlotWidth-1, channelAH);
                        if((mousePos != null) && slotRect.contains(new Point2D.Double(mousePos.x,mousePos.y)) ) {
                            highlightSlotRect = slotRect;
                            g2d.setComposite(transparentComposite);
                            g2d.setColor(slotHighlightInnerColor);
                            final int delta=1;
                            g2d.fillRect(sx-delta,channelAY-delta,miniSlotWidth+delta+delta,channelAH+delta+delta);
                            g2d.setColor(slotHighlightBorderColor);
                            g2d.setComposite(composite);
                            g2d.drawRect(sx-delta,channelAY-delta,miniSlotWidth+delta+delta-1,channelAH+delta+delta-1);
                        }
                    }

                    node = cluster.getNodeForSlot(ChannelName.B,i+clusterConfiguration.gNumberOfStaticSlots());
                    /*
                     * Channel B (minislots)
                     */
                    if( channels.contains(ChannelName.B)) {
                        
                        //   Beschriftung Minislot Kanal B
                        g2d.setColor(Color.LIGHT_GRAY);
                        g2d.setFont(labelFont);
                        slotText = "MS "+(clusterConfiguration.gNumberOfStaticSlots()+i+1);
                        boundsLabel = getFont().getStringBounds(slotText, context);
                  	    slotTextX = sx + miniSlotWidth/2 - (int)boundsLabel.getWidth()/2; 
                  	    slotTextY = channelBY+channelBH/2+(int)boundsLabel.getHeight()/2;
                        g2d.drawString(slotText, slotTextX, slotTextY);
                        
                        g2d.setStroke(lightDashStroke);
                        g2d.setColor(Color.LIGHT_GRAY);
                        g2d.drawRect(sx, channelBY, miniSlotWidth-1, channelBH-1);
                        g2d.setStroke(savedStroke);
                        slotRect = new Rectangle2D.Double(sx, channelBY, miniSlotWidth-1, channelBH);
                        if((mousePos != null) && slotRect.contains(new Point2D.Double(mousePos.x,mousePos.y)) ) {
                            highlightSlotRect = slotRect;
                            g2d.setComposite(transparentComposite);
                            g2d.setColor(slotHighlightInnerColor);
                            final int delta=1;
                            g2d.fillRect(sx-delta,channelBY-delta,miniSlotWidth+delta+delta,channelBH+delta+delta);
                            g2d.setColor(slotHighlightBorderColor);
                            g2d.setComposite(composite);
                            g2d.drawRect(sx-delta,channelBY-delta,miniSlotWidth+delta+delta-1,channelBH+delta+delta-1);
                        }
                    }

                } // for all minislots
                maxMiniSlotX = sx + miniSlotWidth-1;
            } else {
                /*
                 * If sim is running: show current communication
                 */
            	int slotA, slotB;
                int prevSlotA = -2;
                int prevSlotB = -2;
                FlexRayFrame frFrameA=null;
                FlexRayFrame frFrameB=null;
                int numDynSlotA =0;
                int numDynSlotB =0;
                String [] dynSlotLabelsA, dynSlotLabelsB;
                int [] dynSlotLabelsAX, dynSlotLabelsBX;
                int [] dynMiniSlotA, dynMiniSlotB;
                int lastMinislot, dynSlotTicks;
                
                dynSlotLabelsA = new String[clusterConfiguration.gNumberOfMinislots()];
                dynSlotLabelsB = new String[clusterConfiguration.gNumberOfMinislots()];
                dynSlotLabelsAX = new int[clusterConfiguration.gNumberOfMinislots()];
                dynSlotLabelsBX = new int[clusterConfiguration.gNumberOfMinislots()];
                dynMiniSlotA = new int[clusterConfiguration.gNumberOfMinislots()+1];
                dynMiniSlotB = new int[clusterConfiguration.gNumberOfMinislots()+1];
                
                for(int miniSlot = 0 ; miniSlot < clusterConfiguration.gNumberOfMinislots(); miniSlot++) {
                    sx = innerX + dynborder + (miniSlot * mtWidth * clusterConfiguration.gdMinislot());
                    int miniSlotMacroTick = miniSlot * clusterConfiguration.gdMinislot() +
                        clusterConfiguration.gdStaticSlot()*clusterConfiguration.gNumberOfStaticSlots();
                    if( miniSlotMacroTick < macroTickInCycle ) {

                        /*
                         * Channel A
                         */
                    	
                        slotA = cluster.getCurrentCommCycle().getSlotNumberFromMacroTick(ChannelName.A, miniSlotMacroTick);
                        node = cluster.getNodeForSlot(ChannelName.A, slotA);
                        if( (node != null)  && node.isTransmittingInSlot(slotA)) {
                            g2d.setColor(scenario.getNodeComponent(node).getNodeColor());

                            frFrameA = node.getFrameForSlot(ChannelName.A, slotA);
                            if ( (frFrameA == null ) || (!frFrameA.isEmptyFrame()) ) {  
                               g2d.fillRect(sx+1, channelAY+1, miniSlotWidth-2, channelAH-2);
                            } else {
                                g2d.setStroke(boldStroke);
                                g2d.drawRect(sx+2, channelAY+2, miniSlotWidth-1-3, channelAH-1-3);
                                g2d.setStroke(savedStroke);
                            }
                            
                            g2d.setColor(Color.BLACK);
                            g2d.drawRect(sx, channelAY, miniSlotWidth-1, channelAH-1);
                            if( slotA <= prevSlotA ) {
                                g2d.setColor(scenario.getNodeComponent(node).getNodeColor());
                                g2d.fillRect(sx-pad+1, channelAY+1, pad+1, channelAH-2);
                                g2d.setColor(Color.BLACK);
                                g2d.drawLine(sx-pad+1, channelAY, sx, channelAY);
                                g2d.drawLine(sx-pad+1, channelAY+channelAH-1, sx, channelAY+channelAH-1);
                            }
                            
                            //  Position fr Beschriftung dynamischer slot Kanal A bestimmen
                            // Wirklich geschrieben wird nach der for-Schleife, da sonst die
                            // Schrift mit farbigen Quadraten berdeckt wird.
                            if (!node.isTransmittingInSlot(prevSlotA)) {  
	    	                    dynSlotLabelsA[numDynSlotA] = "DS "+(frFrameA.getFrameID()+1);
	    	                    dynSlotLabelsAX[numDynSlotA] = sx;
	    	                    dynMiniSlotA[numDynSlotA] = miniSlot;
	    	                    numDynSlotA++;
                        	}
                            
                            // minislot markieren innerhalb des dynamic slots.
                            if ( (frFrameA == null ) || (!frFrameA.isEmptyFrame()) ) {
                                g2d.setStroke(lightDashStroke);
                                g2d.setColor(Color.BLACK);
                                g2d.drawRect(sx+1, channelAY+1, miniSlotWidth-3, channelAH-4);
                                g2d.setStroke(savedStroke);
                            }
                            
                            
                        }
                        else {
                        	//   Beschriftung leere dyn. slot Kanal A 
                            g2d.setColor(forgroundColor);
                            g2d.setFont(labelFont);
                            dynSlotLabelsA[numDynSlotA] = "DS "+(slotA+1);
    	                    dynSlotLabelsAX[numDynSlotA] = sx;
    	                    dynMiniSlotA[numDynSlotA] = miniSlot;
    	                    numDynSlotA++;
                            
                            g2d.setStroke(lightDashStroke);
                            g2d.drawRect(sx, channelAY, miniSlotWidth-1, channelAH-1);
                            g2d.setStroke(savedStroke);
                        }
                        prevSlotA = slotA;
                        
                        
                        /*
                         * Channel B
                         */
                        if( clusterConfiguration.gChannels().contains(ChannelName.B)) {
                            slotB = cluster.getCurrentCommCycle().getSlotNumberFromMacroTick(ChannelName.B, miniSlotMacroTick);
                            node = cluster.getNodeForSlot(ChannelName.B,slotB);
                            if( (node != null)  && node.isTransmittingInSlot(slotB)) {
                                g2d.setColor(scenario.getNodeComponent(node).getNodeColor());
                                frFrameB = node.getFrameForSlot(ChannelName.B, slotB);
                                if ( (frFrameB == null ) || (!frFrameB.isEmptyFrame()) ) {
                                    g2d.fillRect(sx+1, channelBY+1, miniSlotWidth-2, channelBH-2);
                                } else {
                                    g2d.setStroke(boldStroke);
                                    g2d.drawRect(sx+2, channelBY+2, miniSlotWidth-1-3, channelBH-1-3);
                                    g2d.setStroke(savedStroke);
                                }

                                g2d.setColor(Color.BLACK);
                                g2d.drawRect(sx, channelBY, miniSlotWidth-1, channelBH-1);
                                if( slotB <= prevSlotB ) {
                                    g2d.setColor(scenario.getNodeComponent(node).getNodeColor());
                                    g2d.fillRect(sx-pad+1, channelBY+1, pad+1, channelBH-2);
                                    g2d.setColor(Color.BLACK);
                                    g2d.drawLine(sx-pad+1, channelBY, sx, channelBY);
                                    g2d.drawLine(sx-pad+1, channelBY+channelAH-1, sx, channelBY+channelBH-1);
                                }
                                
                                // Position fr Beschriftung dynamischer slot Kanal B bestimmen
                                // Wirklich geschrieben wird nach der for-Schleife, da sonst die
                                // Schrift mit farbigen Quadraten berdeckt wird.
                                if (!node.isTransmittingInSlot(prevSlotB)) {  
    	    	                    dynSlotLabelsB[numDynSlotB] = "DS "+(frFrameB.getFrameID()+1);
    	    	                    dynSlotLabelsBX[numDynSlotB] = sx;
    	    	                    dynMiniSlotB[numDynSlotB] = miniSlot;
    	    	                    numDynSlotB++;
                            	}
                                
                                //    minislot markieren innerhalb des dynamic slots.
                                if ( (frFrameB == null ) || (!frFrameB.isEmptyFrame()) ) {
                                    g2d.setStroke(lightDashStroke);
                                    g2d.setColor(Color.BLACK);
                                    g2d.drawRect(sx+1, channelBY+1, miniSlotWidth-3, channelBH-4);
                                    g2d.setStroke(savedStroke);
                                }
                            }
                            else {
                            	//   Beschriftung leere dyn. slot Kanal B 
                                g2d.setColor(forgroundColor);
                                g2d.setFont(labelFont);
                                dynSlotLabelsB[numDynSlotB] = "DS "+(slotB+1);
	    	                    dynSlotLabelsBX[numDynSlotB] = sx;
	    	                    dynMiniSlotB[numDynSlotB] = miniSlot;
	    	                    numDynSlotB++;
                                
                                g2d.setStroke(lightDashStroke);
                                g2d.drawRect(sx, channelBY, miniSlotWidth-1, channelBH-1);
                                g2d.setStroke(savedStroke);
                            }
                            prevSlotB = slotB;
                        }
                    } else { // if (miniSlotMacroTick < macroTickInCycle)
                    	// Beschriftung minislot Kanal A und B
                        g2d.setColor(Color.LIGHT_GRAY);
                        g2d.setFont(labelFont);
                        slotText = "MS "+ (clusterConfiguration.gNumberOfStaticSlots()+ miniSlot+1);
                        boundsLabel = getFont().getStringBounds(slotText, context);
                  	    slotTextX = sx + miniSlotWidth/2 - (int)boundsLabel.getWidth()/2; 
                  	    slotTextY = channelAY+channelAH/2+ (int)boundsLabel.getHeight()/2;
                        g2d.drawString(slotText, slotTextX, slotTextY);
                        slotTextX = sx + miniSlotWidth/2 - (int)boundsLabel.getWidth()/2; 
                  	    slotTextY = channelBY+channelBH/2+ (int)boundsLabel.getHeight()/2;
                        g2d.drawString(slotText, slotTextX, slotTextY);
                        
                        g2d.setStroke(lightDashStroke);
                        g2d.setColor(Color.LIGHT_GRAY);
                        g2d.drawRect(sx, channelAY, miniSlotWidth-1, channelAH-1);
                        g2d.drawRect(sx, channelBY, miniSlotWidth-1, channelBH-1);
                        g2d.setStroke(savedStroke);
                    }  // if (miniSlotMacroTick < macroTickInCycle)
                    
                    
                } // for all minislots
                
                g2d.setColor(forgroundColor);
                g2d.setFont(labelFont);
                
                // Bestimme abgelaufene Macroticks...
                if (macroTickInCycle <= clusterConfiguration.gdStaticSlot()*clusterConfiguration.gNumberOfStaticSlots()+ clusterConfiguration.gdMinislot()*clusterConfiguration.gNumberOfMinislots())
                	// Zyklusende wurde noch nicht erreicht !!!
                	dynSlotTicks = macroTickInCycle-clusterConfiguration.gdStaticSlot()*clusterConfiguration.gNumberOfStaticSlots();
                else
                	// Zyklusende wurde noch  erreicht !!!
                	dynSlotTicks = clusterConfiguration.gdMinislot()*clusterConfiguration.gNumberOfMinislots();
                
                // Aus Anzahl der abgelaufenen Macroticks das aktuelle Minislot bestimmen ... 
                if ((dynSlotTicks % clusterConfiguration.gdMinislot()) > 0)
                	 // Fr Visualisierung auf nchste ganze Minislot aufgerunden. 
                	dynSlotTicks = dynSlotTicks + clusterConfiguration.gdMinislot()-(dynSlotTicks % clusterConfiguration.gdMinislot());
                lastMinislot = dynSlotTicks/clusterConfiguration.gdMinislot();
                
                //  Beschriftung dynamischer slot Kanal A
                if (numDynSlotA != 0) {
	                // ENDPOSITION DES LETZTEN SIMULIERTEN MINISLOT 
	                // Ende des letzten dynamischen Slot bestimmen.
                	if (frFrameA != null) {
		                if ((lastMinislot-dynMiniSlotA[numDynSlotA-1]) <= frFrameA.getCurrentFragmentLength())
		                	// dyn Slot ist noch nicht beendet
		                	dynMiniSlotA[numDynSlotA] = lastMinislot;
		                else
		                	// dyn Slot wurde beendet.
		                	dynMiniSlotA[numDynSlotA] = dynMiniSlotA[numDynSlotA-1]+frFrameA.getCurrentFragmentLength();
                	}
                	else
                		dynMiniSlotA[numDynSlotA] = lastMinislot;
                	
	                // FR ALLE DYN. SLOTS DIE BESCHRIFTUNG SCHREIBEN ... 
	                for (int i=0; i < numDynSlotA; i++) {
	                	boundsLabel = getFont().getStringBounds(dynSlotLabelsA[i], context);
	                    g2d.drawString(dynSlotLabelsA[i], dynSlotLabelsAX[i] + (dynMiniSlotA[i+1]-dynMiniSlotA[i])*miniSlotWidth/2 - (int)boundsLabel.getWidth()/2, 
	                    		channelAY+channelAH/2+ (int)boundsLabel.getHeight()/2);
	                }
                }
                
                // Beschriftung dynamischer slot Kanal B
                if (numDynSlotB != 0) {
	                // ENDPOSITION DES LETZTEN SIMULIERTEN MINISLOT 
	                // Ende des letzten dynamischen Slot bestimmen.
                	if (frFrameB != null) {
		                if ((lastMinislot-dynMiniSlotB[numDynSlotB-1]) <= frFrameB.getCurrentFragmentLength())
		                	// dyn Slot ist noch nicht beendet
		                	dynMiniSlotB[numDynSlotB] = lastMinislot;
		                else
		                	// dyn Slot wurde beendet.
		                	dynMiniSlotB[numDynSlotB] = dynMiniSlotB[numDynSlotB-1]+frFrameB.getCurrentFragmentLength();
	                }
                	else
                		dynMiniSlotB[numDynSlotB] = lastMinislot;
                	
	                // FR ALLE DYN. SLOTS DIE BESCHRIFTUNG SCHREIBEN ... 
	                for (int i=0; i < numDynSlotB; i++) {
	                	boundsLabel = getFont().getStringBounds(dynSlotLabelsB[i], context);
	                    g2d.drawString(dynSlotLabelsB[i], dynSlotLabelsBX[i] + (dynMiniSlotB[i+1]-dynMiniSlotB[i])*miniSlotWidth/2 - (int)boundsLabel.getWidth()/2, 
	                    		channelBY+channelBH/2+ (int)boundsLabel.getHeight()/2);
	                }
                }
                
            } // if (sim.stop()) {} else {}

            g2d.setStroke(savedStroke);

            /*
             * Draw the NIT
             */
            int curMTPixels = macroTickInCycle * mtWidth + innerX;
            if( curMTPixels >= nitborder )
                g2d.setColor( new Color(Color.LIGHT_GRAY.getRed()+10,
                                        Color.LIGHT_GRAY.getGreen()+10,
                                        Color.LIGHT_GRAY.getBlue()+10));
            else
                g2d.setColor( Color.LIGHT_GRAY );
            g2d.fillRect(innerX + nitborder,
                         innerY,
                         nitwidth,
                         slotsHeight);

            g2d.setFont(g2d.getFont().deriveFont(Font.BOLD));
            fm = g2d.getFontMetrics();
            Rectangle2D strBounds = fm.getStringBounds("NIT", g2d);
            int strwidth = (int)strBounds.getWidth() + 4;
            if( strwidth < nitwidth ) {
                int t = (nitwidth-strwidth)/2 + 2;
                g2d.setColor(Color.WHITE); // set base color white for XOR
                g2d.setXORMode(Color.BLACK); // XOR with black.
                g2d.drawString("NIT",
                               innerX + nitborder + t,
                               innerY + slotsHeight/2 + (int)strBounds.getHeight()/2 - pad );
                g2d.setPaintMode();
            }
        } // if cluster != null

        /*
         * Mark the beginnings of the various segments
         */
        g2d.setColor(Color.BLACK);
        Stroke savedStroke = g2d.getStroke();
        BasicStroke dashStroke = new BasicStroke(3,
                BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL,
                1, new float[]{ 10,4 }, 1 );
        g2d.setStroke(dashStroke);

        // dynamic segment border
        g2d.drawLine(innerX + dynborder,
                innerY - pad,
                innerX + dynborder,
                innerY + innerHeight - 1 + pad);


        g2d.setColor(Color.BLACK);
        // network idele time border
        g2d.drawLine(innerX + nitborder,
                innerY - pad,
                innerX + nitborder,
                innerY + innerHeight - 1 + pad);


        /*
         * Draw the macrotick marker
         */
        if( !sim.isStop() ) {
            g2d.setComposite(transparentComposite);

            g2d.setStroke(savedStroke);
            g2d.setColor(Color.RED);
            g2d.fillRect(innerX + slotStart + (macroTickInCycle-1)*mtWidth,
                         4,
                         mtWidth,
                         getHeight()-8);
            g2d.setColor(Color.BLACK);
            g2d.setComposite(composite);
//            g2d.setXORMode(Color.WHITE);
            g2d.drawRect(innerX + slotStart + (macroTickInCycle-1)*mtWidth,
                    4,
                    mtWidth - 1,
                    getHeight()-6);
            g2d.setPaintMode();
        } else {
            /*
             * "configuration" label
             */
            g2d.setStroke(savedStroke);
            g2d.setFont(savedFont);
            String configHint = res.getResourceByKey("commcycle.config");
            int center = configHint.length()/2;
            String leftHint = configHint.substring(0,center);
            String rightHint = configHint.substring(center);
            FontMetrics chFm = g2d.getFontMetrics();
            int space = (innerWidth-slotStart-pad);
            Rectangle2D chBounds = chFm.getStringBounds(configHint, g2d);
            while( space < chBounds.getWidth() ) {
                if( leftHint.length() > 1)
                    leftHint = leftHint.substring(0,leftHint.length()-1);
                if( rightHint.length() > 1)
                    rightHint = rightHint.substring(1);
                configHint = leftHint + "..." + rightHint;
                chBounds = chFm.getStringBounds(configHint, g2d);
                if(configHint.length() < 4) break;
            }

            int chY = innerY + innerHeight - fm.getLeading() - fm.getDescent() - 1;
            g2d.setColor(Color.WHITE);
            int x = slotStart + (space/2)-(int)chBounds.getCenterX();
            int y = (int)(innerY + innerHeight - chBounds.getHeight());
            int w = (int)chBounds.getWidth();
            int h = (int)chBounds.getHeight();
            g2d.fillRect(x,y,w, h);
            g2d.setColor(Color.BLACK);
            g2d.drawRect(x-1,y-1,w+1,h+1);
            g2d.drawString(configHint, x, chY);
        }

        g2d.setStroke(savedStroke);
        g2d.setColor(savedColor);


    } // paintComponent()
    
    /*
     * Methode:		setDialogFont
     * 
     * Description:	Setz alle Hintergrund- und Vordergrundfarben der GUI-Element auf die aktuellen
     * 				Werte. 
     *    
     * Parameters:   Aktuelle Font fr Labels und Textfelder.
     * 
     * Date: 		 23.04.2007
     * @author        sirocic
     */	
    private void setPanelFont() {
 	   
 	   labelFont = (Font) module.getConfiguration().getSettingByKey("FontLabel");

 	   setFont(labelFont);
    
    }
    
    /*
     * Methode:		setDialogTexte
     * 
     * Description:	Setz alle Text-Bausteine des Dialoges neu. Wichtig fr Sprachwechsel.
     *    
     * Parameters:   
     * 
     * Date: 		 08.02.2008
     * @author        sirocic
     */	
    private void setPanelTexte() {
    	String tooltip;
    	
    	if (learningModul.isTooltipsVisible()) {
         	configButton.setToolTipText(res.getResourceByKey("cycle.config.tip"));
    	 	sizeButton.setToolTipText(res.getResourceByKey("resize.tip"));
    	}
    	else {
    		configButton.setToolTipText(null);
    	 	sizeButton.setToolTipText(null);
    	}
        
    	staticSegmentString = res.getResourceByKey("flexray.staticsegment");
        dynamicSegmentString = res.getResourceByKey("flexray.dynamicsegment");
        
    }
    
    /*
     * Methode:		setDialogColors
     * 
     * Description:	Setz alle Hintergrund- und Vordergrundfarben der GUI-Element auf die aktuellen
     * 				Werte. 
     *    
     * Parameters:   Aktuelle Farbwerte fr Hintergrund und Vordergrund.
     * 
     * Date: 		 08.02.2008
     * @author        sirocic
     */	
    private void setPanelColors() {
        Color backgroundColor;
		 
 	   if (module.getConfiguration().getSettingByKey("ColorComponentBackground")!=null)
 	   {
 		   backgroundColor = (Color)module.getConfiguration().getSettingByKey("ColorComponentBackground");
 		   forgroundColor = (Color)module.getConfiguration().getSettingByKey("ColorComponentForeground");
 	   }
 	   else {
 		  backgroundColor = (Color)module.getConfiguration().getSettingByKey("ColorModuleBackground");
		   forgroundColor = (Color)module.getConfiguration().getSettingByKey("ColorModuleForeground");
 	   }

  	   setBackground(backgroundColor);
 	   setForeground(forgroundColor);
 	 
    }
    
    /*
     * Methode:		updateLanguageColorFonts
     *
     * Description:	Ldt sprachabhngige Teile neu. In der Toolbar knnen z.B. sprachabhngige
     * 				Buttons sein 
     *
     * Date: 		08.02.2008
     * 
     * @author sirocic
     */
    public void updateLanguageColorFonts()
    {
      
 	  //	 SUBKOMPONENTE UPDATEN...
 	  
 	  // 	 FARBWECHSEL ...
 	   setPanelColors();

 	   // FONTWECHSEL
 	   setPanelFont();
 	  
 	   // SPRACHWECHSEL ...
 	   setPanelTexte();

    }
    
    /*
     * Methode:		fitPanel
     *
     * Description:	Setzt die Ausmasse des Kommunikationszyklus-Panels und alle relevanten Grssen,
     * 				wie z.B. MTwidth usw..
     *
     * Date: 		11.02.2008
     * 
     * @author sirocic
     */
    public void fitPanel()
    {
    	String slotText;
    	Rectangle2D boundsLabel;
  	    FontRenderContext context;
    	
  	    Graphics2D g2d = (Graphics2D)this.getGraphics();
  	    context = g2d.getFontRenderContext();
      
    	//    	 Calculate the height of the slot display
        slotsHeight = innerHeight-scaleHeight;

        /*
         * The static / dynamic segment border
         */
        dynborder = clusterConfiguration.gNumberOfStaticSlots() *
                        clusterConfiguration.gdStaticSlot() *
                        mtWidth + slotStart;

        nitborder = dynborder +
                        clusterConfiguration.gNumberOfMinislots() *
                        clusterConfiguration.gdMinislot() *
                        mtWidth;

        staticSegmentWidth = dynborder;
        dynamicSegmentWidth = nitborder - dynborder;
        nitwidth = clusterConfiguration.gdNIT() * mtWidth - slot_padding;
        
        
        staticSlotWidth = mtWidth * clusterConfiguration.gdStaticSlot() - slot_padding;
        // prfen, ob die Beschriftung der statischen Slots in das entsprechende Kstchen passen..
        // wenn nicht muss die Weite der macroTicks entsprechend grsser gemacht werden.
        slotText = "SS "+clusterConfiguration.gNumberOfStaticSlots();
        boundsLabel = getFont().getStringBounds(slotText, context);
        if ((int)boundsLabel.getWidth() > staticSlotWidth) 
        	mtWidth = ((int)boundsLabel.getWidth() + 3*slot_padding)/clusterConfiguration.gdStaticSlot();
        
        miniSlotWidth = mtWidth*clusterConfiguration.gdMinislot() - slot_padding;
        // prfen, ob die Beschriftung der MiniSlots in das entsprechende Kstchen passen..
        // wenn nicht muss die Weite der macroTicks entsprechend grsser gemacht werden.
        slotText = "MS "+clusterConfiguration.gNumberOfMinislots();
        boundsLabel = getFont().getStringBounds(slotText, context);
        if ((int)boundsLabel.getWidth() > miniSlotWidth) 
        	mtWidth = ((int)boundsLabel.getWidth() + 3* slot_padding)/clusterConfiguration.gdMinislot();
        
    }

    /**
     * Override setLocale for changing the texts if required.
     *
     * @param l
     * @see java.awt.Component#setLocale(java.util.Locale)
     */
    @Override
    public void setLocale(Locale l) {
    	String tooltip;
    	
        super.setLocale(l);
        staticSegmentString = res.getResourceByKey("flexray.staticsegment");
        dynamicSegmentString = res.getResourceByKey("flexray.dynamicsegment");
        resized = true;
        if (learningModul.isTooltipsVisible()) {
	        sizeButton.setToolTipText(res.getResourceByKey("resize.tip"));
	        configButton.setToolTipText(res.getResourceByKey("cycle.config.tip")); 
        }
        else {
        	sizeButton.setToolTipText(null);
	        configButton.setToolTipText(null); 
        }
    }

    /*
     * non-javadoc
     *
     * @see java.awt.Component#setLocation(int, int)
     */
    @Override
    public void setLocation(int x, int y) {
        int oldx = this.getX();
        int oldy = this.getY();
        if( (oldx != x) || (oldy != y)) {
            super.setLocation(x,y);
            firePropertyChange(Property.LOCATION, new Point(oldx,oldy), new Point(x,y));
        }
    }

    private Boolean synch = Boolean.FALSE;
    /*
     * non-javadoc
     * @return the preferred size
     * @see javax.swing.JComponent#getPreferredSize()
     */
    @Override
    public Dimension getPreferredSize() {
        synchronized (synch) {
            int pprefWidth = prefWidth;
            prefWidth = pad + slotStart +
            (mtWidth * clusterConfiguration.gMacroPerCycle()) +
            pad + +autoMoveButtonSize + shadowSize;
            if( prefWidth < minWidth ) prefWidth = pprefWidth;
            /*
             * Recalculate the dimension on first creation
             */
            if( synch.equals(Boolean.FALSE) && getParent() instanceof JComponent ) {
                synch = Boolean.TRUE;
                JComponent parent = (JComponent)getParent();
                int parentVisibleWidth = parent.getVisibleRect().width;
                if(parentVisibleWidth > 0) {
                    int adjX = -1;
                    while( (prefWidth-10) > parentVisibleWidth ) {
                        adjustWidthBy(adjX);
                        adjX--;
                    }
                }
            }
        }
        return new Dimension( prefWidth, prefHeight );
    } // getPreferredSize

    /*
     * non-javadoc
     * @see java.awt.Component#setSize(int, int)
     */
    @Override
    public void setSize(int width, int height) {
        super.setSize(width, height);
        innerWidth = getPreferredSize().width-shadowSize-pad-pad;
        innerHeight = getPreferredSize().height-autoMoveButtonSize-shadowSize-pad-pad;
        scaleHeight = innerHeight/5;
        sizeButton.setLocation(prefWidth-shadowSize-autoMoveButtonSize, 0);
        titleRect.setRect(autoMoveButtonSize,
                1,
                prefWidth-shadowSize-autoMoveButtonSize-autoMoveButtonSize,
                autoMoveButtonSize-1);
        resized = true;
    } // setSize

    /*
     * non-javadoc
     * @param e
     * @see javax.swing.event.ChangeListener#stateChanged(javax.swing.event.ChangeEvent)
     */
    public void stateChanged(ChangeEvent e) {
        this.clusterConfiguration = cluster.getClusterConfiguration();
    }

    /*
     * non-javadoc
     * @param e
     * @see java.awt.event.ComponentListener#componentMoved(java.awt.event.ComponentEvent)
     */
    public void componentMoved(ComponentEvent e) {
        if( autoMove ) {
            adjustLocation();
        }
    }

    /*
     * non-javadoc
     * @param e
     * @see java.awt.event.ComponentListener#componentHidden(java.awt.event.ComponentEvent)
     */
    public void componentHidden(ComponentEvent e) {
        if( autoMove ) {
            adjustLocation();
        }
    }

    /*
     * non-javadoc
     * @param e
     * @see java.awt.event.ComponentListener#componentResized(java.awt.event.ComponentEvent)
     */
    public void componentResized(ComponentEvent e) {
        if( autoMove ) {
            adjustLocation();
        }
        configButton.setLocation(new Point(locationPad,getHeight()-5-configButton.getHeight()-locationPad));
    }

    /*
     * non-javadoc
     * @param e
     * @see java.awt.event.ComponentListener#componentShown(java.awt.event.ComponentEvent)
     */
    public void componentShown(ComponentEvent e) {
        if( autoMove ) {
            adjustLocation();
        }
    }


    /**
     * Adjust the location of this component according to the lower left corner
     * of the parent.
     *
     */
    public void adjustLocation() {
        if( getParent() == null ) return;
        Rectangle visibleRect = ((JComponent)getParent()).getVisibleRect();
        int posX = visibleRect.x + 1;
        int posY = visibleRect.y + visibleRect.height - getPreferredSize().height;
        setLocation(posX, posY);
    }

    /**
     * automatic positioning
     * @return true if automatic positioning is enabled, false otherwise
     */
    public boolean isAutoMove() {
        return autoMove;
    }

    /**
     * Change automatic moving status
     * @param b
     */
    public void setAutoMove(boolean b) {
        boolean old = this.autoMove;
        if( old == b ) return;
        if( (old == true) && (prevLocation == null) ) prevLocation = getLocation();
        this.autoMove = b;
        if( autoMove ) {
            prevLocation = getLocation();
            removeMouseMotionListener(componentDragger);
            adjustLocation();
        } else {
            //addMouseListener(componentDragger);
            addMouseMotionListener(componentDragger);
            if( (prevLocation != null) && (getParent() != null)) {
                if( getParent().contains(prevLocation) && !prevLocation.equals(getLocation()) ){
                    setLocation(prevLocation);
                } else {
                    setLocation(getX()+1, getY()-1);
                }
            }
        }
        firePropertyChange(Property.AUTOMOVE, Boolean.valueOf(old), Boolean.valueOf(b));
    } // setAutoMove

    /**
     * @param x how much should the width be adjusted?
     * @return true if the adjustment is actually in effect
     */
    private boolean adjustWidthBy(int x) {
        /*
         * Width calculation is more complicated because the width depends
         * on the communication cycle configuration.
         */

        int tInnerWidth = prefWidth + x - pad - slotStart - pad - shadowSize - autoMoveButtonSize;
        double macroTickWidth = (double)tInnerWidth / (double)(clusterConfiguration.gMacroPerCycle());
        int iMacroTickWidth = (int)macroTickWidth;

        if( x < 0 ) {
            iMacroTickWidth = (int)Math.ceil(macroTickWidth);
        } else {
            iMacroTickWidth = (int)Math.floor(macroTickWidth);
        }

        if( iMacroTickWidth == mtWidth ) return false;
        if( iMacroTickWidth < minMTWidth ) return false;


        tInnerWidth = iMacroTickWidth * clusterConfiguration.gMacroPerCycle();

        int newPrefW = pad + slotStart + tInnerWidth + pad + shadowSize;
        int newX = newPrefW - prefWidth;

        if(newPrefW > minWidth) {
            prefWidth += newX;
            setSize(prefWidth,prefHeight);
            mtWidth = iMacroTickWidth;
            return true;
        }
        return false;
    } // changeWidth

    public static String recode(String s) {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < s.length(); i++) {
            char c = s.charAt(i);
            if       (c >= 'a' && c <= 'm') c += 13;
            else if  (c >= 'n' && c <= 'z') c -= 13;
            else if  (c >= 'A' && c <= 'M') c += 13;
            else if  (c >= 'A' && c <= 'Z') c -= 13;
            sb.append(c);
        }
        return sb.toString();
    }
    /*
     *  -------------- mousemotionlistener -----------------
     */

    /**
     * When the mouse got dragged on the small resize box, the size of the
     * Communication cycle component has to be recalculated.
     *
     * @param e the MouseEvent
     */
    public void mouseDragged(MouseEvent e) {
        /*
         * Source of the event is the resize-Button
         */
        if( e.getSource().equals(this.sizeButton)) {
            getParent().setComponentZOrder(this, 0);
            // roughly set the center of the button as the origin of the resize
            final int s = (autoMoveButtonSize / 3);
            int x = e.getX() - s;
            int y = e.getY() - s;

            // check if the new height is acceptable and set it if so.
            if((prefHeight - y) > minHeight) {
                prefHeight -= y;
                setSize(prefWidth,prefHeight);
                // in autoMove mode, recalculate the location
                // otherwise set it to reflect the mouse movement
                if( autoMove )
                    adjustLocation();
                else {
                    setLocation(getX(),getY() + y);
                }
            }
            adjustWidthBy(x);
        }
        configButton.setLocation(new Point(locationPad,getHeight()-5-configButton.getHeight()-locationPad));
    } // mouseDragged



    /**
     *
     * @param e
     * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
     */
    public void mouseClicked(MouseEvent e) {
        switch( e.getButton() ) {
        case MouseEvent.BUTTON1: // left
            if(scenario.isEditable()) mouseLeftClicked(e);
            break;
        case MouseEvent.BUTTON2: // mid
            break;
        case MouseEvent.BUTTON3: // right
            if(scenario.isEditable()) mouseRightClicked(e);
            break;
        }
    } // mouseClicked

    /**
     * Code for the left click somewhere on this component
     * @param e
     */
    public void mouseLeftClicked(MouseEvent e) {
        int x = e.getX();
        int y = e.getY();
        ChannelName channel = ChannelName.A;
        SlotDialog d;
        String errorMsg, errorTitle;


        /*
         * click on static slot
         */
        if( (y > channelAY) && (y < channelAY+slotsHeight) &&
            (x > (slotStart + innerX)) && (x < maxStaticSlotX)
          ) {
            if( (y >= channelBY) && (channelBY > 0) ) channel = ChannelName.B;
            else channel = ChannelName.A;
            int slot = (x - (innerX + slotStart)) / (staticSlotWidth + (pad/2));

            // display slot config dialog for slot "slot"
            d = new SlotDialog(scenario, channel, slot);
            d.setLocation(posForClick(d,e));
            d.showDialog(true);
        } // click on static slot

        /*
         * click on dynamic slot
         */
        /* BS rausgenommen, da die Zuordnung des Dynamischen Slots im
         * Knotenkonfig-Dialog erfolgt. es erscheint eine Information,
         * die dies erklrt.
         * */
          if( (y > channelAY) && (y < channelAY+slotsHeight) &&
                (x > (innerX+dynborder)) && (x < maxMiniSlotX)
              ) {
        	  module.getFrame().showInfoMessageDialog("dlg.errortext.18");
        	  
            } // click on dynamic slot


    } // mouseLeftClicked

    /**
     * right click of the mouse
     * @param e the mouse event
     */
    public void mouseRightClicked(MouseEvent e) {
        scenario.getLearningModule().getSimulator().stopSimulation();
        CommCycleConfigDialog d = new CommCycleConfigDialog(scenario);
        d.showDialog(true);
        
    } // mouseRightClicked

    /**
     * Calculate an absolute screen position for displaying a Dialog
     * near the mouse position of a click on this component but not out
     * of bounds of the scenarioPanel.
     * @param d The dialog for which the position to calculate
     * @param e the mouse event (to get mouse position)
     * @return a point containing the position.
     */
    private Point posForClick(JDialog d, MouseEvent e) {
        int dw = d.getWidth();
        int dh = d.getHeight();
        FrameFR frame = scenario.getFrame();

        // horizontal location
        int locX = frame.getX() + getX() + e.getX() -
            scenario.getScenarioPanel().getVisibleRect().x - dw/2;
        if( (locX + dw) > (frame.getX()+frame.getWidth()) ) {
            locX -= ((locX+dw)-(frame.getX()+frame.getWidth()));
        }
        if( locX < frame.getX()) locX=frame.getX();

        // vertical location
        int locY = frame.getY() + getY() + e.getY() -
            scenario.getScenarioPanel().getVisibleRect().y;
        if( (locY + dh) > (frame.getY()+frame.getHeight()) ) {
            locY -= ((locY+dh)-(frame.getY()+frame.getHeight()));
        }
        if( locY < frame.getY()) locY = frame.getY();

        return new Point(locX,locY);
    }

    public void mouseEntered(MouseEvent e) {
    	
        
        if( scenario.isEditable())
            mousePos = new Point(e.getPoint());
        if( !autoMove ) {
            if( (!inMoveZone) && titleRect.contains(e.getPoint())) {
                setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
                inMoveZone = true;
            }
        }
        
        
    }

    private boolean inMoveZone = false;
    public void mouseMoved(MouseEvent e) {
    	String tooltip;
    	int x = e.getX();
        int y = e.getY();
        
        if(mousePos == null) return;
        if( !autoMove ) {
            if( (!inMoveZone) && titleRect.contains(e.getPoint())) {
                setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
                inMoveZone = true;
            }
        }
        if(  inMoveZone && (!titleRect.contains(e.getPoint()))) {
            setCursor(Cursor.getDefaultCursor());
            inMoveZone = false;
        }
        if (learningModul.isTooltipsVisible()) {
	        if ( (y > channelAY) && (y < channelBY+slotsHeight)) {
	        	// Tooltip fr den Kommunikationszyklus setzen
	            tooltip = "<html><br>";
	            // Statische Segment
	            if ((x > (slotStart + innerX)) && (x < maxStaticSlotX) ) {
	            	tooltip=tooltip+res.getResourceByKey("flexray.statsegment.tip")+"<br><br>";
	            	tooltip=tooltip+res.getResourceByKey("flexray.statslot.tip");
	            }
	            if((x > (innerX+dynborder)) && (x < maxMiniSlotX) ) {
	            	tooltip=tooltip+res.getResourceByKey("flexray.dynsegment.tip")+"<br><br>";
	            	tooltip=tooltip+res.getResourceByKey("flexray.dynslot.tip");
	            }
	            if((x > (maxMiniSlotX)) ) {
	            	tooltip=tooltip+res.getResourceByKey("flexray.nit.tip");
	            }
	        	tooltip=tooltip+"</html>";
	        	setToolTipText(tooltip);
	        }
        }
        else {
        	setToolTipText(null);
        }
  
        if ( (y > channelAY) && (y < channelBY+slotsHeight) &&
        		((x > (slotStart + innerX)) && (x < maxStaticSlotX) ) ) {
	        synchronized(highlightSlotRect) {
	            if (!highlightSlotRect.contains(e.getX(),e.getY())) highlightSlotRect = emptySlotRect;
	            if( (highlightSlotRect == emptySlotRect) ) {
	                mousePos.setLocation(e.getPoint());
	                repaint(); // FIXME: could be done better
	            }
	        }
        }

    }
    public void mouseExited(MouseEvent e) {
        prevSlot = new int[]{-1,-1};
        mousePos = null;
        
        if(  inMoveZone && (!titleRect.contains(e.getPoint()))) {
            setCursor(Cursor.getDefaultCursor());
            inMoveZone = false;
        }
    }

    public void mousePressed(MouseEvent e) {/**/}
    public void mouseReleased(MouseEvent e) {/**/}

    /**
     * put this component's attributes into an XML element
     *
     * @param xmlDoc
     * @return XML Element storing this component's data
     */
    public Element toXML(final Document xmlDoc) {
        Element myXML = xmlDoc.createElement(XT_COMMCYCLEVIEW);
        myXML.setAttribute(XT_COMMCYLCEVIEW_AUTO,
                Boolean.toString(autoMove));
        Element screenpos = xmlDoc.createElement(XT_SCREENPOS);
        Integer posX = Integer.valueOf(((autoMove&&prevLocation!=null)?prevLocation.x:getX()));
        Integer posY = Integer.valueOf(((autoMove&&prevLocation!=null)?prevLocation.y:getY()));
        screenpos.setAttribute(XT_SCREENPOS_X, posX.toString());
        screenpos.setAttribute(XT_SCREENPOS_Y, posY.toString());
        screenpos.setAttribute(XT_SCREENPOS_Z,
                Integer.valueOf(getParent().getComponentZOrder(this)).toString());
        myXML.appendChild(screenpos);
        Element dimension = xmlDoc.createElement(XT_DIMENSION);
        dimension.setAttribute(XT_DIMENSION_W, Integer.valueOf(getWidth()).toString());
        dimension.setAttribute(XT_DIMENSION_H, Integer.valueOf(getHeight()).toString());
        myXML.appendChild(dimension);
        return myXML;
    }

    /**
     * restore Component parameters from XML
     * @param element
     * @return the stored z-order value of the component
     */
    public int fromXML(final Element element) {
        int z = 0;
        boolean storedAutoMove = Boolean.valueOf(element.getAttribute(XT_COMMCYLCEVIEW_AUTO)).booleanValue();
        setAutoMove( storedAutoMove );
        autoMoveButton.setSelected( autoMove );
        NodeList nodeList = element.getChildNodes();
        for( int i = 0 ; i < nodeList.getLength() ; i++ ) {
        	if (nodeList.item(i) instanceof Element) {
	            Element item = (Element)nodeList.item(i);
	            if( item.getNodeName().equals(XT_SCREENPOS)) {
	                int x = Integer.parseInt(item.getAttribute(XT_SCREENPOS_X));
	                int y = Integer.parseInt(item.getAttribute(XT_SCREENPOS_Y));
	                z = Integer.parseInt(item.getAttribute(XT_SCREENPOS_Z));
	                if( autoMove ) {
	                    prevLocation = new Point(x,y);
	                } else {
	                    setLocation(x, y);
	                }
	            }
	            if( item.getNodeName().equals(XT_DIMENSION)) {
	                prefWidth = Integer.parseInt(item.getAttribute(XT_DIMENSION_W));
	                prefHeight = Integer.parseInt(item.getAttribute(XT_DIMENSION_H));
	                adjustWidthBy(0); // get the dirty work done
	            }
        	}
        }
        return z;
    } // restoreFromXML

    public static void setSlotHighlightInnerColor(Color slotHighlightInnerColor) {
        CommCycleComponent.slotHighlightInnerColor = slotHighlightInnerColor;
    }

    public Rectangle getDragZone() {
        return titleRect.getBounds();
    }

}
