/*
 * $Id: ComponentDragger.java,v 1.1 2009/04/01 13:57:09 sirocic Exp $
 */
package flexray.gui;

import java.awt.Component;
import java.awt.KeyboardFocusManager;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.util.Vector;

import javax.swing.JPanel;


/**
 * <code>ComponentDragger</code> implements generic Swing component dragging.<br>
 * If the <code>ComponentDragger</code> is used as <code>MouseMotionListener</code> only,
 * the dragging hotspot (where you "grab" the component" will change to the upper left corner
 * of the component (moving the component to accomplish this.<br>
 * If the <code>ComponentDragger</code> is also used as <code>MouseListener</code>,
 * the dragging hotspot will remain the position where you clicked to start dragging.<br>
 * By default, the movement respects the bounds of the component's parent container. If the mouse
 * movement would drag the component outside the container, the component will not be moved further.
 * <br>
 * This Dragger knows about AlphaBlendingComponent.<br>
 * This Dragger knows about DragZone and respects it.<br>
 *
 * <br>
 * Known issues:<br>
 * <ul>
 * <li>There is some random jumping of the dragged component if it is dragged over more than one
 * component because the jump starts in the middle of every single dragged over component.
 * If there are, for example, two components beneath the dragged component, then there are
 * two locations for making the decision to which direction the dragged component should jump.
 * After that jump, the effective decision position changes and so the decision.</li>
 * <li>components that are too big to fit in the bounds of their parent horizontally cannot be
 * dragged over other components vertically and vice versa. Just drag the other components.</li>
 * <li>In some cases</li>
 * </ul>
 *
 * @author Dirk Frsterling
 */
public class ComponentDragger extends ComponentMouseAdapter {

    /*
     * are we in an dragging process?
     */
    private boolean inDrag = false;
    /*
     * relative drag start coordinates
     */
    private int dragX = 0;
    private int dragY = 0;

    /*
     * store a component tat we bumped into as well as the
     * dragging direction that was active at that time
     * (to avoid too dumb jumps of dragged components)
     */
    private DragDirection initialDirection = null;
    private Component initialModule = null;

    // may dragged components overlap other components in the same container?
    private boolean overlapAllowed = true;
    // is movement restricted to the parent's bounds
    private boolean restrictMovement = false;

    private boolean wasInDragZone = false;
    
    private Component findDestComponent = null;
    private int prevY = -1;
    private int prevX = -1;

    private Vector<Component> ignoreComponents = new Vector<Component>();
    {
        ignoreComponents.add( null );
    }

    private int componentZOrder = 0;

    /**
     * Enumerate the basic dragging directions left, right, up
     * and down and provide some means to check if a direction
     * is horizonzal or vertical.
     *
     * @author Dirk Frsterling
     */
    private enum DragDirection {
        LEFT,
        RIGHT,
        UP,
        DOWN,
        NONE;

        /**
         * check if the direction is vertical. Please note that this
         * is NOT the same as !isHorizontal() because of the NONE direction
         * which is neither horizontal nor vertical.
         *
         * @return true if the direction is vertical, false otherwise.
         */
        public boolean isVertical() {
            if ( this.equals(UP) || this.equals(DOWN)) return true;
            return false;
        }

        /**
         * check if the direction is horizontal. Please note that this
         * is NOT the same as !isVertical() because of the NONE direction
         * which is neither horizontal nor vertical.
         *
         * @return true if the direction is horizontal, false otherwise.
         */
        public boolean isHorizontal() {
            if ( this.equals(LEFT) || this.equals(RIGHT)) return true;
            return false;
        }
    } // DragDirection

    /**
     * default constructor
     */
    public ComponentDragger() {
        super();
    }


    /**
     * record potential starting positions of the dragging process.
     * Right now, any mouse button press will record coordinates.
     * The component will also be brought to the foreground if overlapping
     * is allowed.
     *
     * @param e the MouseEvent to process
     */
    @Override
    public void mousePressed( MouseEvent e ) {
        if( e.getButton() != MouseEvent.BUTTON1 ) return;
        if( !inDrag && ((e.getModifiers() & InputEvent.CTRL_MASK) == 0)) {
            Component c = e.getComponent();
            if( c instanceof DragZone ) {
                if( !((DragZone)c).getDragZone().contains(e.getPoint()) ) return;
                wasInDragZone = true;
            }
            dragX = e.getX();
            dragY = e.getY();
            if( overlapAllowed ) {
                Component parent = c.getParent();
                if( parent instanceof JPanel ) {
                    JPanel p = (JPanel)parent;
                    if( p.getComponentZOrder(c) != 0 ) {
                        p.setComponentZOrder( c, 0 );
                        c.repaint();
                    }
                }
            }
        } // if !inDrag
    } // mousePressed()

    /**
     * Re-set dragging on mouse release and correct focus: Requesting focus
     * on the dragged component (because it was surely clicked for dragging)
     * and resetting keyboard focus in case there is exactly one keyboard
     * focusable component and it or its container was just moved.
     *
     * @param e the MouseEvent to process
     */
    @Override
    public void mouseReleased(MouseEvent e) {
        if( e.getButton() != MouseEvent.BUTTON1 ) return;
        Component c = e.getComponent();
        if( c instanceof DragZone ) {
            if( !((DragZone)c).getDragZone().contains(e.getPoint()) ) return;
            wasInDragZone = false;
        }
        if( inDrag ) {
            inDrag = false;
            initialDirection = null;
            Component parent = c.getParent();
            if( parent instanceof JPanel ) {
                JPanel p = (JPanel)parent;
                p.setComponentZOrder( c, componentZOrder );
            }
            KeyboardFocusManager.getCurrentKeyboardFocusManager().clearGlobalFocusOwner();
            if( !e.getComponent().isFocusOwner() ) {
                e.getComponent().requestFocus();
            }
            if( c instanceof Resizable ) reAddOtherListeners( c );
        }
        if( c instanceof AlphaBlendingComponent ) {
            if( ((AlphaBlendingComponent)c).isBlendOnDrag() ) {
                ((AlphaBlendingComponent)c).blendToLight( false );
                c.repaint();
            }
        }
    } // mouseReleased

    /**
     * If the mouse is dragged, immediately update the component's
     * location to reflect the drag. Ensure that the component
     * remains completely visible within its comtainers bounds.
     *
     * @param e the MouseEvent to process
     */
    @Override
    public void mouseDragged(MouseEvent e) {
        if ((e.getModifiers() & InputEvent.CTRL_MASK) != 0) return;
        if( (e.getModifiers() & InputEvent.BUTTON1_MASK) == 0  ) return;
        // If not already ongoing, dragging has definitively started now.
        Component c = e.getComponent();
        if( (c instanceof DragZone) && !wasInDragZone ) return;
        if( (c instanceof Resizable) && !inDrag ) {
            if( ( e.getX()> (c.getWidth()  - ((Resizable)c).getSensitiveX())) &&
                ( e.getY()> (c.getHeight() - ((Resizable)c).getSensitiveY())) ) {
                return;
            }
        }
        if( !inDrag ) {
            inDrag = true;
            removeOtherListeners( c );
            // bring the component to the foreground
            c.getParent().setComponentZOrder(c, 0);
            if( c instanceof AlphaBlendingComponent ) {
                if( ((AlphaBlendingComponent)c).isBlendOnDrag() ) {
                    ((AlphaBlendingComponent)c).blendToLight( true );
                    c.repaint();
                }
            }
        }



        int oldX = c.getX();
        int oldY = c.getY();
        int width = c.getWidth();
        int height = c.getHeight();


        // calculate the component's new location:
        // the dragging distance plus the current position of the component.
        int newX = e.getX() - dragX + oldX;
        int newY = e.getY() - dragY + oldY;

        if( restrictMovement ) {
            // get the parent's bounds
            Rectangle parBounds = c.getParent().getBounds();

            // ensure the upper/left boundary
            if( newY < parBounds.y ) newY = parBounds.y;
            if( newX < parBounds.x ) newX = parBounds.x;

            // ensure the lower/right boundary
            if( (newX + width) >  parBounds.width )
                newX = parBounds.width - c.getWidth();
            if( (newY + height) >  parBounds.height)
                newY = parBounds.height - c.getHeight();
        }

        int diffX = newX - oldX;
        int diffY = newY - oldY;


        if( !overlapAllowed ) {

            /*
             * exclude the dragged component and its parent from the
             * collision check.
             */
            Component parent = c.getParent();
            ignoreComponent( c );
            ignoreComponent( parent );

            if( parent instanceof JPanel ) {
                JPanel p = (JPanel)parent;
                componentZOrder = p.getComponentZOrder( c );
                p.setComponentZOrder( c, 0 );
            }

            /* one attempt at overlapAllowed prevention. This doesn't work 100% yet.
             * If the mouse pointer moved far enough into a collisioned component,
             * the dragged component should just jump over it and possibly the
             * following components in the same direction.
             */
            // flag: was the location set by "findDestination"
            boolean setByFind = false;
            // dragged component doesn't overlap (stands free)
            boolean standingFree = true;
            if( diffX < 0 ) { // drag to the left
                // find out if the new location is problematic
                for( int x = oldX; ( x >= newX ) && !setByFind; x--) {
                    for( int y = oldY; y <= (oldY + height); y++ ) {
                        Component comp = parent.getComponentAt( x, y );
                        if( !isIgnoredComponent( comp ) ) {
                            standingFree = false; // we hit a component
                            /*
                             *  Now, if we're not already dragging over another component,
                             *  let's now record which component we ran into and that we were
                             *  dragging to the LEFT. This information is used to prevent
                             *  unnecessary UP/DOWN jumps if the dragging changes direction
                             *  while the mouse is over another component.
                             *  Update the initial direction also if the mouse, dragging one
                             *  component, is moved from within a second onto a third component,
                             *  changing dragging direction
                             */
                            if ( (initialDirection == null) || (!comp.equals(initialModule)) ) {
                                initialDirection = DragDirection.LEFT;
                                initialModule = comp;
                            }
                            /*
                             * Leave the dragged component where it is until the mouse moved far
                             * enough into the component we bumped into. This gives the user some
                             * "grip" to it.
                             */
                            // jump horizontally only if we approached horizontally at this component.
                            if( initialDirection.isHorizontal() ) {
                                // mouse has to be moved half over the other component before the dragged
                                // component jumps over it.
                                int inComponent = comp.getX() + (comp.getWidth() / 2) - oldX;
                                if( e.getX() >= inComponent ) {
                                    // reset the x position of the dragged component to the outer edge of the
                                    // component over which we are dragging
                                    newX = comp.getX() + comp.getWidth();
                                } else {
                                    // Mouse moved far enough. Now calculate a location for the component
                                    // according to the dragging direction
                                    int testX = findDestination( c, newX, newY, DragDirection.LEFT ).x;
                                    // correction to reduce jumping in case the mouse did not move
                                    // far enough (because of an obstacle): Only accept the new position
                                    // if it's at least half the way from start to end.
                                    int midX = testX + (oldX-testX)/2;
                                    if( newX < midX ) {
                                        newX = testX;
                                    } else {
                                        newX = oldX;
                                    }
                                    setByFind = true; // stop further checking. We moved.
                                }
                            }
                            // we already found components in the way. No further test needed.
                            break;
                        }
                    }
                }
            } else if ( diffX > 0 ) { // drag to the right. Works just like drag to the left
                for( int x = oldX + width; (x < newX + width) && !setByFind; x++ ) {
                    for( int y = oldY ; y < oldY + height ; y++ ) {
                        Component comp = parent.getComponentAt( x, y );
                        if( !isIgnoredComponent( comp ) ){
                            standingFree = false;
                            if ( (initialDirection == null) || (!comp.equals(initialModule)) ) {
                                initialDirection = DragDirection.RIGHT;
                                initialModule = comp;
                            }
                            if( initialDirection.isHorizontal() ) {
                                int inComponent = comp.getX() + (comp.getWidth() / 2) - oldX;
                                if( e.getX() <= inComponent ) {
                                    newX = x - width;
                                } else {
                                    int testX = findDestination( c, newX, newY, DragDirection.RIGHT ).x;
                                    int midX = oldX + (testX - oldX) / 2;
                                    if( newX > midX ) {
                                        newX = testX;
                                    } else {
                                        newX = oldX;
                                    }
                                    setByFind = true;
                                }
                            }
                            break;
                        }
                    }
                }
            } // drag to left/right

            setByFind = false;
            // vertical -  works just like horizontal
            if( diffY < 0 ) { // drag upwards
                for( int y = newY; (y < oldY) && !setByFind; y++ ) {
                    for( int x = newX ; x < newX + width ; x++ ) {
                        Component comp = parent.getComponentAt( x, y );
                        if( !isIgnoredComponent( comp ) ){
                            standingFree = false;
                            if ( (initialDirection == null) || (!comp.equals(initialModule)) ) {
                                initialDirection = DragDirection.UP;
                                initialModule = comp;
                            }
                            if( initialDirection.isVertical() ) {
                                int inComponent = comp.getY() + (comp.getHeight() / 2) - oldY;
                                if( e.getY() >= inComponent ) {
                                    newY = comp.getY() + comp.getHeight();
                                } else {
                                    int testY = findDestination( c, newX, newY, DragDirection.UP ).y;
                                    int midY = testY + (oldY-testY)/2;
                                    if( newY < midY ) {
                                        newY = testY;
                                    } else {
                                        newY = oldY;
                                    }
                                    setByFind = true;
                                }
                            }
                            break;
                        }
                    }
                }
            } else if ( diffY > 0 ) { // drag downwards
                for( int y = oldY + height; (y < newY + height) && !setByFind; y++ ) {
                    for( int x = newX ; x < newX + width ; x++ ) {
                        Component comp = parent.getComponentAt( x, y );
                        if( !isIgnoredComponent( comp ) ) {
                            standingFree = false;
                            if ( (initialDirection == null) || (!comp.equals(initialModule)) ) {
                                initialDirection = DragDirection.DOWN;
                                initialModule = comp;
                            }
                            if( initialDirection.isVertical() ) {
                                int inComponent = comp.getY() + (comp.getHeight() / 2) - oldY;
                                if( e.getY() <= inComponent ) {
                                    newY = y - height;
                                } else {
                                    int testY  = findDestination( c, newX, newY, DragDirection.DOWN ).y;
                                    int midY = oldY + (testY - oldY) / 2;
                                    if( newY > midY ) {
                                        newY = testY;
                                    } else {
                                        newY = oldY;
                                    }
                                    setByFind = true;
                                }
                            }
                            break;
                        }
                    }
                }
            } // drag up/down

            /*
             * if no (more) components to overlapAllowed were encontered, the dragged component
             * stands free for itself and no component is approached from any direction
             */
            if( standingFree ) {
                initialDirection = null;
                initialModule = null;
            }

            unIgnoreComponent( c );
            unIgnoreComponent( parent );
        } // if (!overlapAllowed)
        // set the final location
        c.setLocation( newX , newY );
        //c.repaint();
        c.getParent().repaint();
    } // mouseDragged()

    static long cumu = 0;
    static long num = 0;
    /**
     * Checks if the component is to be ignored in regard to overlapping
     * @param c the component in question
     * @return true if the component is to be ignored, false otherwise
     */
    private boolean isIgnoredComponent( Component c ) {
        boolean result = false;
        synchronized( ignoreComponents ) {
            result = ignoreComponents.contains(c);
        }
        return result;
    }

    /**
     * Add a component to the vector of components that can be overlapped.
     * @param c
     */
    public void ignoreComponent( Component c ) {
        synchronized( ignoreComponents ) {
            ignoreComponents.add(c);
        }
    }

    public void unIgnoreComponent( Component c ) {
        synchronized( ignoreComponents ) {
            ignoreComponents.remove( c );
        }
    }

    public void unIgnoreAllModules() {
        synchronized( ignoreComponents ) {
            ignoreComponents.clear();
        }
    }

    /**
     * Find a destination location for the dragged component. This is, in dragging direction,
     * the next location within the component's parent where the dragged component can be
     * placed without overlapping any other components.<br>
     * If there is no such space, <code>findDestination</code> returns the best possible
     * location so far, moving the dragged component as near as possible to any component
     * or the parent's bounds in the dragging direction.<br>
     * However, there's one bug left: If the dragging position is out of bounds in dragging
     * direction, and there are at least two components involved, then <code>findDestination</code>
     * fails to adjust the coordinate to the nearest possible value.
     *
     *
     * @param dragged The dragged component.
     * @param destX x coordinate of the drag target
     * @param destY y coordinate of the drag target
     * @param d dragging direction
     * @return the point to finally move the component to.
     */
    private Point findDestination( Component dragged, int destX, int destY, DragDirection d ) {
        int x = destX;
        int y = destY;
        int w = dragged.getWidth();
        int h = dragged.getHeight();


        // Get the environment of the dragged component: parent and its bounds.
        Component parent = dragged.getParent();
        Rectangle bounds = parent.getBounds();
        Component obstacle = null;


//        System.err.println( "finding (" + destX + "," + destY + " -> " + (destX + dragged.getWidth()) + "," + (destY+dragged.getHeight()) +  ") @ " + DebugUtil.getCallSource());
//        Graphics2D g = (Graphics2D)parent.getGraphics();
//        g.setColor(Color.RED);
//        g.drawRect( destX, destY, dragged.getWidth(), dragged.getHeight()  );
//        try{Thread.sleep(1000);}catch(InterruptedException ie){/**/}

        /*
         * as long as the destination area is not free (of obstacles ;-), inspect the
         * points of the edge of the component to be dragged (the rectangle around
         * the component). If at one of those points is a component, the destination
         * area cannot be used.
         * Then, according to the dragging direction, we need to find another
         * destination area.
         */
        // prepare the flag so the loop will run
        boolean free = false;

        while( !free ) {
            free = true;
            int tx = x;
            int ty = y;
            int tw = w;
            int th = h;
            while( free && (tw > 0) && (th > 0) ){
                int step = 0;
                int stepEnd = Math.max( tw, th );
                while ( free && ( step <= stepEnd) ) {
                    if( step < tw ) {
                        obstacle = parent.getComponentAt( tx+step, ty );
                        if( !isIgnoredComponent( obstacle ) ) {
                                free = false;
                        }
                        if( free ) {
                            obstacle = parent.getComponentAt( tx+tw-step, ty+th );
                            if( !isIgnoredComponent( obstacle ) ) {
                                free = false;
                            }
                        }
                    }

                    if(  step < th ) {
                        obstacle = parent.getComponentAt( tx+tw, ty+step );
                        if( !isIgnoredComponent( obstacle ) ) {
                            free = false;
                        }
                        if( free ) {
                            obstacle = parent.getComponentAt( tx, ty+th-step );
                            if( !isIgnoredComponent( obstacle ) ) {
                                free = false;
                            }
                        }
                    }
                    step++;
                }
                tx++;
                ty++;
                tw-=2; if( tw < 0 ) tw = 0;
                th-=2; if( th < 0 ) th = 0;
            }

            assert obstacle != null;

            /*
             * adjust the coordinates for checking in the next round,
             * according to the dragging direction. If the component was
             * dragged with the mouse outside the bounds of the parent,
             * then adjust the nearest component and the dragged component's
             * position to it.
             */
            if( !free  ) {
                switch( d ) {
                case LEFT:
                    // dragged left, but there is something in the way. Decrease x (drag more left)
                    x--;
                    // if the component that is in the way meanwhile changed (mouse was not ONLY
                    // moved leftwards), then correct this, too and move the dragged component
                    // near it.
                    if( (!obstacle.equals(dragged)) && (!isIgnoredComponent(obstacle) && (!obstacle.equals(findDestComponent)) || (prevY!=destY))) {
                        findDestComponent = obstacle;
//                        System.err.println( "initiate refind L, dragged=" + dragged.getClass().getSimpleName() + ":" + dragged.hashCode() + " @ " + coord(dragged) + ", obstacle=" + obstacle.getClass().getSimpleName() + ":" + obstacle.hashCode() + " @ "+ coord(obstacle));
                        if( prevY != destY ) {
                            prevY = destY;
                            return findDestination( dragged, obstacle.getX() + obstacle.getWidth() , y, DragDirection.LEFT );
                        }
                        return findDestination( dragged, obstacle.getX() - dragged.getWidth() - 1, y, DragDirection.LEFT);
                    }
                    break;
                case RIGHT:
                    x++;
                    if( !isIgnoredComponent(obstacle) && (!obstacle.equals(dragged)) && (!obstacle.equals(findDestComponent) || (prevY!=destY))) {
                        findDestComponent = obstacle;
//                        System.err.println( "initiate refind R, dragged=" + coord(dragged) + ", obstacle=" + obstacle.getClass().getSimpleName() + ":" + obstacle.hashCode() + " @ "+ coord(obstacle));
                        if( prevY != destY ) {
                            prevY = destY;
                            return findDestination( dragged, obstacle.getX() - dragged.getWidth() - 1, y, DragDirection.RIGHT );
                        }
                        return findDestination( dragged, obstacle.getX() + obstacle.getWidth(), y, DragDirection.RIGHT);
                    }
                    break;
                case UP:
                    y--;
                    if( !isIgnoredComponent(obstacle) && (!obstacle.equals(dragged)) && (!obstacle.equals(findDestComponent) || (prevX!=destX))) {
                        findDestComponent = obstacle;
//                        System.err.println( "initiate refind U, dragged=" + coord(dragged) + ", obstacle=" + coord(obstacle));
                        if( prevX != destX ) {
                            prevX = destX;
                            return findDestination( dragged, x, obstacle.getY() + obstacle.getHeight(), DragDirection.UP);
                        }
                        int findY = obstacle.getY() - dragged.getHeight() - 1;
                        if ( findY < parent.getBounds().y) {
                            return findDestination( dragged, x, y, DragDirection.UP );
                        }
                        return findDestination( dragged, x, findY, DragDirection.UP );
                    }
                    break;
                case DOWN:
                    y++;
                    if( !isIgnoredComponent(obstacle) && (!obstacle.equals(dragged)) && (!obstacle.equals(findDestComponent) || (prevX!=destX)) ) {
                        findDestComponent = obstacle;
//                        System.err.println( "initiate refind D, dragged=" + coord(dragged) + ", obstacle=" + coord(obstacle));
                        if( prevX != destX ) {
                            prevX = destX;
                            return findDestination( dragged, x, obstacle.getY() - dragged.getHeight() - 1, DragDirection.DOWN);
                        }
                        return findDestination( dragged, x, obstacle.getY() + obstacle.getHeight(), DragDirection.DOWN );
                    }
                    break;
                case NONE:
                    // nothing to do
                    break;
                }
                if( !isIgnoredComponent(obstacle) && !findDestComponent.equals(obstacle) ) {
                    findDestComponent = obstacle;
                }
            }
            /* there's no use to continue this loop if we already
             * exceeded the parent's bounds.
             */
            if( !bounds.contains( new Rectangle(x,y,w,h) )) {
                return new Point( dragged.getX(), dragged.getY() );
            }
        } // while !free
        // correction for same directions as loop directions
        findDestComponent = null;
        return new Point( x, y );
    } // findDestination


    @SuppressWarnings("unused")
    private String coord( Component c ) {
        return "(" + c.getX() + "," + c.getY() + " -> " + (c.getX() + c.getWidth()) + "," + (c.getY()+c.getHeight()) + ")";
    }

    public void setOverlapAllowed( boolean overlap ) {
        this.overlapAllowed = overlap;
    }

    // other MouseListener methods remain unimplemented
    @Override public void mouseClicked(@SuppressWarnings("unused") MouseEvent e) { /* dummy */ }
    @Override public void mouseEntered(@SuppressWarnings("unused") MouseEvent e) { /* dummy */ }
    @Override public void mouseExited(@SuppressWarnings("unused") MouseEvent e) { /* dummy */  }
    @Override public void mouseMoved(@SuppressWarnings("unused") MouseEvent e) { /* dummy */ }


    /**
     * @return the restrictMovement
     */
    public boolean isRestrictMovement() {
        return restrictMovement;
    }


    /**
     * @param restrictMovement the restrictMovement to set
     */
    public void setRestrictMovement(boolean restrictMovement) {
        this.restrictMovement = restrictMovement;
    }

}