package flexray.gui;

import java.awt.Dimension;
import java.awt.DisplayMode;
import java.awt.EventQueue;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.HeadlessException;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JRootPane;
import javax.swing.RepaintManager;

/**
 * Vollbild-Umschaltung mit und ohne nderung der Anzeigebetriebsart.
 * Keine Objekte werden bei der Umschaltung gelscht.
 *
 * @author Dirk Frsterling
 *
 */
@SuppressWarnings("serial")
public class FullScreenDemo extends JFrame implements ActionListener {

    static FullScreenDemo w = null;

    // GraphicsDevice: Zugriff auf Grafikmodus
    private GraphicsDevice device;
    private boolean canChangeDisplayMode = false;

    // Variablen zur Speicherung der Bildschirmeigenschaften
    // vor der Umschaltung
    private DisplayMode previousDisplayMode;
    private Dimension previousSize;
    private Point previousLocation;
    private boolean previousUndecorated;
    private int previousDecorationStyle;

    // Variablen fr den Vollbild-Knopf.
    private boolean fullscreen = false;
    private JButton fullScreenButton  = null;

    // Variablen fr den Moduswechselknopf
    private boolean changeDisplayMode = false;
    private JButton changeDisplayModeButton = null;

    // Auflsungsauswahlbox
    private JComboBox displayModesBox = null;

    // Ende-Knopf
    private JButton exitButton = null;

    private Dimension fullScreenSize = null;

    /**
     * Innere Klasse, um Anzeigebetriebsarten sortiert
     * zu verwalten.
     *
     * @author Dirk Frsterling
     *
     */
    private class Anzeigebetriebsart implements Comparable<Anzeigebetriebsart> {
        private int breite;
        private int hhe;
        private int bittiefe;
        private int wiederholrate;
        private DisplayMode mode;

        /**
         * Objekt konstruieren mittels eines DisplayMode.
         *
         * @param dm DisplayMode
         */
        public Anzeigebetriebsart( DisplayMode dm ) {
            if( dm == null ) {
                throw new NullPointerException( "null-Betriebsart?" );
            }
            breite = dm.getWidth();
            hhe = dm.getHeight();
            bittiefe = dm.getBitDepth();
            wiederholrate = dm.getRefreshRate();
            mode = dm;
        }

        /**
         * Abholen des DisplayMode
         * @return the current DisplayMode
         */
        public DisplayMode getDisplayMode() {
            return mode;
        }

        /**
         * compareTo fr den Vergleich zur Sortierung. (Comparable)
         * Die Sortierung erfolgt _umgekehrt_ nach Breite, Hhe, Bittiefe
         * und Wiederholrate. 320x200 ist also grer als 800x600!
         *
         * @param c Anzeigebetriebsart im Vergleich
         * @return Wert analog compareTo
         * @see java.lang.Comparable#compareTo( Object )
         */
        public int compareTo( Anzeigebetriebsart c ) {
            if( c == null ) {
                throw new NullPointerException( "Vergleich mit null!" );
            }


            if (c.equals( this )) return 0; // Identitt

            //
            // Nun: DIESES mit dem BERGEBENEN vergleichen
            //

            // Breite ist schon kleiner
            if( breite < c.breite ) return 1;

            // Breite gleich, Hhe kleiner
            if( (breite == c.breite) &&
                (hhe < c.hhe) ) return 1;

            // Breite & Hhe gleich, Bittiefe kleiner
            if( (breite == c.breite) &&
                (hhe == c.hhe) &&
                (bittiefe < c.bittiefe) ) return 1;

            // Alles gleich bis auf kleinere Wiederholrate.
            if( (breite == c.breite) &&
                (hhe == c.hhe) &&
                (bittiefe == c.bittiefe) &&
                (wiederholrate < c.wiederholrate)) return 1;

            // Alles gleich
            if( (breite == c.breite) &&
                (hhe == c.hhe ) &&
                (bittiefe == c.bittiefe) &&
                (wiederholrate == c.wiederholrate )
              ) return 0;

            // Bleibt nur noch: Ich bin grer.
            return -1;
        } // compareTo()

        /**
         * Textreprsentation dem Anzeigebetriebsart zurckgeben
         * @return Text mit den Eigenschaften der Anzeigebetriebsart
         */
        @Override
        public String toString() {
            return breite + "x" + hhe +
            "x" + bittiefe + ", " + wiederholrate + "Hz";
        }
    } // private class Anzeigebetriebsart

    /**
     * Konstruktor. Hier wird das Fenster zusammengebaut:
     * Ein 640x480-Fenster, das zwei Buttons hat.
     *
     */
    public FullScreenDemo() {
        // Device holen
        GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
        try {
            device = ge.getDefaultScreenDevice();
        } catch( HeadlessException he ) {
            System.err.println( "Grafik ohne Grafikkarte? Nein!" );
            System.exit( 1 );
        }

        /*
         * Testen, ob Betriebsartwechsel mglich ist.
         * Da der Wechsel abhngig sein kann von der Vollbilddarstellung
         * des Fensters mittels setFullScreenWindow, mu zunchst diese
         * Darstellung aktiviert werden.
         */
        if( device.isFullScreenSupported() ) {
            device.setFullScreenWindow( this );
            canChangeDisplayMode = device.isDisplayChangeSupported();
            device.setFullScreenWindow( null );
        }

        // Fenstergre bestimmen, Layout auswhlen und
        // Fenster zentriert platzieren
        int myWidth = 640;
        int myHeight = 480;
        setSize( myWidth, myHeight );
        setLayout( new FlowLayout() );
        int nw = device.getDisplayMode().getWidth();
        int h = device.getDisplayMode().getHeight();
        fullScreenSize = new Dimension( nw, h );
        setLocation( nw/2 - (myWidth/2), h/2 - (myHeight/2) );
        setTitle( "Vollbild-Demo" );

        // Knopf: Vollbild
        fullScreenButton = new JButton( "Vollbild" );
        fullScreenButton.setActionCommand( "fullscreen" );
        fullScreenButton.addActionListener( this );
        add( fullScreenButton );

        if( canChangeDisplayMode ) {
            // Knopf: Bildschirmauflsung ndern
            changeDisplayModeButton = new JButton( "Anzeigebetriebsart ndern: " + (changeDisplayMode?"ja":"nein") );
            changeDisplayModeButton.setActionCommand( "cd" );
            changeDisplayModeButton.addActionListener( this );
            changeDisplayModeButton.setEnabled( canChangeDisplayMode );
            add( changeDisplayModeButton );

            // Auflsungen abholen und zum sortieren in ein
            // TreeSet packen
            DisplayMode modes[] = device.getDisplayModes();
            SortedSet<Anzeigebetriebsart> betriebsarten = new TreeSet<Anzeigebetriebsart>();
            for( DisplayMode mode : modes ) {
                if( !mode.equals(device.getDisplayMode()) ) {
                    Anzeigebetriebsart a = new Anzeigebetriebsart( mode );
                    betriebsarten.add(a);
                }
            }

            // Combobox: Bildschirmauflsungen (sortiert)
            displayModesBox = new JComboBox();
            for( Anzeigebetriebsart a: betriebsarten ) {
                displayModesBox.addItem( a );
            }
            add( displayModesBox );
        }

        // Knopf: Ende
        exitButton = new JButton( "Ende" );
        exitButton.setActionCommand( "exit" );
        exitButton.addActionListener( this );
        add( exitButton );

        // Das ist sicher, da der Fenster-Schlie-Knopf nur in der
        // Fensterdarstellung vorhanden ist.
        // Wre das Fenster im Vollbild decorated knnte der Knopf
        // bettigt werden und ggf. bliebe die Anzeigebetriebsart bestehen.
        this.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
        // Aufrumen
        dispose();
    } // FullScreenDemo()

    /**
     * actionPerformed fr die Buttons. Hier passiert die Umschaltung.
     * Da der DisplayMode-Wechsel sowieso einen leeren Bildschirm bedeutet,
     * macht es nix, wenn Swing hier solange stecken bleibt.
     *
     * @param ae Das Event
     */
    public void actionPerformed( ActionEvent ae ) {

        //
        // Kommando "fullscreen"
        // Vollbild an/aus schalten
        //
        if( ae.getActionCommand().equals( "fullscreen" ) ) {
            // Fenster fr den Vorgang unsichtbar machen und
            // fr ungltig erklren.
            setVisible( false );
            invalidate();
            //
            // Teil 1: Vollbild einschalten
            if( !fullscreen ) {
                //
                // aktuelle Fensterinstellungen speichern.
                // Andernfalls ist das Fenster nach der Rckkehr in die
                // Nicht-Maximierte Fensterdarstellung eventuell in
                // einer anderen Gre und/oder an andere Position
                previousSize = getSize();
                previousLocation = getLocation();
                previousDisplayMode = device.getDisplayMode();
                //
                // displayable loswerden mit dispose()
                // (ansonsten funktioniert setUndecorated nicht)
                dispose();
                previousUndecorated = isUndecorated();
                setUndecorated( true ); // Titelbalken und Fensterrahmen ausschalten
                previousDecorationStyle = getRootPane().getWindowDecorationStyle();
                getRootPane().setWindowDecorationStyle(JRootPane.NONE);

                if( changeDisplayMode ) {
                    Anzeigebetriebsart a = (Anzeigebetriebsart)displayModesBox.getSelectedItem();
                    device.setFullScreenWindow( this );
                    device.setDisplayMode( a.getDisplayMode() );
                } else {
                    setSize( fullScreenSize );
                    setLocation(0,0);
                }
                this.setExtendedState( Frame.MAXIMIZED_BOTH );
                setVisible( true ); // um Ergebnis unter Linux zu verbessern...
                //
                // Anzeigebetriebsart-Knopf und ComboBox abschalten,
                // solange eine fremde Betriebsart aktiv ist.
                if( canChangeDisplayMode ) {
                    changeDisplayModeButton.setEnabled( false );
                    displayModesBox.setEnabled( false );
                }
                fullScreenButton.setText( "Fenster" );
                fullscreen = true;
            //
            // Teil 2: Vollbild ausschalten
            } else {
                // Anzeigebetriebsart wiederherstellen, weil setFullScreenWindow( null )
                // trotz Behauptung in der Doku es manchmal nicht schafft, die alte
                // Betriebsart wieder herzustellen.
                if( changeDisplayMode ) {
                    device.setDisplayMode( previousDisplayMode );
                }
                //
                // Vollbild verlassen
                device.setFullScreenWindow( null );
                //
                // Vorherige Fenstergre und -position wiederherstellen
                dispose(); // displayable loswerden
                setUndecorated( previousUndecorated );
                setLocation( previousLocation );
                setSize( previousSize );
                setExtendedState( Frame.NORMAL );
                //getRootPane().setWindowDecorationStyle( previousDecorationStyle );
                //
                // Knpfe wieder einschalten
                if( canChangeDisplayMode ) {
                    changeDisplayModeButton.setEnabled( canChangeDisplayMode );
                    displayModesBox.setEnabled( canChangeDisplayMode );
                }
                fullScreenButton.setText( "Vollbild" );
                fullscreen = false;
                // nderungen haben sich ergeben. Fenster verstecken
                // und Container fr ungltig erklren.
                invalidate();
            }
            setVisible( true ); // Fenster auf jeden Fall wieder darstellen ...
            EventQueue.invokeLater( new Runnable() {
                public void run() {
                    validate();
                    repaint();
                }
            });
        } // kommando "fullscreen"

        //
        // Kommando "cd"
        // aktivieren/deaktivieren der nderung der Anzeigebetriebsart
        //
        if( ae.getActionCommand().equals("cd")) {
            changeDisplayMode = !changeDisplayMode;
            JButton source = (JButton)ae.getSource();
            source.setText( "Anzeigebetriebsart ndern: " + (changeDisplayMode?"ja":"nein") );
        } // kommando "cd"

        //
        // Kommando "exit"
        // Eventuell Anzeigebetriebsart zurcksetzen und dann exit()
        //
        if( ae.getActionCommand().equals( "exit" ) ) {
            if( fullscreen ) {
                if( changeDisplayMode ) {
                    device.setDisplayMode( previousDisplayMode );
                }
                dispose(); // displayable loswerden
                device.setFullScreenWindow( null );
            }
            System.exit( 0 );
        } // kommando "exit"
    } // actionPerformed

    /**
     * Hauptmethode zum Starten der Demo
     * @param args
     */
    public static void main(String[] args) {
        JFrame.setDefaultLookAndFeelDecorated( true );
        RepaintManager.setCurrentManager( new ThreadCheckingRepaintManager( System.err ) );

        EventQueue.invokeLater( new Runnable() {
            public void run() {
                w = new FullScreenDemo();
                w.setVisible( true );
            }
        });

    } // main()
} // AMainWindow
