/*
 * $Id: MemoryBar.java,v 1.1 2009/04/01 13:57:09 sirocic Exp $
 */
package flexray.gui;

import static flexray.gui.MemoryBar.PeakMode.PEAK_MAXIMUM;
import static flexray.gui.MemoryBar.PeakMode.PEAK_MIN;
import static flexray.gui.MemoryBar.PeakMode.PEAK_NONE;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;


/**
 * Module to display a horizontal bar that indicates JVM memory usage.<br>
 * Visual semantics is like this:<br>
 * <ul>
 * <li>GREEN: The "real free" memory: That much is left from the current heap size
 *         up to the maximum heap size. In other words, the right edge represents
 *         the maximum heap size ("-Xmx")</li>
 * <li>YELLOW: The "free" memory: This represents the amount of unused but allocated
 *         Heap. The right edge of the bar starts with "-Xms", the minimum heap size.</li>
 * <li>RED   : This part of the bar represents the allocated and used memory.</li>
 * </ul>
 * <br>
 * There is a peak display for the maximum values and the previous values displayed.
 * The maximum values display never moves from the right to the left. The previous
 * values display will regularily update to the current value.<br>
 * The previous value peak displays will not be drawn
 *
 * @author Dirk Frsterling
 *
 */
public class MemoryBar extends AlphaBlendingComponent {

    private static final long serialVersionUID = 3534658957309342591L;

    // runtime object for autonomous memory bar
    private static Runtime runtime = Runtime.getRuntime();

    private long maxMem;
    private long totMem;
    private long maxTotMem = -1;
    private long minTotMem  = Integer.MAX_VALUE;
    private long freeMem;
    private long maxUsedMem = -1;
    private long minUsedMem = Integer.MAX_VALUE;

    private int maxHeap = -1;
    private int minHeap = Integer.MAX_VALUE;
    private int curUsed = -1;
    private int maxUsed = -1;
    private int minUsed = Integer.MAX_VALUE;

    private Object statLock = new Object();

    private int updateInterval = 1;

    private boolean doPaint = true;
    
    private int desiredWidth = 100;

    /**
     * Enumeration for the peak display modes
     * @author Dirk Frsterling
     */
    public enum PeakMode {
        PEAK_NONE,
        PEAK_MAXIMUM,
        PEAK_MIN,
        PEAK_BOTH;

        /**
         * Binary peakMode check.
         *
         * @param peakMode mode to check against
         * @return true if this PeakMode means that peakMode is enabled
         */
        public boolean enables( PeakMode peakMode ) {
            return ( this.ordinal() | peakMode.ordinal() ) != 0;
        }
    }

    // default: no peaks.
    private PeakMode peakMode = PEAK_NONE;

    // autonomous mode: If true, the MemoryBar works without furher interaction.
    // The memory status is then gathered from the Runtime directly.
    private boolean autonomous = false;

    // Executor and Future for autonomous mode and peak updates
    private ScheduledExecutorService scheduler = null;
    ScheduledFuture<?> scheduledFuture = null;

    static final BufferedImage loHeap = new BufferedImage( 5, 6, BufferedImage.TYPE_INT_ARGB );
    static final BufferedImage hiHeap = new BufferedImage( 5, 6, BufferedImage.TYPE_INT_ARGB );
    static final BufferedImage loUsed = new BufferedImage( 5, 6, BufferedImage.TYPE_INT_ARGB );
    static final BufferedImage hiUsed = new BufferedImage( 5, 6, BufferedImage.TYPE_INT_ARGB );
    BufferedImage myImage = null;

    /**
     * Construtor creates the images for the peak indicators.
     *
     */
    public MemoryBar() {
        int[] poly_x = { 0, 2, 4 };
        int[] bottom_y = { 6, 0, 6 };
        int[] top_y = { 0, 6, 0 };

        Graphics2D g = (Graphics2D)loHeap.getGraphics();
        g.setColor( new Color( 0, 0, 0, 0f ) );
        g.fillRect( 0, 0, 5, 6 );
        g.setColor( Color.YELLOW );
        g.fillPolygon( poly_x, bottom_y ,3);
        g.setColor( Color.BLACK );
        g.drawLine(  0, 5, 2, 0 );
        g.drawLine(  2, 0, 4, 5 );

        g = (Graphics2D)hiHeap.getGraphics();
        g.setColor( new Color( 0, 0, 0, 0f ) );
        g.fillRect( 0, 0, 5, 6 );
        g.setColor( Color.YELLOW );
        g.fillPolygon( poly_x, top_y ,3);
        g.setColor( Color.BLACK );
        g.drawLine(  0, 0, 2, 5 );
        g.drawLine(  2, 5, 4, 0 );

        g = (Graphics2D)loUsed.getGraphics();
        g.setColor( new Color( 0, 0, 0, 0f ) );
        g.fillRect( 0, 0, 5, 6 );
        g.setColor( Color.RED );
        g.fillPolygon( poly_x, bottom_y ,3);
        g.setColor( Color.BLACK );
        g.drawLine(  0, 5, 2, 0 );
        g.drawLine(  2, 0, 4, 5 );

        g = (Graphics2D)hiUsed.getGraphics();
        g.setColor( new Color( 0, 0, 0, 0f ) );
        g.fillRect( 0, 0, 5, 6 );
        g.setColor( Color.RED );
        g.fillPolygon( poly_x, top_y ,3);
        g.setColor( Color.BLACK );
        g.drawLine(  0, 0, 2, 5 );
        g.drawLine(  2, 5, 4, 0 );
    }

    /**
     * ensure minimum size
     *
     * @return Dimension with the current size, at least the minimum size
     */
    @Override
    public Dimension getPreferredSize() {
        int w = getWidth();
        int h = getHeight();
        if( w < desiredWidth ) w = desiredWidth;
        if( h < 16 ) h = 16;
        return new Dimension( w, h );
    }

    public void setWidth( int pWidth ) {
        if( pWidth < 20 ) return;
        desiredWidth = pWidth;
        setSize(pWidth,getHeight());
    }

    public void setPaint( boolean p ) {
        doPaint = p;
    }

    /**
     * Paint this thing without respect to the current L&F
     * @param g the <code>Graphics</code> context in which to paint
     */
    @Override
    public void paintComponent( Graphics g ) {
        if( !doPaint ) return;

        super.paintComponent( g );

        if( (myImage == null) || (myImage.getWidth() != getWidth()) || (myImage.getHeight() != getHeight()) ) {
            myImage = new BufferedImage( getWidth(), getHeight(), BufferedImage.TYPE_INT_ARGB );
        }

        Graphics2D g2d = (Graphics2D)myImage.getGraphics();

        // 1) calculate the pixel-positions of the ends of the
        //    two bars, based on this module's current size.
        int fw = 0;
        int fh = 0;
        double oneGraphPercent = 0;
        double oneMemPercent = 0;
        double totPercent = 0;
        int totPixel = 0;
        double freePercent = 0;
        int freePixel = 0;

        synchronized( statLock ) {
            fw = getWidth();
            fh = getHeight();
            g2d.clearRect( 0, 0, fw, fh );
            oneGraphPercent = ((fw-4))/100.0;
            oneMemPercent = maxMem/100.0;
            totPercent = totMem/oneMemPercent;
            totPixel = (int)Math.round((totPercent * oneGraphPercent));

            minHeap = (int)Math.round( (minTotMem/oneMemPercent) * oneGraphPercent);
            maxHeap = (int)Math.round( (maxTotMem/oneMemPercent) * oneGraphPercent);
            minUsed = (int)Math.round( (minUsedMem/oneMemPercent) * oneGraphPercent);
            maxUsed = (int)Math.round( (maxUsedMem/oneMemPercent) * oneGraphPercent);
            curUsed = (int)Math.round( ((totMem-freeMem)/oneMemPercent) * oneGraphPercent);

            freePercent = freeMem/oneMemPercent;
            freePixel = (int)Math.round((freePercent * oneGraphPercent));

            // paint base border
            g2d.setColor( Color.BLACK );
            g2d.fillRect( 0, 0, fw, fh );

            // paint total memory green
            if( (fw-4 > totPixel) ) {
                g2d.setColor( Color.GREEN );
                g2d.fill3DRect( 2 + totPixel, 2, fw-4-totPixel , fh-4, true);
            }

            // paint heap yellow
            if( freePixel > 0 ) {
                g2d.setColor( Color.YELLOW );
                g2d.fill3DRect( 2, 2, totPixel, fh-4, true );
            }

            // paint used memory red
            if( totPixel > freePixel ) {
                g2d.setColor( Color.RED );
                g2d.fill3DRect( 2, 2, curUsed, fh-4, true);
            }

            // draw peak indicators only if there is enough room
            if( (fh > 15) && (fw > 30) ) {
                // peak heap
                if( peakMode.enables( PEAK_MIN ) && (minHeap < Integer.MAX_VALUE)  ) {
                    g2d.drawImage( loHeap, minHeap-1, fh-2-6, null);
                }
                // maximum heap indicator
                if( peakMode.enables( PEAK_MAXIMUM ) && (maxHeap > 0) ) {
                    g2d.drawImage( hiHeap, maxHeap-1, 2, null);
                }

                // peak
                if( peakMode.enables( PEAK_MIN ) && (minUsed < Integer.MAX_VALUE) ) {
                    g2d.drawImage( loUsed, minUsed-1, fh-2-6, null );
                }
                // maximum used indicator
                if( peakMode.enables( PEAK_MAXIMUM ) && (maxUsed > 0) ) { // && (maxUsed+6 < maxHeap) ) {
                    g2d.drawImage( hiUsed, maxUsed-1, 2, null );
                }
            }
            g.drawImage( myImage, 0, 0, null );
        } // synchronized()
    } // paint()

    /**
     * Enable or disable autonomous mode. In autonomous mode, the MemoryBar
     * will gather the memory status values by itself and update the display
     * accordingly.
     * @param bAutonomous true to enable, false to disable
     */
    public void setAutonomous( boolean bAutonomous ) {
        autonomous = bAutonomous;
        checkScheduler();
    }

    public boolean isAutonomous() {
        return autonomous;
    }

    /**
     * Define which peak displays should be drawn
     * @param pPeakMode
     */
    public void setPeakMode( PeakMode pPeakMode ) {
        peakMode = pPeakMode;
        checkScheduler();
    }

    public PeakMode getPeakMode() {
        return peakMode;
    }

    /**
     * Set the values to display. This has no effect in
     * autonomous mode.
     * @param max A value for Runtime.getRuntime().maxMemory()
     * @param tot A value for Runtime.getRuntime().totalMemory()
     * @param free A value for Runtime.getRuntime().freeMemory()
     */
    public void set( long max, long tot, long free ) {
        if( isAutonomous() ) return;
        synchronized( statLock ) {
            maxMem = max;
            totMem = tot;
            freeMem = free;
            adjustPeakValues();
        }
        update(); // adjust and repaint
    }

    /**
     * Internal method for display update. In autonomous mode this
     * method also fetches current memory information.
     */
    protected void update() {
        if( autonomous ) {
            synchronized( statLock ) {
                maxMem = runtime.maxMemory();
                totMem = runtime.totalMemory();
                freeMem = runtime.freeMemory();
                adjustPeakValues();
            }
        }

        /*
         * update() gets called by the internal blendScheduler,
         * so it WILL be called outside the EDT sometimes.
         * Compensate here for that.
         */
        if( !EventQueue.isDispatchThread() ) {
            EventQueue.invokeLater( new Runnable() {
                public void run() {
                    repaint();
                }
            });
        } else {
            repaint();
        }
    } // update()

    private void adjustPeakValues() {
        if( minTotMem > totMem ) minTotMem = totMem;
        if( maxTotMem < totMem ) maxTotMem = totMem;
        if( minUsedMem > (totMem-freeMem) ) minUsedMem = totMem-freeMem;
        if( maxUsedMem < (totMem-freeMem) ) maxUsedMem = totMem-freeMem;
    }

    /**
     * Check if the Executor is executing the update Runnable.<br>
     * Stop and remove it if unneeded or create and run it if
     * necessary.
     *
     */
    private void checkScheduler() {
        if( autonomous || (peakMode.ordinal() != 0) ) {
            startScheduler();
        } else {
            stopScheduler();
        }
    }

    /**
     * Create an Executor and start the scheduled update.
     */
    private void startScheduler() {
        if( scheduledFuture != null ) return; // already running
        scheduler = Executors.newScheduledThreadPool( 1 );
        scheduledFuture = scheduler.scheduleAtFixedRate(
            new Runnable() {
                public void run() {
                    update();
                }
            }, 0, updateInterval, TimeUnit.SECONDS );
    }

    /**
     * Stop the scheduled updates and remove the Executor.
     */
    private void stopScheduler() {
        if( scheduledFuture == null ) return; // already stopped
        scheduler.shutdown();
        scheduledFuture = null;
        scheduler = null;
    }

    public int getUpdateInterval() {
        return updateInterval;
    }

    public void setUpdateInterval(int updateInterval) {
        this.updateInterval = updateInterval;
    }
}
