/*
 * $Id: ScenarioPanel.java,v 1.1 2009/04/01 13:57:09 sirocic Exp $
 *
 * Created on 29.05.2006 09:30:01
 * by Dirk Frsterling <r@zorbla.de>
 */
package flexray.gui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.HashMap;

import javax.swing.BorderFactory;
import javax.swing.JPanel;

import learningVisualizations.LeviResourceManager;


import flexray.cluster.Cluster;
import flexray.cluster.Node;
import flexray.communication.FlexRayFrame;
import flexray.configuration.ChannelName;
import flexray.configuration.ChannelTopology;
import flexray.scenario.FlexRayScenario;

/**
 * A Panel especially for the display of the FlexRay scenario.
 * This panel draws the network connections
 *
 * @author Dirk Frsterling <r@zorbla.de>
 *
 */
public class ScenarioPanel extends JPanel
                     implements MouseListener,
                                MouseMotionListener {

    private FlexRayScenario scenario;
    private Cluster cluster;
    private HashMap<ChannelName, ActiveStarCouplerComponent> starCouplers = new HashMap<ChannelName, ActiveStarCouplerComponent>();
    private static boolean antiAlias = true;
    private static ScenarioPanel self = null;

    /**
     * Construct a new scenarioPanel
     * @param m
     */
    public ScenarioPanel(final FlexRayScenario m) {
        super(null);
        this.scenario = m;
        this.cluster = scenario.getCluster();
        setBorder(BorderFactory.createEmptyBorder());
        self = this;
        addMouseListener(this);
    }

    /*
     *
     * @param g
     * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
     */
    @Override
    protected void paintComponent(final Graphics g) {
        super.paintComponent(g);
        Graphics2D g2d = (Graphics2D)g;

           /*
            * raster for debugging
            */
        if( false ) {
            Color saveColor = g.getColor();
            g.setColor(Color.LIGHT_GRAY);
            for( int x = 10 ; x < getWidth() ; x+=10)
                g.drawLine(x, 0, x, getHeight());
            g.setColor(Color.RED);
            for( int x = 100 ; x < getWidth() ; x+=100)
                g.drawLine(x, 0, x, getHeight());
            g.setColor(saveColor);
        }

        if( cluster != null ) {
            if( antiAlias ) {
                g2d.setRenderingHint( RenderingHints.KEY_ANTIALIASING,
                        RenderingHints.VALUE_ANTIALIAS_ON);
            }

            drawChannel(g2d, ChannelName.A);
            drawChannel(g2d, ChannelName.B);

            if( antiAlias ) {
                g2d.setRenderingHint( RenderingHints.KEY_ANTIALIASING,
                        RenderingHints.VALUE_ANTIALIAS_OFF);
            }
        } // if channel != null
    }

    /*
     * drawChannel - delegate drawing to BUS and STAR painters
     * @param g2d
     * @param channel
     */
    private void drawChannel( final Graphics2D g2d, final ChannelName channel ) {
        if( scenario.getTopology(channel) == ChannelTopology.BUS)
            drawChannelBus(g2d,channel);
        if( scenario.getTopology(channel) == ChannelTopology.STAR)
            drawChannelStar(g2d,channel);
    }

    /*
     * Draw the lines that represent a channel in BUS topology
     */
    private void drawChannelBus(final Graphics2D g2d, final ChannelName channel ) {
        Stroke savedStroke = g2d.getStroke();
        Stroke noTransStroke = new BasicStroke(2f,0,0,1f,new float[]{4,4},0f);
        Stroke transStroke = new BasicStroke(2.0f);
        Stroke transStrokeBig = new BasicStroke(4.0f);
        Stroke channelStroke = noTransStroke;
        Stroke deadChannelStroke = new BasicStroke(2f,0,0,1f,new float[]{2,16},0f);

        Color channelColor = getForeground();

        // find out the current (sending) node
        final Node currentNode = cluster.getNodeForSlot(channel, cluster.getSlotCounter(channel));
        if( (currentNode != null) && (cluster.getClusterConfiguration().gChannels().contains(channel))) {
            if( currentNode.getNodeConfiguration().pChannels().contains(channel)) {
                FlexRayFrame frFrame = currentNode.getCurrentFrame(channel);
                if( ((frFrame != null) && (frFrame.getLength()>FlexRayFrame.EMPTY_FRAME_LENGTH)) || (cluster.inStaticSegment() && currentNode.isSendingStatic(channel)) ) {
                    channelColor = scenario.getNodeComponent(currentNode).getColor();
                    channelStroke = transStroke;
                }
            }
        }

        g2d.setColor(channelColor);

        ActiveStarCouplerComponent asc = starCouplers.get(channel);
        // determine, where to draw the bus. Start with the position
        // of the ASC and expand the x direction if necessary
        int busY = asc.getY() + (asc.getHeight()/2);
        int maxx = asc.getX() + asc.getWidth() + 10; // we want the bus to "stand out"
        int minx = asc.getX() - 10; // we want the bus to "stand out"

        // find the left- and rightmost nodes and their connection position
        for( final Node node : cluster.getNodes() ) {
            NodeComponent nc = scenario.getNodeComponent(node);
            ChannelName nodeChannels = node.getNodeConfiguration().pChannels();
            if( asc != null && nodeChannels.contains(channel)) {
                int tminx = nc.getX();
                int tmaxx = nc.getX() + nc.getWidth();
                if( tminx < minx ) minx = tminx;
                if( tmaxx > maxx ) maxx = tmaxx;
            }
        }

        // now we know the min and max x positions.

        // draw the main cable.
        if( (asc != null)  && (asc.isVisible())) {
            // draw an outlne of the main cable if it is white.
            if( channelStroke.equals(transStroke) && channelColor.equals(getBackground()) ) {
                g2d.setStroke(transStrokeBig);
                g2d.setColor(getForeground());
                g2d.drawLine(minx, busY, maxx, busY);
                g2d.drawLine(minx,busY-4, minx, busY+4);
                g2d.drawLine(maxx,busY-4, maxx, busY+4);
            }

            g2d.setColor(channelColor);
            if( cluster.getClusterConfiguration().gChannels().contains(channel))
                g2d.setStroke(channelStroke);
            else
                g2d.setStroke(deadChannelStroke);

            g2d.drawLine(minx, busY, maxx, busY);
            g2d.setStroke(transStroke);
            g2d.drawLine(minx,busY-4, minx, busY+4);
            g2d.drawLine(maxx,busY-4, maxx, busY+4);
        }

        for( final Node node : cluster.getNodes() ) {
            NodeComponent nc = scenario.getNodeComponent(node);
            int ncwh = nc.getWidth() / 2;
            int nchh = nc.getHeight() / 2;
            int busConnY = nc.getY() + nchh;

            ChannelName nodeChannels = node.getNodeConfiguration().pChannels();
            if( asc != null && nodeChannels.contains(channel)) {
                if( channelStroke.equals(transStroke) && channelColor.equals(getBackground()) ) {
                    g2d.setStroke(transStrokeBig);
                    g2d.setColor(getForeground());
                    g2d.drawLine(nc.getX()+ncwh, busConnY, nc.getX()+ncwh, busY);
                    g2d.drawOval( nc.getX() + ncwh-3, busY-3, 7, 7);
                }
                g2d.setColor(channelColor);
                if( cluster.getClusterConfiguration().gChannels().contains(channel))
                    g2d.setStroke(channelStroke);
                else
                    g2d.setStroke(deadChannelStroke);

                g2d.drawLine(nc.getX()+ncwh, busConnY, nc.getX()+ncwh, busY);
                g2d.setStroke(transStroke);
                // draw a connection dot
                g2d.fillOval( nc.getX() + ncwh-3, busY-3, 7, 7);
            } // if asc != null
        } // for all nodes

        g2d.setStroke(savedStroke);
    } // drawChannelBus

    /*
     * Draw the lines that represent a channel in STAR topology
     *
     */
    private void drawChannelStar(final Graphics2D g2d, final ChannelName channel ) {
        Stroke savedStroke = g2d.getStroke();
        Stroke noTransStroke = new BasicStroke(2f,0,0,1f,new float[]{4,4},0f);
        Stroke transStroke = new BasicStroke(2.0f);
        Stroke transStrokeBig = new BasicStroke(4.0f);
        Stroke channelStroke = noTransStroke;
        Stroke deadChannelStroke = new BasicStroke(2f,0,0,1f,new float[]{2,16},0f);

        Color channelColor = getForeground();

        // Current (sending) node
        final Node currentNode = cluster.getNodeForSlot(channel, cluster.getSlotCounter(channel));
        if( (currentNode != null) && (cluster.getClusterConfiguration().gChannels().contains(channel))) {
            if( currentNode.getNodeConfiguration().pChannels().contains(channel)) {
                FlexRayFrame frFrame = currentNode.getCurrentFrame(channel);
                if( ((frFrame != null) && (frFrame.getLength()>FlexRayFrame.EMPTY_FRAME_LENGTH)) || cluster.inStaticSegment() ) {
                    channelColor = scenario.getNodeComponent(currentNode).getColor();
                    channelStroke = transStroke;
                }
            }
        }
        ActiveStarCouplerComponent asc = starCouplers.get(channel);
        for( final Node node : cluster.getNodes() ) {
            NodeComponent nc = scenario.getNodeComponent(node);
            int ncwh = nc.getWidth() / 2;
            int nchh = nc.getHeight() / 2;
            ChannelName nodeChannels = node.getNodeConfiguration().pChannels();
            if( asc != null && nodeChannels.contains(channel)) {
                int ascwh = asc.getWidth()/2;
                int aschh = asc.getHeight()/2;
                if( channelStroke.equals(transStroke) && channelColor.equals(getBackground()) ) {
                    g2d.setStroke(transStrokeBig);
                    g2d.setColor(getForeground());
                    g2d.drawLine(asc.getX() + ascwh,
                                 asc.getY() + aschh,
                                  nc.getX() + ncwh,
                                  nc.getY() + nchh);
                }
                g2d.setColor(channelColor);
                if( cluster.getClusterConfiguration().gChannels().contains(channel))
                    g2d.setStroke(channelStroke);
                else
                    g2d.setStroke(deadChannelStroke);
                g2d.drawLine(asc.getX() + ascwh,
                             asc.getY() + aschh,
                              nc.getX() + ncwh,
                              nc.getY() + nchh);
            } // if asc != null
        } // for all nodes
        g2d.setStroke(savedStroke);
    } // drawChannelStar

    /**
     * Add an active star coupler component
     * @param asc
     */
    public void add(ActiveStarCouplerComponent asc) {
        super.add(asc);
        starCouplers.put(asc.getChannel(), asc);
    }

    public static void setAntiAlias(boolean b) {
        antiAlias = b;
        if( self != null ) {
            self.repaint();
        }
    }

    public static boolean isAntiAlias() {
        return antiAlias;
    }


    /*
     * mouselistener / rubberbanding
     */

    public void mouseClicked(MouseEvent e) {/**/}

    private Component lastRepaintComponent = null;
    /**
     * mouseEntered for "inverse highlighting". The highlighting of the nodes
     * does not work precisely if listeners are listening for  mouseExited on
     * the children (some events just don't happen). It works much better
     * mouseEntered on the parent is checked instead
     * @param e
     * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
     */
    public void mouseEntered(MouseEvent e) {
        if( lastRepaintComponent != null ) lastRepaintComponent.repaint();
        Component hitComponent = getComponentAt(e.getPoint());
        if( (hitComponent instanceof NodeComponent) ||
            (hitComponent instanceof CommCycleComponent) ||
            (hitComponent instanceof ActiveStarCouplerComponent)
          ) {
            hitComponent.repaint();
            lastRepaintComponent = hitComponent;
        }
    }

    public void mouseExited(MouseEvent e) {
        if( lastRepaintComponent != null ) lastRepaintComponent.repaint();
        Component hitComponent = getComponentAt(e.getPoint());
        if( (hitComponent instanceof NodeComponent) ||
            (hitComponent instanceof CommCycleComponent) ||
            (hitComponent instanceof ActiveStarCouplerComponent)
          ) {
            hitComponent.repaint();
            lastRepaintComponent = hitComponent;
        }
    }

    public void mousePressed(MouseEvent e) {/**/}

    public void mouseReleased(MouseEvent e) {/**/}

    public void mouseDragged(MouseEvent e) {/**/}

    public void mouseMoved(MouseEvent e) {/* */}
}
