/*
 * $Id: ThreadCheckingRepaintManager.java,v 1.1 2009/04/01 13:57:09 sirocic Exp $
 */
package flexray.gui;

import java.awt.Component;
import java.io.PrintStream;
import java.util.LinkedList;
import java.util.concurrent.atomic.AtomicInteger;

import javax.swing.JComponent;
import javax.swing.RepaintManager;
import javax.swing.SwingUtilities;

/**
 * Based on an idea and implementation by Scott Delap. <br>
 * See http://www.clientjava.com/blog/2004/08/31/1093972473000.html
 * <br><br>
 * This repaint manager checks if a request came from within the Event Dispatch Thread,
 * and if it didn't prints some usefull information to correct the problem.<br>
 * By default, this repaint manager prints a call tree and a component tree down to the
 * current component if a request outside EDT is detected.
 * <br><br>
 * Using this repaint manager as a replacement for the Swing RepaintManager:
 * <pre>
 *         RepaintManager.setCurrentManager( new ThreadCheckingRepaintManager( ) );
 * </pre>
 *
 * @author Dirk Frsterling
 *
 */
public class ThreadCheckingRepaintManager extends RepaintManager {

    private int indentationDepth = 0;

    private boolean checkOnlyShowing = true;
    private boolean printCallTree = true;
    private boolean printComponentTree = true;
    private boolean simpleComponentTree = true;

    // AtomicInteger just because of the increment method.
    private static AtomicInteger faultcount = new AtomicInteger( 1 );

    private PrintStream output = System.out;


    /**
     * Just create the ThreadCheckingRepaintManager
     */
    public ThreadCheckingRepaintManager() {
        super();
    }

    /**
     * Create the ThreadCheckingRepaintManager selecting its behaviour
     * on non-showing components.
     *
     * @param pCheckOnlyShowing check only showing components?
     */
    public ThreadCheckingRepaintManager( boolean pCheckOnlyShowing ) {
        this( pCheckOnlyShowing, System.out );
    }

    /**
     * Create the ThreadCheckingRepaintManager selecting the
     * PrintStream to print to
     *
     * @param pOutput print the output to this PrintStream
     */
    public ThreadCheckingRepaintManager( PrintStream pOutput ) {
        this( true, pOutput );
    }

    /**
     * Create the ThreadCheckingRepaintManager selecting its behaviour
     * on non-showing components and the PrintStream to print to
     *
     * @param pCheckOnlyShowing  check only showing components?
     * @param pOutput print the output to this PrintStream
     */
    public ThreadCheckingRepaintManager( boolean pCheckOnlyShowing, PrintStream pOutput ) {
        super();
        checkOnlyShowing = pCheckOnlyShowing;
        output = pOutput;
    }

    /* (non-javadoc)
     * @see javax.swing.RepaintManager#addInvalidComponent(JComponent)
     */
    @Override
    public synchronized void addInvalidComponent(JComponent jComponent) {
        checkEDT(jComponent);
        super.addInvalidComponent(jComponent);
    }

    /* (non-javadoc)
     * @see javax.swing.RepaintManager#addDirtyRegion(JComponent,int,int,int,int)
     */
    @Override
    public synchronized void addDirtyRegion(JComponent jComponent, int i, int i1, int i2, int i3) {
        checkEDT(jComponent);
        super.addDirtyRegion(jComponent, i, i1, i2, i3);
    }

    /**
     * Check if we are in the EDT. If so, everything is fine. If not, write a Warning
     * and optional print the component tree and/or write a call trace
     *
     * @param c current component
     */
    private void checkEDT( JComponent c ) {
        if (!SwingUtilities.isEventDispatchThread() && (c.isShowing() | !checkOnlyShowing)) {
            synchronized( faultcount ) {
                output.printf( ">>>>> Outside EDT #%06d", Integer.valueOf( faultcount.get() ) );
                Exception e = new Exception();
                StackTraceElement[] st =  e.getStackTrace();
                // only the last line - the "root" of the problem ?
                StringBuffer sb = new StringBuffer( st[st.length-1].toString() );
                if( printCallTree ) {
                    output.print( " >>>>>" );
                    printCallTree();
                } else {
                    output.println( " [ " + sb + " ]");
                }
                if( printComponentTree ){
                    printComponentTree( c );
                }
                if( printCallTree || printComponentTree)
                    output.printf( "<<<<< Outside EDT #%06d <<<<<\n\n", Integer.valueOf(faultcount.get()) );
                faultcount.addAndGet( 1 );
            }
        }
    } // checkEDT

    /**
     * Print a "call tree". This is basically a stack trace of a new
     * Exception in reverse order, formatted treelike and stripped of
     * the java* classes as well as the current class.<br>
     * This leaves us with something that usually is the essence of
     * where something happened here. <br>
     * Example:
     * <pre>
     * public class Foo extends JComponent {
     *       // something...
     *    }
     * }
     *
     * public class Bar {
     *    public Bar() {
     *       Foo f = new Foo();
     *       f.setVisible();
     *    }
     *
     *    public static void main( String[] args ) {
     *       Bar b = new Bar();
     *       // do something
     *    }
     * }
     * </pre>
     * If, finally paintComponent if JComponent calls the RepaintManager, a stack
     * trace would contain many entries from <code>javax.swing.*</code> and
     * <code>java.awt.*</code>. <code>printCallTree()</code> would just print:
     * <pre>
     *            Bar.main(Bar.java:8)
     *               Bar.<init>(Bar.java:4)
     * </pre>
     * and that's it. So you know immediately every place where you can repair the problem.
     */
    private void printCallTree() {
        resetIndentationDepth();
        StackTraceElement[] st = new Exception().getStackTrace();
        output.println();
        for( int i = st.length-1 ; i > 0 ; i-- ) {
            StackTraceElement ste = st[i];
            String className = ste.getClassName();
            if( (className.indexOf(this.getClass().getSimpleName()) < 0) &&
                    (className.indexOf("java")!= 0) ) {
                printIndentation();
                output.println( ste + "#" + ste.getMethodName() + "(" + ste.getFileName() + ":" + ste.getLineNumber() + ")");
                incrementIndentation();
            }
        }
    } // printCallTree()

    /**
     * Print the component tree down to the given component. In simple
     * mode, component names will be prefixed with two letters in brackets,
     * like <code>[sV]</code>. This is an <code>s</code> for <code>isShowing()</code> and
     * a <code>v</code> for <code>isVisible()</code>. Lower case letter means <code>false</code>
     * and upper case letter means <code>true</code>. So, the example above means: the Component
     * is visile but not showing.
     *
     * @see java.awt.Component#isShowing()
     * @see java.awt.Component#isVisible()
     * @param pC Component to print the (parental) tree for.
     */
    private void printComponentTree(Component pC) {
        Component c = pC;
        resetIndentationDepth();
        LinkedList<Component> components = new LinkedList<Component>();
        for (; c != null; c = c.getParent()) {
            components.addFirst(c);
        }
        if( components.size() > 0 ) {
            output.println( "      ________________________________________\n" );
            for (Component component : components) {
                printIndentation();
                component.isShowing();
                if( simpleComponentTree ) {
                    output.print( "["+(component.isShowing()?"S":"s") + (component.isVisible()?"V":"v") + "] " );
                    output.println( component.getClass().getName().trim() + " ");
                } else {
                    output.println( component );
                    printIndentation();
                    output.println( "[ " + (component.isShowing()?"":"not ") + "showing, " + (component.isVisible()?"":"in") + "visible ]" );
                }
                incrementIndentation();
            }
        }
    } // printComponentTree

    /*------------------------------------------------------------*
     * helpers for printing the tree indentation
     *------------------------------------------------------------*/

    // 51 spaces. requests over 51 will remain at 51
    static StringBuffer spaces =
    new StringBuffer("                                                   ");
    private void printIndentation() {
        output.print( spaces.subSequence( 0, ( indentationDepth + 6 ) % spaces.length() ) );
    }

    private void resetIndentationDepth() {
        indentationDepth = 0;
    }

    private void incrementIndentation() {
        indentationDepth += 3;
    }

    @SuppressWarnings("unused")
    private void decrementIndent() {
        if( indentationDepth >= 3 ) indentationDepth -= 3;
    }

    /*------------------------------------------------------------*
     * autogenerated getters and setters
     *------------------------------------------------------------*/

    public boolean isCheckOnlyShowing() {
        return checkOnlyShowing;
    }

    public void setCheckOnlyShowing(boolean checkOnlyShowing) {
        this.checkOnlyShowing = checkOnlyShowing;
    }

    public boolean isPrintCallTree() {
        return printCallTree;
    }

    public void setPrintCallTree(boolean printCallTree) {
        this.printCallTree = printCallTree;
    }

    public boolean isPrintComponentTree() {
        return printComponentTree;
    }

    public void setPrintComponentTree(boolean printComponentTree) {
        this.printComponentTree = printComponentTree;
    }

    public boolean isSimpleComponentTree() {
        return simpleComponentTree;
    }

    public void setSimpleComponentTree(boolean simpleComponentTree) {
        this.simpleComponentTree = simpleComponentTree;
    }

    public void setOutput(PrintStream output) {
        this.output = output;
    }

} // ThreadCheckingRepaintManager
