/*
 * $Id: TurningKnob.java,v 1.1 2009/04/01 13:57:09 sirocic Exp $
 */
package flexray.gui;

import static java.lang.Math.PI;
import static java.lang.Math.cos;
import static java.lang.Math.sin;

import java.awt.AWTException;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.RenderingHints;
import java.awt.Robot;
import java.awt.Stroke;
import java.awt.Transparency;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;

import javax.swing.JButton;
import javax.swing.JToolBar;
import javax.swing.SwingConstants;
import javax.swing.event.AncestorEvent;
import javax.swing.event.AncestorListener;

import learningVisualizations.LeviModule;
import learningVisualizations.gui.LeviButton;


import flexray.util.DebugUtil;


/**
 *
 * @author Dirk Frsterling
 *
 */
public class TurningKnob extends LeviButton
                      implements SwingConstants,
                                 MouseMotionListener,
                                 MouseListener,
                                 AncestorListener {

    private static final long serialVersionUID = -588079421897557388L;

    private static final double PI180 = (PI / 180.0);

    private int posX = 5;
    private int posY = 5;
    private int width = 15;
    private int height = 15;
    private int drawwidth = 0;
    private int drawheight = 0;
    private int angle = 0;
    private int displacement = -225;
    private int maxAngle = 270;
    private static Color highlightColor;
    private Robot robot;
    private String text = null;
    private final StringBuffer tText = new StringBuffer();
    private int orientation = -1;
    private Font hFont;
    private Font vFont;
    private Container parent;

    /**
     * Must define explicit default constructor to allow
     * default TurningKnobs
     */
    public TurningKnob(String keyLabel, LeviModule module) {
        super(keyLabel, module);
        setOpaque( true );
        setAlignmentX(LEFT_ALIGNMENT);
        setAlignmentY(CENTER_ALIGNMENT);
        addMouseMotionListener(this);
        addMouseListener(this);
        highlightColor = new Color(255,204,0);
        try {
            robot = new Robot();
        } catch (AWTException ae) {
            robot = null;
        }
    }

    /**
     * Constructor for a TurningKnob with a specific size
     * @param kWidth size of the bounding rectangle of this TurningKnob
     */
    public TurningKnob( String keyLabel, LeviModule module, int kWidth ) {
        this(keyLabel, module);
        width = kWidth;
        height = width;
        drawwidth = width - (2*posX);
        drawheight = width - (2*posY);
        setMaximumSize(new Dimension(width,height));
        setMinimumSize(new Dimension(width,height));
    }

    /* ------------------------------------------------------------ *
     *                      Module methods
     * ------------------------------------------------------------ */

    /*  (non-javadoc)
     * @see java.awt.Component#getPreferredSize()
     */
    @Override
    public Dimension getPreferredSize() {
        return new Dimension( width, height );
    }

    /**
     * Paint method for the TurningKnob<br>
     *
     * @author Dirk Frsterling
     * @param g the <code>Graphics</code> context in which to paint
     */
    @Override
    public void paintComponent( Graphics g ) {

        if( parent == null ) {
            parent = getParent();
            if( parent instanceof JToolBar ) {
                ((JToolBar)parent).addAncestorListener(this);
                setOrientation(((JToolBar)parent).getOrientation());
                hFont = g.getFont().deriveFont(7f);
                vFont = g.getFont().deriveFont(11f);
                parent.validate();
            }
        }

        super.paintComponent( g );

        Graphics2D g2d = (Graphics2D)g;
        // make circles look good
        g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
        g2d.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);

        /*
         * fill the complete background (opaque) or just the
         * circular shape (not opaque)
         */
        g2d.setColor( getBackground() );
        if( isOpaque() ) {
            g2d.fillRect( 0, 0, width, height );
        } else {
            g2d.fillOval( posX, posY, drawwidth, drawheight );
        }


        /*
         * draw the knob interna
         */

        Stroke savedStroke = g2d.getStroke();

        if( drawwidth != drawheight ) {
            DebugUtil.println("no circle:" + drawwidth + "x" + drawheight);
        }

        /*
         * Position calculations
         */
        int radius = drawwidth / 2;
        int cx = posX + radius;
        int cy = posY + radius;
        // minimum (left) range marker
        double rad = (0-displacement) * PI180;
        double minMarkX1 = sin( rad ) * radius;
        double minMarkY1 = cos( rad ) * radius;
        double bigMinMarkX1 = sin( rad ) * (radius*2);
        double bigMinMarkY1 = cos( rad ) * (radius*2);
        double minMarkX2 = sin( rad ) * (radius + 5);
        double minMarkY2 = cos( rad ) * (radius + 5);
        // current value marker
        rad = (angle - displacement) * PI180;
        double x1 = sin( rad ) * radius;
        double y1 = cos( rad ) * radius;
        double bigX1 = sin(rad) * (radius*2);
        double bigY1 = cos(rad) * (radius*2);
        // maximum (right) range marker
        rad = (maxAngle-displacement) * PI180;
        double maxMarkX1 = sin( rad ) * radius;
        double maxMarkY1 = cos( rad ) * radius;
        double bigMaxMarkX1 = sin( rad ) * (radius*2);
        double bigMaxMarkY1 = cos( rad ) * (radius*2);
        double maxMarkX2 = sin( rad ) * (radius + 5);
        double maxMarkY2 = cos( rad ) * (radius + 5);

        Stroke knobStroke = new BasicStroke(3f,BasicStroke.CAP_ROUND,BasicStroke.JOIN_BEVEL);

        /*
         * Arc2D and Graphics[2D]#drawArc are not usable to draw "stable" arcs.
         * "stable" means, that all segments with the same boundaries overlap perfectly in
         * the same regions.. The Java arcs are not stable: Two arcs of the same circle
         * have different pixels in the same parts.Thus, drawing an animated growing
         * arc using the java arcs gives a jumpy animation.
         * Solution: Draw complete circles with appropriate clipping.
         */
        g2d.setColor(highlightColor);
        g2d.setStroke(knobStroke);
        // clipping polygon for left hand arc
        Polygon poly = new Polygon();
        poly.addPoint(cx,cy);
        poly.addPoint(cx+(int)bigMinMarkX1, cy-(int)bigMinMarkY1+1);
        if( (cx+x1) > (cx-cx/2))
            poly.addPoint(0, 0);
        if( (cx+x1) > (cx+cx/2))
            poly.addPoint(width, 0);
        poly.addPoint(cx+(int)bigX1,cy-(int)bigY1);
        g2d.setClip(poly);
        g2d.drawOval(posX+1, posY+1, drawwidth-1, drawheight-1);

        g2d.setStroke(knobStroke);
        g2d.setColor(getForeground());
        // clipping polygon for right hand arc
        poly = new Polygon();
        poly.addPoint(cx,cy);
        poly.addPoint(cx+(int)bigMaxMarkX1, cy-(int)bigMaxMarkY1+1);
        if(  ((cx+x1) < (cx+cx/2)) )
            poly.addPoint(width,0);
        if( (cx+x1) < (cx-cx/2) )
            poly.addPoint(0,0);
        poly.addPoint(cx+(int)bigX1,cy-(int)bigY1);
        g2d.setClip(poly);
        g2d.drawOval(posX+1, posY+1, drawwidth-1, drawheight-1);

        g2d.setClip(0,0,width,height);

        g2d.setStroke(knobStroke);

        // minimum (left) range marker
        g2d.setColor(highlightColor);
        g2d.drawLine(cx + (int)minMarkX1, cy - (int)minMarkY1,
                   cx + (int)minMarkX2, cy - (int)minMarkY2);
        g2d.setColor(getForeground());


        // maximum (right) range marker
        g2d.drawLine(cx + (int)maxMarkX1, cy - (int)maxMarkY1,
                   cx + (int)maxMarkX2, cy - (int)maxMarkY2);


        int vDisplace;
        if(orientation == VERTICAL){
            g2d.setFont(vFont);
            vDisplace = 2;
        } else {
            g2d.setFont(hFont);
            vDisplace = 0;
        }

        if( text == null) tText.append( angle + "" );
        else tText.append(text);

        FontMetrics f = g.getFontMetrics();
        int fx, fy , fw;
        fw = (int)(f.getStringBounds( tText.toString(), g ).getWidth());
        fx = ( width / 2 )-( fw / 2 ) + 1;
        fy = height - ( f.getLeading() + f.getAscent() ) / 2;
        g.setColor( getForeground() );
        g.drawString( tText.toString(), fx , fy + vDisplace );

        // current position marker
        // half value correction for the "jumpy" tip of the hand
        rad = (angle - displacement) * PI180;
        x1 = sin( rad ) * (radius-0.5);
        y1 = cos( rad ) * (radius-0.5);

        tText.delete(0, tText.length());

        g2d.drawLine( cx, cy /* + (int)x1, cy - (int)y1*/,
                      cx + (int)x1, cy - (int)y1 );

        g2d.setStroke(savedStroke);

    } // paintComponent()


    public static void setHighlightColor(Color color) {
        if( color == null )
            highlightColor = new Color(255,204,0);
        else
            highlightColor = color;
    }

    /**
     * Set the orientation.
     * @param newOrientation
     */
    public void setOrientation(int newOrientation) {
        if( newOrientation != orientation ) {
            switch(newOrientation) {
            case HORIZONTAL:
                orientation = newOrientation;
                height = width;
                break;
            case VERTICAL:
                orientation = newOrientation;
                height = width+10;
                break;
            default:
                throw new IllegalArgumentException("orientation must be one of: VERTICAL, HORIZONTAL");
            }
            Dimension dimension = new Dimension(width,height);
            setPreferredSize(dimension);
            setMaximumSize(dimension);
            setMinimumSize(dimension);
            invalidate();
        }
    } // setOrientation

    /* ------------------------------------------------------------ *
     *            Control specific (scenario) methods
     * ------------------------------------------------------------ */

    /**
     * Set the angle that should be displayed by this TurningKnob.<br>
     * Please note that the actual display depends on the <tt>growing</tt>
     * value if in <tt>DOUBLECIRCLE</tt> mode.
     *
     * @param angle the angle to display
     */
    public void setAngle( int angle ) {
        if( (angle < 0) || angle > 360 )
            throw new IllegalArgumentException( "angle must be between 0 and 360" );
        this.angle = angle;
        repaint();
    }

    public int getAngle() {
        return angle;
    }

    /**
     * Let this TurningKnob display an angle that corresponds to
     * the given percent value.<br>
     * Please note that the actual display depends on the <tt>growing</tt>
     * value if in <tt>DOUBLECIRCLE</tt> mode.
     *
     * @param percent percentage this symbol should draw.
     */
    public void setPercent( int percent ) {
        if( (percent < 0) || percent > 100 )
            throw new IllegalArgumentException( "percent must be between 0 and 100" );
        this.angle = (int)(3.6 * percent);
        repaint();
    }


    /**
     * Set the text to display in the lower part of the button
     * (may be null)
     * @param newText
     */
    @Override
    public void setText(String newText) {
        text = newText;
        repaint();
    }

    int prevX = 0;
    int prevY = 0;
    Cursor cursor;
    Cursor emptyCursor = null;
    boolean cursorSet = false;

    /* ------------------------------------------------------------ *
     *                     Mouse[Motion]Listener
     * ------------------------------------------------------------ */

    /**
     * Mouselistener part to implement the control
     * @param e
     * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
     */
    public  void mouseDragged(MouseEvent e) {
        if(!cursorSet) {
            setCursor(emptyCursor);
            cursorSet = true;
        }
        // synchronize to prevent handling of mouseDragged from
        // the robot#mouseMove action
        synchronized(robot) {
            int diffX = prevX - e.getX();
            int diffY = prevY - e.getY();
            int diff = diffX;
            if( Math.abs(diffX) < Math.abs(diffY))
                diff = -diffY;
            if(diff != 0) {
                int newAngle = angle-diff;
                if(newAngle < 0)
                    newAngle = 0;
                if(newAngle > maxAngle)
                    newAngle = maxAngle;
                int oldAngle = angle;
                if( oldAngle != newAngle ){
                    setAngle(newAngle);
                    firePropertyChange("angle", oldAngle, newAngle);
                }
                // put back the mouse pointer to where it was
                robot.mouseMove(getLocationOnScreen().x + prevX,
                        getLocationOnScreen().y + prevY);
            }
        }
    } // mouseDragged

    public void mouseMoved(MouseEvent e) {/**/}

    public void mouseClicked(MouseEvent e) {/**/}

    public void mouseEntered(MouseEvent e) {/**/}

    public void mouseExited(MouseEvent e) {/**/}

    /**
     * If the mouse is pressed, record the position for reference
     * and change the cursor to an empty (translucent) cursor
     * @param e
     * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
     */
    public void mousePressed(MouseEvent e) {
        cursor = getCursor();
        prevX = e.getX();
        prevY = e.getY();
        if(emptyCursor == null) {
            BufferedImage cursorImage = new BufferedImage(1,1,Transparency.TRANSLUCENT);
            emptyCursor = getToolkit().createCustomCursor(cursorImage, new Point(0,0), "empty");
        }
    }

    /**
     * reset the cursor to the previous cursor if the mouse is released
     * @param e
     * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
     */
    public void mouseReleased(MouseEvent e) {
        setCursor(cursor);
        cursorSet = false;
    }

    /* ------------------------------------------------------------ *
     *                      AncestorListener
     * ------------------------------------------------------------ */


    /**
     * when the ancestor (JToolBar) was added (made visible) then we
     * should change our alignment
     * @param event
     * @see javax.swing.event.AncestorListener#ancestorAdded(javax.swing.event.AncestorEvent)
     */
    public void ancestorAdded(AncestorEvent event) {
        if( event.getAncestor() instanceof JToolBar)
            setOrientation(((JToolBar)event.getAncestor()).getOrientation());
    }

    public void ancestorMoved(AncestorEvent event) {/**/}
    public void ancestorRemoved(AncestorEvent event) {/**/}




} // class TurningKnob
