package flexray.util;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

public class DebugUtil {

    private static boolean printMark = true;
    private static boolean printPrint = true;
    private static boolean printHint = true;


    private static SimpleDateFormat formatter
        = new SimpleDateFormat ("EEE MMM dd HH:mm:ss yyyy", Locale.ENGLISH);

    private static StringBuffer dashes = new StringBuffer(160);
    static {
        dashes.append("--------------------"); // 20
        dashes.append(dashes); // 40;
        dashes.append(dashes); // 80;
        dashes.append(dashes); // 160;
    }

    /**
     * Return the calling point (just like a single line stack trace) of the method that
     * just used getCaller(). That means, that
     * <pre>
     * public class Foo {
     *    public void bar() {
     *       System.out.println( DebugTools.getCaller() );
     *    }
     *    public static void main( String args[] ) {
     *       Foo foo = new Foo();
     *       foo.bar();
     *    }
     * }
     * </pre>
     * will print something like:
     * <pre>
     * com.plete.package.Foo.main(Foo.java:7)
     * </pre>
     *
     * @return single line of stack trace indicating the where source of a method call.
     */
    public static String getCaller() {
        // return the stack trace element two steps back from this Exception:
        // one step to reach our own caller and another to find the caller's caller.
        return getCaller(2);
    }

    /**
     * Get a single line of stack trace which is <tt>index</tt>
     * steps in the past.
     * @param index
     * @return single line of stack trace
     */
    public static String getCaller(int index) {
        return new Exception().getStackTrace()[index].toString();
    }

    /**
     * Get a single line of stack trace which is <tt>index</tt>
     * steps in the past.
     * @param index
     * @param simple - indicate if just the class name should be returned.
     * @return single line of stack trace
     */
    public static String getCaller(int index, boolean simple) {
        if(simple)
            return new Exception().getStackTrace()[index].getClassName();
        return getCaller(index+1);
    }

    /**
     * Basically, this is a cleaned up indented reverse stack trace down to the
     * point where the method was called that just used getCallerHistory.
     *
     * @param length How long should the history go back?
     * @return a String with a formatted caller history
     */
    public static String getCallerHistory( int length ) {
        StackTraceElement[] stackTrace = new Exception().getStackTrace();
        StringBuffer callerHistory = new StringBuffer("---[ caller history ]---\n");
        StringBuffer indentation = new StringBuffer();
        int count = 0;
        for( int i = stackTrace.length-1 ; i > 1; i-- ) {
            StackTraceElement stackTraceElement = stackTrace[i];
            String className = stackTraceElement.getClassName();
            if( (className.indexOf("javxa") != 0) ) {
                callerHistory.append(indentation).append( stackTrace[i].toString() ).append("\n");
                indentation.append( "   " );
                count++;
            }
            if((length > 0) && (count >= length)) break;
        }
        return callerHistory.append("---").toString();
    }

    /**
     * Convenience method to get a complete caller history
     * @return a String with a formatted caller history
     */
    public static String getCallerHistory() {
        return getCallerHistory(-1);
    }

    /**
     * Print a formatted message indicating the whereabouts
     * @param message
     */
    public static void hint( String message ) {
        if( !printHint ) return;
        String r = formatter.format(new Date());
        System.err.println(r + " " + message);
        System.err.println(r + "     " + getCaller(2));
        System.err.println(r + "        " + getCaller(3));
    }


    /**
     * Print a simple message preceded by date/time, including the
     * Caller's full qualified class name
     *
     * @param message the message to be printed.
     */
    public static void println( String message ) {
        if( !printPrint ) return;
        String r = formatter.format(new Date());
        System.err.println( r + " [ " + getCaller(2,true) + " ]: " + message );
    }


    /**
     * Print a simple message preceded by date/time, including the
     * Caller's full whereabouts.
     * This allows for example eclipse to transfer the developer
     * directly to the line containing the "mark" call.
     *
     * @param message the message to be printed.
     */
    public static void mark( String message ) {
        if( !printMark ) return;
        String r = formatter.format(new Date());
        String s1 = " | " + getCaller(2,false);
        String s2 = " | " + message;
        int maxlen = Math.max(s1.length(), s2.length()) - 1;
        maxlen = Math.min(maxlen, 160);
        System.err.println( r + " +" + dashes.subSequence(0, maxlen));
        System.err.println( r + s1 );
        System.err.println( r + s2);
        System.err.println( r + " +" + dashes.subSequence(0, maxlen));
    }

    /**
     * silence or unsilence the whole println() and mark() stuff
     * @param b
     */
    public static void setSilent(boolean b) {
        if( b ) {
            printMark = false;
            printPrint = false;
            printHint = false;

        } else {
            printMark = true;
            printPrint = true;
            printHint = true;
        }
    }

    /**
     * silence or unsilence the mark() stuff
     * @param b
     */
    public static void setMarkSilent(boolean b) {
        if( b ) printMark = false;
        else printMark = true;
    }

    /**
     * silence or unsilence the println() stuff
     * @param b
     */
    public static void setPrintSilent(boolean b) {
        if( b ) printPrint = false;
        else printPrint = true;
    }

    /**
     * silence or unsilence the hint() stuff
     * @param b
     */
    public static void setHintSilent(boolean b) {
        if( b ) printHint = false;
        else printHint = true;
    }

    /**
     * Just wait for a few seconds, ignoring the InterruptedException
     * @param secs
     */
    public static void sleepSec(int secs) {
        try {
            Thread.sleep(1000*secs);
        } catch( InterruptedException ie ) {
            //
        }
    } // sleepSec

}
